/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/************************************************
             Cube.cpp

 ************************************************/

#include <iostream>
#include <fstream>
#include <exception>
#include <algorithm>

#include "CUBException.h"
#include "Metric.h"
#include "Module.h"
#include "Region.h"
#include "Callsite.h"
#include "Cnode.h"
#include "Machine.h"
#include "Thread.h"

#include "Cube.h"
#include "Space.h"
#include "Severity.h"
#include "Parser.h"

using namespace std;
using namespace cube;


// 3 types of user data supported
#define CALL_TREE_W_LINE_NUM   0
#define CALL_TREE_WO_LINE_NUM  1
#define FLAT_TREE_ONLY         2

Cube::Cube() {
  space = new Space();
  severity = new Severity();
  user_data_type = -1; // by default
}

Cube::~Cube() {
  // delete space; !!! seg fault
  // delete severity; seg fault on solaris
}

Cube::Cube(const Cube& cube) {
  
}

Cube::Cube(Space* space, Severity* sev) {
  this->space = space;
  severity = sev;
}

// CUBE_CT_AGGR -> SUM or MAX
void Cube::def_attr(string key, string value) {
  space->def_attr(key, value);
}

string Cube::get_attr(std::string key) {
  return space->get_attr(key);
}

void Cube::def_mirror(std::string url) {
  space->def_mirror(url);
}
vector<string>&  Cube::get_mirrors() {
  return space->get_mirrors();
}

const Metric* Cube::def_met(string name, string uom, string url,
			    string descr, const Metric* parent) { 
  return space->def_met(name, uom, url, descr, const_cast<Metric*>(parent)); 
}

const Module* Cube::def_module(string name) {
  return space->def_module(name);
}

const Region* Cube::def_region(string name, long begln, long endln, 
			       string man_url, string descr, const Module* mod) {
  return space->def_region(name, begln, endln, man_url, descr, const_cast<Module*>(mod));
}

const Callsite* Cube::def_csite(const Module* mod, int line, const Region* callee) {
  return space->def_csite(const_cast<Module*>(mod), line, const_cast<Region*>(callee));
}

const Cnode* Cube::def_cnode(const Callsite* csite, const Cnode* parent) {
  if (user_data_type == -1) user_data_type = CALL_TREE_W_LINE_NUM;
  if (user_data_type != CALL_TREE_W_LINE_NUM) {
    cerr << "Error: not building call-tree with line number data!" << endl;
  }
  return space->def_cnode(const_cast<Callsite*>(csite), const_cast<Cnode*>(parent));
}

const Cnode* Cube::def_cnode(const Region* region, const Cnode* parent) {
  if (user_data_type == -1) user_data_type = CALL_TREE_WO_LINE_NUM;
  if (user_data_type != CALL_TREE_WO_LINE_NUM) {
    cerr << "Error: not building call-tree without line number data!" << endl;
  }
  // after passing error checking
  Callsite* v_csite = const_cast<Callsite*>(def_csite(region->get_mod(), -1, region));
  const Cnode* ret =  space->def_cnode(v_csite, const_cast<Cnode*>(parent));
  // correct the module for the fake csite: inherits its caller's module
  if (parent != NULL) {
    v_csite->set_mod
      (const_cast<Module*>(parent->get_csite()->get_caller()->get_mod()));
  }
  return ret;
}

const Machine* Cube::def_mach(string name) {
  return space->def_mach(name);
}

const Node* Cube::def_node(string name, const Machine* mach) {
  return space->def_node(name, const_cast<Machine*>(mach));
}

const Process* Cube::def_proc(string name, int rank, const Node* node) {
  return space->def_proc(name, rank, const_cast<Node*>(node));
}

const Thread* Cube::def_thrd(string name, int rank, const Process* proc) {
  return space->def_thrd(name, rank, const_cast<Process*>(proc));
}

void Cube::set_sev(const Metric* met, const Cnode*  cnode,  const Thread* thrd, 
		   double value) {
  if (user_data_type == FLAT_TREE_ONLY) {
    cerr << "Error: not building call trees!" << endl;
  }
  severity->set_sev(const_cast<Metric*>(met), const_cast<Cnode*>(cnode), 
			   const_cast<Thread*>(thrd), value);
}

void Cube::add_sev(const Metric* met, const Cnode*  cnode,  const Thread* thrd, 
		   double incr) {
  double val = 
    severity->get_sev(const_cast<Metric*>(met), const_cast<Cnode*>(cnode), 
		      const_cast<Thread*>(thrd));
  severity->set_sev(const_cast<Metric*>(met), const_cast<Cnode*>(cnode), 
		    const_cast<Thread*>(thrd), val+incr);
}

void Cube::set_sev(const Metric* met, const Region* region, const Thread* thrd, 
		   double value) {
  if (user_data_type == -1) user_data_type = FLAT_TREE_ONLY;
  if (user_data_type != FLAT_TREE_ONLY) {
    cerr << "Error: not building flat-tree!" << endl;
  }
  // after passing error checking
  Cnode* v_cnode = NULL;
  const vector<Cnode*>& cnodes = space->get_cnodev();
  // existent already?
  for (int i = 0; i < cnodes.size(); i++) {
    const Region* tmp = cnodes[i]->get_csite()->get_callee();
    if (tmp == region) {
      v_cnode = cnodes[i];
      break;
    }
  }
  if (v_cnode == NULL) {
    const Callsite* v_csite = def_csite(region->get_mod(), -1, region);
    v_cnode = const_cast<Cnode*>
      (space->def_cnode(const_cast<Callsite*>(v_csite), NULL));//everyone is a root
  }
  severity->set_sev(const_cast<Metric*>(met), const_cast<Cnode*>(v_cnode), 
		    const_cast<Thread*>(thrd), value);
}

void Cube::add_sev(const Metric* met, const Region* region,  const Thread* thrd, 
		   double incr) {
  int i;
  if (user_data_type != FLAT_TREE_ONLY) {
    cerr << "Error: not building flat-tree!" << endl;
  }
  const vector<Cnode*>& cnodes = space->get_cnodev();
  Cnode* cn;
  for (i = 0; i < cnodes.size(); i++) {
    const Region* tmp = cnodes[i]->get_csite()->get_callee();
    if (tmp == region) {
      cn = cnodes[i];
      break;
    }
  }
  if (i == cnodes.size()) throw CUBException("add_sev(): region undefined");
  double val = 
    severity->get_sev(const_cast<Metric*>(met), cn, const_cast<Thread*>(thrd));
  severity->set_sev(const_cast<Metric*>(met), cn, const_cast<Thread*>(thrd), val+incr);
}

vector<const Thread*>
get_ordered_thrdv(const map<const Location*, int>& loc_mapping,
		  const vector<Thread*>& thrdv) {
  map<const Location*, int>::const_iterator i;
  map<int, const Location*> reversed;
  map<int, const Location*>::iterator j;
  for (i = loc_mapping.begin(); i != loc_mapping.end(); i++) {
    reversed[i->second] = i->first;
  }
  vector<const Thread*> retv;
  for (j = reversed.begin(); j != reversed.end(); j++) {
    if (find(thrdv.begin(),thrdv.end(), j->second) != thrdv.end()) {
      Location* l = const_cast<Location*>(j->second);
      retv.push_back(static_cast<Thread*>(l));
    }
  }
  if (retv.size() != thrdv.size()) 
    cerr << "Error: write a row of severities in loc-id order!" << endl;
  
  return retv;
}

void Cube::write(string path) {
  ofstream out;
  out.open(path.c_str());
  if (!out.good()) {
    cerr << "Open file error: " << path << endl;
    exit(1);
  }
  out << *this;
  out.close();		    
} 

namespace cube {
  ostream& operator<< (ostream& out, const Cube& cube) {
    int i;
    int idcounter;
    int level = 2;
    map<const Metric*, int>   met_mapping;
    map<const Module*, int>   mod_mapping;
    map<const Region*, int>   regn_mapping;
    map<const Callsite*, int> csite_mapping;
    map<const Cnode*, int>    cnode_mapping;
    map<const Location*, int> loc_mapping;
    
    const Space* space = cube.get_space();
    const vector<Metric*>& pv = space->get_met_roots();
    const vector<Module*>& mv = space->get_modv();
    const vector<Region*>& rv = space->get_regnv();
    const vector<Callsite*>& csv = space->get_csitev();
    const vector<Cnode*>&  cnv = space->get_prog_roots();
    const vector<Machine*>& lv = space->get_sys_roots();
    const map<string, string>& attrs = space->get_attrs();
    const vector<string>& mirrors = const_cast<Space*>(space)->get_mirrors();

    // xml header
    out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl << endl;
    out << "<cube " << "version=\"2.0\">" << endl;

    // metadata info
    map<string, string>::const_iterator ai;
    for(ai = attrs.begin(); ai != attrs.end(); ai++) {
      out << "  <attr " << "key=" << "\"" << ai->first << "\" "
	  << "value=" << "\"" << ai->second << "\"/>" << endl;
    }
    // mirrored URLs
    out << "  <doc>" <<endl;
    out << "    <mirrors>" << endl;
    for (i = 0; i < mirrors.size(); i++) {
      out << "      <murl>" << mirrors[i] << "</murl>" << endl;
    }
    out << "    </mirrors>" << endl;
    out << "  </doc>" <<endl;

    // behavior 
    out << "  <behavior>"   << endl;
    idcounter = 0;
    for(i = 0; i < pv.size() ; i++) {
      pv[i]->set_level(level);
      pv[i]->writeXML(out, idcounter, met_mapping);
    }
    out << "  </behavior>"  << endl;
    // program
    out << "  <program>"    << endl;
    idcounter = 0;
    for(i = 0; i < mv.size(); i++) { // write modules
      mv[i]->set_level(level);
      mod_mapping[mv[i]] = idcounter;
      mv[i]->writeXML(out, idcounter);
    }
    idcounter = 0;
    for(i = 0; i < rv.size(); i++) {// write regions
      rv[i]->set_level(level);
      regn_mapping[rv[i]] = idcounter;
      rv[i]->writeXML(out, idcounter, mod_mapping[rv[i]->get_mod()]);
    }
    idcounter = 0;
    for(i = 0; i < csv.size(); i++) {// write csites
      csv[i]->set_level(level);
      csite_mapping[csv[i]] = idcounter;
      csv[i]->writeXML(out, idcounter, mod_mapping[csv[i]->get_mod()], 
		       regn_mapping[csv[i]->get_callee()]);
    }
    idcounter = 0;
    for(i = 0; i < cnv.size(); i++) {// write cnodes
      cnv[i]->set_level(level);
      cnv[i]->writeXML(out, idcounter, cnode_mapping, csite_mapping); 		 
    }
    out << "  </program>"   << endl;
    // system
    idcounter = 0;
    out << "  <locations>"  << endl;
    for(i = 0; i < lv.size(); i++) {
      lv[i]->set_level(level);
      lv[i]->writeXML(out, idcounter, loc_mapping);
    }
    out << "  </locations>" << endl;
    // severity matrices
    out << "  <severity>"   << endl;
    map<const Metric*, map<const Cnode*, map<const Thread*, double> > >::
      iterator ii;
    map<const Cnode*,  map<const Thread*, double> >::iterator j;
    // map<const Thread*, double>::const_iterator k;
    Severity* sevp = const_cast<Severity*>(cube.get_sev()); 
    map<const Metric*, map<const Cnode*, map<const Thread*, 
      double> > >& sev = sevp->get_sev(); //get_sev() can't be const
    vector<const Thread*> thrdv = 
      get_ordered_thrdv(loc_mapping, space->get_thrdv()); // reorder threadv.
    out.precision(12);
    // out.setf(ios::fixed);
    for (ii = sev.begin(); ii != sev.end(); ii++) {
      out << "    <matrix metricId=\"" << met_mapping[ii->first]; // 4 space indent
      out << "\">" << endl;
      map<const Cnode*, map<const Thread*, double> >& map2d=ii->second;
      for (j = map2d.begin(); j!= map2d.end(); j++) {
	out << "      <row cnodeId=\"" << cnode_mapping[j->first];//6 space indent
	out << "\">" <<endl;
	map<const Thread*, double>& map1d = j->second;
	//for (k = map1d.begin(); k!= map1d.end(); k++) {
	//  out << "        " << k->second << endl;   // 8 space indent
	//} // removed since this order is in mem address.
	for (int k = 0; k < thrdv.size(); k++) {
	  if (map1d[thrdv[k]] == 0) {
	    out << /*"        " <<*/ "0" << endl;
	  } else {
	    out << /*"        " <<*/ map1d[thrdv[k]] << endl;   // 8 space indent
	  }
	}
	out <<"      </row>" << endl;
      }
      out << "    </matrix>" << endl;
    }
    out << "  </severity>"  << endl;
    out << "</cube>"      << endl;
    
    return out;
  }
  
  istream& operator>> (std::istream& in, Cube& cb) {
    try {
      Parser parser(cb);
      int ret = parser.do_parse(in);
      // cout << "parser done" << endl;
      if (ret < 0) {
	if (ret == -1) {
	  throw CUBExceptionWarning("version " + parser.get_version());
	}
      }
    } catch (CUBException e) {
      cerr << e;
    }
    return in;      
  }

  /*******************************************************************
   *
   *                  ALGEBRA FUNCTIONS
   *
   *******************************************************************/

  Metric* get_myroot(const Metric* m) {
    Metric* tmp = const_cast<Metric*>(m);
    if (tmp == NULL) return NULL; 
    while(tmp->get_parent() != NULL) {
      tmp = const_cast<Metric*>(m->get_parent());
    }
    return tmp;
  }

  Cube* cube_merge(const Cube& lhs, const Cube& rhs) {
    double d;
    Space* newspace;
    SpaceMapping lhsm, rhsm;
    
    newspace = space_merge(lhs.get_space(), rhs.get_space(), lhsm, rhsm);
    Cube* newcube = new Cube(newspace, new Severity());
    Severity* lhssev = const_cast<Severity*>(lhs.get_sev());
    lhssev->align(&lhsm);
    Severity* rhssev = const_cast<Severity*>(rhs.get_sev());
    rhssev->align(&rhsm);

    const vector<Metric*>&  metv = newspace->get_metv();
    const vector<Cnode*>&  cnodev= newspace->get_cnodev();
    const vector<Thread*>& thrdv = newspace->get_thrdv();
    // cout << "total # of metrics: " << metv.size() << endl;
    // cout << "total # of cnodes: " << cnodev.size() << endl;
    // cout << "total # of threads: " << thrdv.size() << endl;
    
    for (int i = 0; i < metv.size(); i++) 
      for (int j = 0; j < cnodev.size(); j++) 
	for (int k = 0; k < thrdv.size(); k++) {
	  if (lhsm.metm[metv[i]] != NULL) {
	    d = lhssev->get_sev(metv[i], cnodev[j], thrdv[k]);
	  } else {
	    Metric* root = get_myroot(metv[i]);
	    if (lhsm.metm[root] == NULL) { // RHS will be included to LHS only tree by tree.
	      d = rhssev->get_sev(metv[i], cnodev[j], thrdv[k]);
	    }
	  }
	  newcube->set_sev(metv[i], cnodev[j], thrdv[k], d);	
	}
    
    return newcube;
  }
  
  Cube* cube_diff (const Cube& lhs, const Cube& rhs) {
    double d;
    Space* newspace;
    SpaceMapping lhsm, rhsm;

    newspace = space_merge(lhs.get_space(), rhs.get_space(), lhsm, rhsm);
    Cube* newcube = new Cube(newspace, new Severity());
    Severity* lhsev = const_cast<Severity*>(lhs.get_sev());
    lhsev->align(&lhsm);
    Severity* rhsev = const_cast<Severity*>(rhs.get_sev());
    rhsev->align(&rhsm);

    const vector<Metric*>& metv = newspace->get_metv();
    const vector<Cnode*>& cnodev = newspace->get_cnodev();
    const vector<Thread*>& thrdv = newspace->get_thrdv();
    // cout << "total # of metrics: " << metv.size() << endl;
    // cout << "total # of cnodes: " << cnodev.size() << endl;
    // cout << "total # of threads: " << thrdv.size() << endl;  
    
    for (int i = 0; i < metv.size(); i++) 
      for (int j = 0; j < cnodev.size(); j++) 
	for (int k = 0; k < thrdv.size(); k++) {
	  d = lhsev->get_sev(metv[i], cnodev[j], thrdv[k]) -
	    rhsev->get_sev(metv[i], cnodev[j], thrdv[k]);
	  newcube->set_sev(metv[i], cnodev[j], thrdv[k], d);	
	}
    
    return newcube;
  }
  
  Cube* cube_mean (const vector<Cube*>& cubev) {
    double d;
    Space* newspace;
    SpaceMapping dummy1, dummy2;

    // union of a collection of spaces
    newspace = space_merge(cubev[0]->get_space(), cubev[1]->get_space(), dummy1, dummy2);
    for (int i = 2; i < cubev.size(); i++) {
      Space* space = space_merge(newspace, cubev[i]->get_space(), dummy1, dummy2);
      delete newspace;
      newspace = space;
    }

    Cube* sumcube = new Cube(newspace, new Severity());
    Cube* newcube;
    Severity *lhsev, *rhsev;
    for (int i = 0; i < cubev.size(); i++) {
      SpaceMapping tmp1, tmp2;
      Space* space = space_merge(sumcube->get_space(), cubev[i]->get_space(), tmp1, tmp2);
      lhsev = const_cast<Severity*>(sumcube->get_sev());
      rhsev = const_cast<Severity*>(cubev[i]->get_sev());
      lhsev->align(&tmp1);
      rhsev->align(&tmp2);
      const vector<Metric*>& metv = space->get_metv();
      const vector<Cnode*>& cnodev = space->get_cnodev();
      const vector<Thread*>& thrdv = space->get_thrdv();
      newcube = new Cube(space, new Severity());
      for (int t = 0; t < metv.size(); t++) 
	for (int j = 0; j < cnodev.size(); j++) 
	  for (int k = 0; k < thrdv.size(); k++) {
	    d = lhsev->get_sev(metv[t], cnodev[j], thrdv[k]) +
	      rhsev->get_sev(metv[t], cnodev[j], thrdv[k]);
	    newcube->set_sev(metv[t], cnodev[j], thrdv[k], d);
	    if (i == cubev.size()-1) {
	      newcube->set_sev(metv[t], cnodev[j], thrdv[k], d / cubev.size());
	    }
	  }
      delete sumcube;
      sumcube = newcube;
    }

    return sumcube;
  }


} // end of namespace cube

  



