/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


#ifndef CUBE_CUBE_H
#define CUBE_CUBE_H   

/*
 *----------------------------------------------------------------------------
 *
 * class Cube
 *
 *----------------------------------------------------------------------------
 */

#include <iostream>
#include <string>
#include <vector>


namespace cube {
  class Callsite;
  class Cnode;
  class Machine;
  class Metric;
  class Module;
  class Node;
  class Process;
  class Region;
  class Severity;
  class Space;
  class Thread;
  
  class Cube {
  public:
    
    Cube();
    ~Cube();

    Cube(const Cube& cube);
    Cube(Space* space, Severity* sev);
    
    const Space*    get_space() const        { return space; }  
    const Severity* get_sev()   const        { return severity; }
    

    /**************************************************************
     * 
     * CUBE write API
     *
     * This part is intended to be used by tools writing CUBE data.
     * 
     *************************************************************/ 
    
    // The order of children in any hierarchical structure is defined by
    // the order of child creation. For example, the first performance metric
    // created as child of a given metric is the first child.
    
    // set cube attributes
    void                        def_attr(std::string key, std::string value);
    std::string                 get_attr(std::string key);
    
    // set mirror URLs
    void                             def_mirror(std::string url);
    std::vector<std::string>&        get_mirrors();

    // build metric dimension.

    // url is a web link or a file path, or a string of "@mirror@...". 
    // descr is simply a short string.
    const Metric*   def_met(std::string name, std::string uom, std::string url, 
			    std::string descr, const Metric* parent);
    
    // build program dimension
    const Module*   def_module(std::string name);
    const Region*   def_region(std::string name, long begln, long endln, 
			       std::string url, std::string descr, const Module* mod);
    const Callsite* def_csite(const Module* mod, int line, const Region* callee);
    const Cnode*    def_cnode(const Callsite* csite, const Cnode* parent);
   
    // for call-tree w/o line num.
    const Cnode*    def_cnode(const Region* region, const Cnode* parent);

    
    // build system dimension
    const Machine*  def_mach(std::string name);
    const Node*     def_node(std::string name, const Machine* mach);
    const Process*  def_proc(std::string name, int rank, const Node* node);
    const Thread*   def_thrd(std::string name, int rank, const Process* proc);

    // set severity data
    void set_sev(const Metric* met, const Cnode*  cnode,  const Thread* thrd,
		 double value); 
    void add_sev(const Metric* met, const Cnode*  cnode,  const Thread* thrd,
		 double incr); 
    // flat-tree only
    void set_sev(const Metric* met, const Region* region, const Thread* thrd,
		 double value);
    void add_sev(const Metric* met, const Region* region, const Thread* thrd,
		 double incr);
    // write to a file
    void write(std::string path);

  private:

    Space* space;                             // metadata
    Severity* severity;                       // data
    int user_data_type;                       // user input data type
  };

  // non-member functions

  // I/O operators
  std::istream& operator>> (std::istream& in, Cube& cb);
  std::ostream& operator<< (std::ostream& out, const Cube& cb);
  
  // Algebra operators
  Cube* cube_merge(const Cube& lhs, const Cube& rhs);
  Cube* cube_mean (const std::vector<Cube*>& cubev);
  Cube* cube_diff (const Cube& lhs, const Cube& rhs);
  
}

#endif






