/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/************************************************
             CubeCache.cpp


 ************************************************/

#include <iostream>
#include <vector>
#include <math.h>

#include "CubeCollect.h"
#include "CubeCached.h"
#include "Cube.h"
#include "Space.h"
#include "Metric.h"
#include "Cnode.h"
#include "Location.h"
#include "Module.h"
#include "Region.h"
#include "Machine.h"

using namespace std;
using namespace cube;

CubeCached::CubeCached(CubeCollect* cub_col) {
  cube_collect = cub_col;
  cnode_met_tag=flat_met_tag = NULL;
  loc4cnode_met_tag=loc4mod_met_tag=loc4reg_met_tag = NULL;
  loc_cnode_tag = NULL;
  loc_mod_tag = NULL;
  loc_reg_tag = NULL;
  cnode_metmode_tag=flat_metmode_tag = -1;
  loc4cnode_metmode_tag=loc_cnodemode_tag = -1;
  loc4mod_metmode_tag=loc_modmode_tag = -1;
  loc4reg_metmode_tag=loc_regmode_tag = -1;
  // compute for met cache only once
  const Cube* cube = cube_collect->get_cube();
  const Space* space = cube->get_space();
  const vector<Metric*>& met_roots = space->get_met_roots();
  for(int i = 0; i < met_roots.size(); i++) {
    update_met_cache(met_roots[i]); 
  }
  fill_met_max();
}

CubeCached::~CubeCached() {
  delete cube_collect;
}

void CubeCached::update_met_cache(const Metric* root) {
  Metric* m = const_cast<Metric*>(root);
  int childnum = root->num_children();
  int mode_in[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  int mode_ex[3] = {CUBE_EXCL, CUBE_INCL, CUBE_INCL};

  /*
  met_buf_in[m] = cube_collect->get_vcsev(mode_in, root);
  met_buf_ex[m] = cube_collect->get_vcsev(mode_ex, root);
  if (childnum > 0) {
    for (int i = 0; i < childnum; i++) {
      const Metric* child = root->get_child(i);
      update_met_cache(child);
    }
  } */

  if (childnum == 0) {
    met_buf_in[m] = cube_collect->get_vcsev(mode_in, root);
    met_buf_ex[m] = cube_collect->get_vcsev(mode_ex, root);
    return; // if leaf node, return
  } else {
    for (int i = 0; i < childnum; i++) {
      const Metric* child = root->get_child(i);
      update_met_cache(child);
    }
    met_buf_ex[m] = cube_collect->get_vcsev(mode_ex, root);
    met_buf_in[m] = cube_collect->get_vcsev(mode_in, root);
    return;
  }
}

double CubeCached::
update_cnode_cache(const Metric* met, int met_mode, const Cnode* root) {
  double val_ex, val_in;
  Cnode* cn = const_cast<Cnode*>(root);
  int mode[3];
  mode[CUBE_MET]  = met_mode;
  mode[CUBE_PROG] = CUBE_EXCL;
  mode[CUBE_LOC]  = CUBE_INCL;

  val_ex = cube_collect->get_vcsev(mode, met, root); // exclusive cnode
  cnode_metmode_tag = met_mode;
  cnode_met_tag = const_cast<Metric*>(met);
  
  int childnum = root->num_children();
  if (childnum == 0) {
    cnode_buf_ex[cn] = val_ex;
    cnode_buf_in[cn] = val_ex;
    return val_ex;
  } else {
    val_in = val_ex;
    for (int i = 0; i < childnum; i++) {
      const Cnode* child = root->get_child(i);
      val_in += update_cnode_cache(met, met_mode, child);
    }
    cnode_buf_ex[cn] = val_ex;
    cnode_buf_in[cn] = val_in;
    return val_in;
  }
}

double CubeCached::
update_loc_cache (const Metric* met, int met_mode, 
		  const Cnode* cnode, int cnode_mode,
		  const Location* root) {
  double val_ex, val_in;
  // refresh the flags for the location cache
  loc4cnode_met_tag = const_cast<Metric*>(met);
  loc4cnode_metmode_tag = met_mode;
  loc_cnode_tag = const_cast<Cnode*>(cnode);
  loc_cnodemode_tag = cnode_mode;

  int childnum = root->num_children();
  Location* loc = const_cast<Location*>(root);
  int mode[3];
  mode[CUBE_MET] = met_mode;
  mode[CUBE_PROG] = cnode_mode;
  mode[CUBE_LOC]  = CUBE_EXCL;
  val_ex = cube_collect->get_vcsev(mode, met, cnode, root);
  if (childnum == 0) {
    loc_buf_cnode_ex[loc] = val_ex;
    loc_buf_cnode_in[loc] = val_ex;
    return val_ex;
  } else {
    val_in = val_ex;
    for (int i = 0; i < childnum; i++) {
      Vertex* tmp = const_cast<Vertex*>(root->get_child(i));
      const Location* child = static_cast<Location*>(tmp);
      val_in += update_loc_cache(met, met_mode, cnode, cnode_mode, child);
    }
    loc_buf_cnode_ex[loc] = val_ex;
    loc_buf_cnode_in[loc] = val_in;

    return val_in;
  }

}

void CubeCached::update_loc_cache(const Metric* met, int met_mode,
				  const Module* mod, int mod_mode,
				  const Location* root) {
  Location* loc = const_cast<Location*>(root);
  int childnum = root->num_children();
  double sum = 0;
  // refresh the flags for the location cache
  loc4mod_met_tag = const_cast<Metric*>(met);
  loc4mod_metmode_tag = met_mode;
  loc_mod_tag = const_cast<Module*>(mod);
  loc_modmode_tag = mod_mode;
  int mode[3];
  mode[CUBE_MET] = met_mode;
  mode[CUBE_PROG] = mod_mode;
  mode[CUBE_LOC]  = CUBE_EXCL;
  sum = cube_collect->get_vmsev(mode, met, mod, root);
  // leaf tree node
  if (childnum == 0) {
    loc_buf_mod_ex[loc] = sum;
    loc_buf_mod_in[loc] = sum;
    return;
  } 
  loc_buf_mod_ex[loc] = sum;
  mode[CUBE_LOC] = CUBE_INCL;
  loc_buf_mod_in[loc] = 
    cube_collect->get_vmsev(mode, met, mod, root);
  for (int i = 0; i < childnum; i++) {
    Vertex* tmp = const_cast<Vertex*>(root->get_child(i));
    const Location* child = static_cast<Location*>(tmp);
    update_loc_cache(met, met_mode, mod, mod_mode, child);
  }  
}

void CubeCached::update_loc_cache(const Metric* met, int met_mode,
				  const Region* reg, int reg_mode,
				  const Location* root) {
  Location* loc = const_cast<Location*>(root);
  int childnum = root->num_children();
  double sum = 0;
  // refresh the flags for the location cache
  loc4reg_met_tag = const_cast<Metric*>(met);
  loc4reg_metmode_tag = met_mode;
  loc_reg_tag = const_cast<Region*>(reg);
  loc_regmode_tag = reg_mode;
  int mode[3];
  mode[CUBE_MET] = met_mode;
  mode[CUBE_PROG] = reg_mode;
  mode[CUBE_LOC]  = CUBE_EXCL;
  sum = cube_collect->get_vrsev(mode, met, reg, root);
  if (childnum == 0) {
    // leaf tree node
    loc_buf_reg_ex[loc] = sum;
    loc_buf_reg_in[loc] = sum;
    return;
  }
  // internal tree node
  loc_buf_reg_ex[loc] = sum;
  mode[CUBE_LOC] = CUBE_INCL;
  loc_buf_reg_in[loc] = cube_collect->get_vrsev(mode, met, reg, root);
  for (int i = 0; i < childnum; i++) {
    Vertex* tmp = const_cast<Vertex*>(root->get_child(i));
    const Location* child = static_cast<Location*>(tmp);
    update_loc_cache(met, met_mode, reg, reg_mode, child);
  }
}

void CubeCached::update_mod_cache(const Metric* met, int met_mode) {
  const Cube* cube = cube_collect->get_cube();
  const Space* space = cube->get_space();
  const vector<Module*>& modv = space->get_modv();
  int mode[3];
  mode[CUBE_MET] = met_mode;
  for (int i = 0; i < modv.size(); i++) {
    mode[CUBE_PROG] = CUBE_INCL;
    mod_buf_in[modv[i]] = 
      cube_collect->get_vmsev(mode, met, modv[i]);
    mode[CUBE_PROG] = CUBE_EXCL;
    mod_buf_ex[modv[i]] = 
      cube_collect->get_vmsev(mode, met, modv[i]);
  }
  flat_met_tag = const_cast<Metric*>(met);
  flat_metmode_tag = met_mode;
}

void CubeCached::update_reg_cache(const Metric* met, int met_mode) {
  const Cube* cube = cube_collect->get_cube();
  const Space* space = cube->get_space();
  const vector<Region*>& regv = space->get_regnv();
  int mode[3];
  mode[CUBE_MET] = met_mode;
  for (int i = 0; i < regv.size(); i++) {
    Region* r = regv[i];
    mode[CUBE_PROG] = CUBE_INCL;
    reg_buf_in[r] = cube_collect->get_vrsev(mode, met, r);
    mode[CUBE_PROG] = CUBE_EXCL;
    reg_buf_ex[r] = cube_collect->get_vrsev(mode, met, r);
    reg_buf_sub[r] = reg_buf_in[r] - reg_buf_ex[r]; // update subregn meanwhile
  }
  flat_met_tag = const_cast<Metric*>(met);
  flat_metmode_tag = met_mode;
}

double CubeCached::
get_vcsev(int mode[3], const Metric* met,const Cnode* cnode, 
	  const Location* loc) {
  double sum = 0;
  Metric* m = const_cast<Metric*>(met);
  Cnode*  c = const_cast<Cnode*>(cnode);
  Location* l = const_cast<Location*>(loc);
  
  if (cnode == NULL && loc == NULL) {
    if (mode[CUBE_MET] == CUBE_INCL) sum = met_buf_in[m];
    else sum = met_buf_ex[m];
  } else if (loc == NULL) {
    validate_cnode_cache(met, mode[CUBE_MET]);
    if (mode[CUBE_PROG] == CUBE_INCL) sum = cnode_buf_in[c];
    else sum = cnode_buf_ex[c];
  }else {
    validate_loc_cache(met, mode[CUBE_MET], cnode, mode[CUBE_PROG]); 
    if (mode[CUBE_LOC] == CUBE_INCL) sum = loc_buf_cnode_in[l];
    else sum = loc_buf_cnode_ex[l];
  }
  return sum;  
}

double CubeCached::
get_vmsev(int mode[3], const Metric* met, const Module* mod, const Location* loc){
  double sum = 0;
  Module*  module = const_cast<Module*>(mod);
  Location* l = const_cast<Location*>(loc);
  if (loc == NULL) {
    validate_flat_cache(met, mode[CUBE_MET]);    
    if (mode[CUBE_PROG] == CUBE_INCL) sum = mod_buf_in[module];
    else sum = mod_buf_ex[module];
  } else {
    validate_loc_cache(met, mode[CUBE_MET], mod, mode[CUBE_PROG]);
    if (mode[CUBE_LOC] == CUBE_INCL) sum = loc_buf_mod_in[l];
    else sum = loc_buf_mod_ex[l];
  }
  return sum;  
}

double CubeCached::
get_vrsev(int mode[3], const Metric* met, const Region* reg, const Location* loc) {
  double sum = 0;
  Region*  r = const_cast<Region*>(reg);
  Location* l = const_cast<Location*>(loc);
  if (loc == NULL) {
    validate_flat_cache(met, mode[CUBE_MET]);    
    if (mode[CUBE_PROG] == CUBE_INCL) sum = reg_buf_in[r];
    else sum = reg_buf_ex[r];
  } else {
    validate_loc_cache(met, mode[CUBE_MET], reg, mode[CUBE_PROG]);
    if (mode[CUBE_LOC] == CUBE_INCL) sum = loc_buf_reg_in[l];
    else sum = loc_buf_reg_ex[l];
  }
  return sum;  
}

void CubeCached::validate_cnode_cache(const Metric* met, int met_mode) {
  if (met == cnode_met_tag && met_mode == cnode_metmode_tag) {
    return;
  }
  const Cube* cube = cube_collect->get_cube();
  const Space* space = cube->get_space();
  const vector<Cnode*>& roots = space->get_prog_roots();
  for(int i = 0; i < roots.size(); i++) {
    update_cnode_cache(met, met_mode, roots[i]); 
  }
}

void CubeCached::validate_flat_cache(const Metric* met, int met_mode) {
  if (met == flat_met_tag && met_mode == flat_metmode_tag) {
    return;
  }
  update_mod_cache(met, met_mode); 
  update_reg_cache(met, met_mode);
}

void CubeCached::
validate_loc_cache(const Metric* met, int met_mode,
		   const Cnode* cnode, int cnode_mode) {
  if (met == loc4cnode_met_tag && met_mode == loc4cnode_metmode_tag &&
      cnode == loc_cnode_tag && cnode_mode == loc_cnodemode_tag) {
    return;
  }
  const Cube* cube = cube_collect->get_cube();
  const Space* space = cube->get_space();
  const vector<Machine*>& roots = space->get_sys_roots();
  for(int i = 0; i < roots.size(); i++) {
      update_loc_cache(met, met_mode, cnode, cnode_mode, roots[i]); 
  }
}

void CubeCached::
validate_loc_cache(const Metric* met, int met_mode,
		   const Module* mod, int mod_mode) {
  if (met == loc4mod_met_tag && met_mode == loc4mod_metmode_tag &&
      mod == loc_mod_tag && mod_mode == loc_modmode_tag) {
    return;
  }
  const Cube* cube = cube_collect->get_cube();
  const Space* space = cube->get_space();
  const vector<Machine*>& roots = space->get_sys_roots();
  for(int i = 0; i < roots.size(); i++) {
    update_loc_cache(met, met_mode, mod, mod_mode, roots[i]); 
  }
}

void CubeCached::
validate_loc_cache(const Metric* met, int met_mode,
		   const Region* reg, int reg_mode) {
  if (met == loc4reg_met_tag && met_mode == loc4reg_metmode_tag &&
      reg == loc_reg_tag && reg_mode == loc_regmode_tag) {
    return;
  }
  const Cube* cube = cube_collect->get_cube();
  const Space* space = cube->get_space();
  const vector<Machine*>& roots = space->get_sys_roots();
  for(int i = 0; i < roots.size(); i++) {
    update_loc_cache(met, met_mode, reg, reg_mode, roots[i]); 
  }
}

void CubeCached::fill_met_max() {
  const Cube* cube = cube_collect->get_cube();
  const Space* space = cube->get_space();
  const vector<Metric*>& roots = space->get_met_roots();

  for (int i = 0; i < roots.size(); i++) {
    max_sevs[roots[i]] = find_metree_max(roots[i]);
  }
}

double CubeCached::find_metree_max(const Metric* root) {
  double max = 0;  // inclusive state
  double max2 = 0; // exclusive state
  double val;
  int mode[3]  = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  int mode2[3] = {CUBE_EXCL, CUBE_INCL, CUBE_INCL};

  max  = get_vcsev(mode,  root);
  if (root->num_children() == 0) {
    return max;
  }

  max2 = get_vcsev(mode2, root);
  
  if (fabs(max) < fabs(max2)) max = fabs(max2);
  // left-right-root (depth-first)
  for (int i = 0; i < root->num_children(); i++) {
    val = find_metree_max(root->get_child(i));
    if (fabs(val) > fabs(max)) max = val;
  }
  return fabs(max);
}

double CubeCached::get_max(const Metric* met) {
  if (met == NULL) return 0;

  Metric* m = const_cast<Metric*>(met);
  while(m->get_parent() != NULL) {
    m = const_cast<Metric*>(m->get_parent());
  }
  return max_sevs[m];
}
