/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#ifndef CUBE_CACHED_H
#define CUBE_CACHED_H   

/*
 *----------------------------------------------------------------------------
 *
 * class CubeCached (with cache support)
 *
 *----------------------------------------------------------------------------
 */

#include <map>
#include "Matrix.h"

namespace cube {
  class CubeCollect;
  class Metric;
  class Cnode;
  class Region;
  class Location;
  class Module;
  

  class CubeCached {
    
  public:
    
    CubeCached(CubeCollect* cub_col); 
    CubeCollect* get_cube() { return cube_collect; }
    ~CubeCached();
    
    // call-tree node severity using cache
    double get_vcsev (int mode[3], const Metric* met, const Cnode* cnode = NULL,
		      const Location* loc = NULL);
    
    // region severity using cache
    double get_vrsev (int mode[3], const Metric* met, const Region* reg, 
		      const Location* loc = NULL);

    // module severity using cache
    double get_vmsev (int mode[3], const Metric* met, const Module*  mod, 
		      const Location* loc = NULL);
    
    // get metric max using cache
    double get_max(const Metric* met);

  private:
    
    CubeCollect* cube_collect;
    
    // do nothing if $ still valid
    void validate_cnode_cache(const Metric* met, int met_mode);
    void validate_flat_cache (const Metric* met, int met_mode);
    void validate_loc_cache  (const Metric* met, int met_mode,
			      const Cnode* cnode, int cnode_mode);
    void validate_loc_cache  (const Metric* met, int met_mode,
			      const Module* mod, int mod_mode);
    void validate_loc_cache  (const Metric* met, int met_mode,
			      const Region* reg, int reg_mode);
    
    // update caches
    double update_cnode_cache (const Metric* met, int met_mode, const Cnode* root);
    double update_loc_cache   (const Metric* met,  int met_mode,
			       const Cnode* cnode, int cnode_mode,
			       const Location* root);
    void update_met_cache     (const Metric* root); // called only once
    void update_loc_cache     (const Metric* met, int met_mode,
			       const Module* mod, int mod_mode,
			       const Location* root);
    void update_loc_cache     (const Metric* met, int met_mode,
			       const Region* reg, int reg_mode,
			       const Location* root);    
    void update_mod_cache     (const Metric* met, int met_mode);
    void update_reg_cache     (const Metric* met, int met_mode);
    
    void   fill_met_max(); // each metric tree has a max val.
    double find_metree_max(const Metric* root);

    // metric cache (no tags)
    std::map<Metric*, double>   met_buf_in;
    std::map<Metric*, double>   met_buf_ex;

    // call path cache
    std::map<Cnode*, double>    cnode_buf_in;
    std::map<Cnode*, double>    cnode_buf_ex;
    Metric* cnode_met_tag; // 2 tags associated with this $
    int cnode_metmode_tag;

    // module cache
    std::map<Module*, double>   mod_buf_in; 
    std::map<Module*, double>   mod_buf_ex; 
    // region cache
    std::map<Region*, double>   reg_buf_in; 
    std::map<Region*, double>   reg_buf_ex;
    // subregion cache
    std::map<Region*, double>   reg_buf_sub;
    Metric* flat_met_tag; // 2 tags associated with this $
    int flat_metmode_tag;
    
    // 3 location caches for cnode, module, and regn accordingly:

    std::map<Location*, double> loc_buf_cnode_in;
    std::map<Location*, double> loc_buf_cnode_ex;
    Metric* loc4cnode_met_tag; // 4 tags
    int loc4cnode_metmode_tag;
    Cnode* loc_cnode_tag;
    int loc_cnodemode_tag;
    
    std::map<Location*, double> loc_buf_mod_in;
    std::map<Location*, double> loc_buf_mod_ex;
    Metric* loc4mod_met_tag;
    int loc4mod_metmode_tag;
    Module* loc_mod_tag;
    int loc_modmode_tag;
    
    std::map<Location*, double> loc_buf_reg_in;
    std::map<Location*, double> loc_buf_reg_ex;
    Metric* loc4reg_met_tag;
    int loc4reg_metmode_tag;
    Region* loc_reg_tag;
    int loc_regmode_tag;

    // max value of each metric tree (may not be root)
    std::map<Metric*, double>  max_sevs;    
  };
  
}

#endif

/*****************************************************************************
How many caches($) used:

met$------->cnode$--------->loc_cnode$(state: left met, cnode)
    \
     \
      \
       \--->module$-------->loc_module$(state: left met, module)
        \-->region$-------->loc_region$(state: left met, region)



*****************************************************************************/
