/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/************************************************
             CubeCollect.cpp
  
  
 ************************************************/

#include <vector>
#include <map>
#include <algorithm>
#include "math.h"

#include "Cube.h"
#include "CubeCollect.h"
#include "Matrix.h"
#include "Space.h"
#include "Severity.h"
#include "Location.h"
#include "Metric.h"
#include "Callsite.h"
#include "Region.h"
#include "Module.h"
#include "Machine.h"
#include "Node.h"
#include "Process.h"
#include "Thread.h"
#include "Cnode.h"

using namespace std;
using namespace cube;


CubeCollect::CubeCollect(Cube* cube) {
  this->cube = cube;
  const vector<Metric*>& metv = cube->get_space()->get_met_roots();
  for(int i = 0; i < metv.size(); i++) {
    calc_excl_matrix(metv[i]);
  }
}

CubeCollect::~CubeCollect() {
  delete this->cube;
}

Matrix CubeCollect::calc_excl_matrix(const Metric* root) {
  Severity* severity = const_cast<Severity*>(cube->get_sev());
  map<const Metric*, Mat2D>& sev = severity->get_sev();
  Matrix incl_mat = static_cast<Matrix>(sev[root]);
  // leaf node
  if(root->num_children() == 0) {
    excl_severity[root] = incl_mat;
    return incl_mat;
  }
  // internal node
  Matrix tmp = incl_mat; // backup incl_mat to return
  for (int i = 0; i < root->num_children(); i++) {
    const Metric* c = root->get_child(i);
    Matrix child = calc_excl_matrix(c);
    tmp = tmp - child;
  }
  excl_severity[root] = tmp;

  return incl_mat;
}

double CubeCollect::get_vcsev(int mode[3], const Metric* met, 
			      const Cnode* cnode, const Location* loc) {
  double sum = 0;

  if (cnode == NULL && loc == NULL) { // met
    const vector<Cnode*>& cnodev = cube->get_space()->get_prog_roots();
    const vector<Machine*>& machv = cube->get_space()->get_sys_roots();
    if (cube->get_attr("CUBE_CT_AGGR") == "SUM"){//calc the sum of all cnode-roots
      for(int i = 0; i < cnodev.size(); i++) {
	for (int j = 0; j < machv.size(); j++) {
	  sum += sev(mode[CUBE_MET], mode[CUBE_PROG], mode[CUBE_LOC], 
		     met, cnodev[i], machv[j]);
	}
      }
    } else if (cube->get_attr("CUBE_CT_AGGR") == "MAX") {
      for(int i = 0; i < cnodev.size(); i++) {
	double candidate = 0;
	for (int j = 0; j < machv.size(); j++) {
	  candidate += sev(mode[CUBE_MET], mode[CUBE_PROG], mode[CUBE_LOC], 
		     met, cnodev[i], machv[j]);
	}
	if (fabs(sum) < fabs(candidate)) sum = candidate; // try to keep the max value.
      }
    } else {
      cerr << "Error: call-tree aggregation mode undefined!" << endl;
    }
  } else if (loc == NULL) {// cnode
    const vector<Machine*>& machv = cube->get_space()->get_sys_roots();
    for (int j = 0; j < machv.size(); j++) {
      sum += sev(mode[CUBE_MET], mode[CUBE_PROG], mode[CUBE_LOC], 
		 met, cnode, machv[j]);
    } 
  } else { // loc
    sum = sev(mode[CUBE_MET], mode[CUBE_PROG], mode[CUBE_LOC], 
	      met, cnode, loc);
  }
  
  return sum;
}

const Thread* CubeCollect::loc2thrd(const Location* loc) {
  int i;
  const vector<Thread*>& thrdv = cube->get_space()->get_thrdv();
  // see whether the location is a thread
  for (i = 0; i < thrdv.size(); i++) {
    if (thrdv[i] == loc) break;
  }
  if (i == thrdv.size()) return NULL;
  return thrdv[i];
}

double CubeCollect::get_esev(const Metric* met, const Cnode* cnode, 
			     const Location* loc) {
  const Thread* t = loc2thrd(loc);
  if (t == NULL) return 0;
  Severity* sev = const_cast<Severity*>(cube->get_sev());
  return sev->get_sev(met, cnode, t);
}

double CubeCollect::get_esev_ex(const Metric* met, const Cnode* cnode, 
				const Location* loc) {
  double val;
  const Thread* t = loc2thrd(loc);
  if (t == NULL) return 0;
  val = excl_severity[const_cast<Metric*>(met)]
    [const_cast<Cnode*>(cnode)][const_cast<Thread*>(t)];
  return val;
}

/* x, y, z are inclusive/exclusive state */
double CubeCollect::sev(int x, int y, int z, const Metric* b, 
			const Cnode* n, const Location* l) {
  if (x == CUBE_EXCL) excl_met_flag = true;
  else excl_met_flag = false;

  return sev(y, z, b, n, l);
}

double CubeCollect::sev(int y, int z, const Metric* b, 
			const Cnode* n, const Location* l) {
  double sum = 0.0;
  sum = sev(z, b, n, l);

  if (n->num_children() == 0 || y == CUBE_EXCL) {
    return sum;
  }
  // if inclusive mode
  for (int i = 0; i < n->num_children(); i++) {
    const Cnode* c = n->get_child(i);
    sum += sev(CUBE_INCL, z, b, c, l);
  }
  return sum;
}

double CubeCollect::sev(int z, const Metric* b, const Cnode* n, 
			const Location* l ) {
  double sum = 0.0;

  if (!excl_met_flag)
    sum = get_esev(b, n, l);
  else // if focused metric is in Exclusive state, use exclusive matrix.
    sum = get_esev_ex(b, n, l);

  if (l->num_children() == 0 || z == CUBE_EXCL) {
    return sum;
  }

  // if inclusive mode
  for (int i = 0; i < l->num_children(); i++) {
    Vertex* tmp = const_cast<Vertex*>(l->get_child(i));
    const Location* c = static_cast<Location*>(tmp);
    // add its children up
    sum += sev(CUBE_INCL, b, n, c);
  }
  return sum;
}

// how many cnodes are pointing to this <regn> through csites
vector<const Cnode*> CubeCollect::get_callees(const Region* regn) {
  const vector<Cnode*>& cnodev = cube->get_space()->get_cnodev();
  vector<const Cnode*> v;

  for (int i = 0; i < cnodev.size(); i++) {
    Cnode* cnode = cnodev[i];
    const Callsite* csite = cnode->get_csite();
    const Region* callee = csite->get_callee();
    if (callee == regn) {
      v.push_back(cnode);
    }
  }
  
  return v;
}

// locally used by get_vrsev() only
bool is_simple(vector<const Cnode*>& cnodes) {
  if (cnodes.size() == 1) return true;  
  vector<const Cnode*>::iterator i; 
  for(i = cnodes.begin(); i != cnodes.end(); i++) {
    if ((*i)->get_rcpath().size() == 1) {
      const Cnode* backup = *i;
      cnodes.clear();
      cnodes.push_back(backup);
      return true;
    }
  }
  return false;
}
vector<string> largest_suffix(const vector<string>& a, const vector<string>& b) {
  if (a.size() > b.size()) {
    cerr << "Error: sorting STL didn't work" << endl;
    exit(1);
  }
  vector<string> lsuf;
  for (int i = 0; i < a.size(); i++) {
    if (a[i] == b[i]) {
      lsuf.push_back(a[i]);
    }
    else break;
  }
  return lsuf;
}
vector<string> largest_prefix(const vector<string>& a, const vector<string>& b) {
  if (a.size() > b.size()) {
    cerr << "Error: sorting STL didn't work" << endl;
    exit(1);
  }
  vector<string> prefix, prefix2;
  int i, j;
  i = a.size()-1;
  j = b.size()-1;
  for (; i >= 0; i--, j--) {
    if (a[i] == b[j]) {
      prefix.push_back(a[i]);
    }
    else break;
  }
  if (!prefix.empty()) {
    // reverse the order:
    int k = prefix.size()-1;
    for (; k >= 0; k--) {
      prefix2.push_back(prefix[k]);
    }
    return prefix2;
  }
  return prefix;
}

void print_cpath(const vector<string>& a) {
  for (int i = 0; i < a.size(); i++) {
    cout << a[i] <<"<==";
  }
  cout << endl;
}
void rm_suffix_prefix(vector<const Cnode*>& cnodes) {
  sort(cnodes.begin(), cnodes.end(), Less());
  vector<const Cnode*>::iterator i, j;
  vector<string> lsuf; // larget possible suffix
  vector<string>  lpre; // larget possible prefix

  for (i = cnodes.begin(); i != cnodes.end(); i++) {
    j = i; j++;
    while (j != cnodes.end()) {
      lsuf = largest_suffix((*i)->get_rcpath(),(*j)->get_rcpath());
      lpre = largest_prefix((*i)->get_rcpath(),(*j)->get_rcpath());
      /*
      if((*i)->get_rcpath()[0] == "ESMF_GRIDCOMPMOD::ESMF_GRIDCOMPRUN") {
	cout << "*********** COMPARE ***********" << endl;
	cout << "1:    ";
	print_cpath((*i)->get_rcpath());
	cout << "2:    ";
	print_cpath((*j)->get_rcpath());
	cout <<"suffix:   ";
	print_cpath(lsuf);
	cout <<"prefix:   ";
	print_cpath(lpre);
      }
      */
      if( lsuf == (*i)->get_rcpath() || lpre == (*i)->get_rcpath() ) {
	// cout << "removed " << (*j)->get_rcpath()[0] << endl;
	j = cnodes.erase(j);
      } 	
      else j++;
    } // end of while loop    
  }

}

double CubeCollect::
get_vrsev(int mode[3], const Metric* met, const Region* reg, const Location* loc) {
  double sum = 0.0;
  vector<const Cnode*> cns;

  cns = get_callees(reg);
  if (cube->get_attr("CUBE_CT_AGGR") == "SUM"){
    for (int i = 0; i < cns.size(); i++) {
      sum += get_vcsev(mode, met, cns[i], loc);
    }
  } else if (cube->get_attr("CUBE_CT_AGGR") == "MAX") {
    if(!is_simple(cns)) {
      rm_suffix_prefix(cns); // suffix and prefix call-path
    }
    for (int i = 0; i < cns.size(); i++) {
      sum += get_vcsev(mode, met, cns[i], loc);
    }
  } else {
    cerr << "Error: call-tree aggregation mode undefined!" << endl;
  }
  return sum;
}

double CubeCollect::
get_vmsev(int mode[3], const Metric* met, const Module* mod, const Location* loc){
  double sev = 0;
  int mode2[3];
 
  if (mode[CUBE_PROG] == CUBE_INCL) {
    // add up all its callees' values
    for (int i = 0; i < mod->num_children(); i++) {
      const Region* r = mod->get_child(i);
      mode2[CUBE_MET] = mode[CUBE_MET];
      mode2[CUBE_PROG] = CUBE_EXCL;
      mode2[CUBE_LOC]  = mode[CUBE_LOC];
      sev += get_vrsev(mode2, met, r, loc);
    }
  }
  else { 
    sev = 0; // If expanded modul
  }
  
  return sev;
}
