/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/*******************************
				
     FlatTree.cpp	
					
********************************/

#include <iostream>
#include <fstream>

#include "Tree.h"
#include "MidTree.h"
#include "FlatTree.h"
#include "LocationTree.h"

#include "PresenterApp.h"
#include "PresenterFrame.h"
#include "SourceCodeDlg.h"
#include "GUIData.h"
#include "CubeCollect.h"
#include "Cube.h"

#include "Cnode.h"
#include "Callsite.h"
#include "Region.h"
#include "Module.h"
#include "Browser.h"

using namespace std;
using namespace cube;

int FlatTree::mydepth(wxTreeItemId id) {
  int depth = 0;
  while (id != GetVRootItem()) {
    id = GetItemParent(id);
    depth++;
  }
  return depth;
}

bool FlatTree::is_module (wxTreeItemId id) {
  return (mydepth(id) == 1);
}

bool FlatTree::is_region (wxTreeItemId id) {
  return (mydepth(id) == 2);
}

bool FlatTree::is_subreg (wxTreeItemId id) {
  return (mydepth(id) == 3);
}

void FlatTree::OnLoc(wxCommandEvent& event) {
  PresenterFrame* frame = myApp->get_frame();
  Vertex* obj = GetItemData(right_click_item)->get_obj();
  
  if (is_subreg(right_click_item)) {
    wxMessageBox(_T("Subroutines of its parent node."), _T("Location"), wxOK, this);
  } else if (is_module(right_click_item)) {
    wxMessageBox(_T("This is a module."), _T("Location"), wxOK, this);
  } else if (is_region(right_click_item)){
    show_descr_dlg4reg(static_cast<Region*>(obj));
  } else {
    cerr << "Warning: Can't decide which type of node in Flat Tree" << endl;
  }
}

void FlatTree::OnDes(wxCommandEvent& event) {
  PresenterFrame* frame = myApp->get_frame();
  wxString text;
  Vertex* obj = GetItemData(right_click_item)->get_obj();
  if (obj != NULL) {
    if (is_subreg(right_click_item)) {
      wxMessageBox(_T("Subroutines of its parent node."), _T("Subroutines Information"), 
		   wxOK, this);
    } else if (is_module(right_click_item)) {
      wxMessageBox(_T("This is a module."), _T("Module Information"), wxOK, this);
    } else if (is_region(right_click_item)){
      Region* r = static_cast<Region*>(obj);
      string descr = r->get_descr();
      text.Printf("%s\n", descr.c_str());
      wxMessageBox(text, _T("Region Information"), wxOK, this);
    } else {
      cerr << "Warning: Can't decide which type of node in Flat Tree" << endl;
    }
  }
  
}

void FlatTree::OnURL (wxCommandEvent& event) {
  PresenterFrame* frame = myApp->get_frame();
  Vertex* obj = GetItemData(right_click_item)->get_obj();
  if (obj != NULL) {
    Region* r = static_cast<Region*>(obj);
    string url = r->get_url();
    if (frame->get_browser_dlg() == NULL)
      frame->set_browser_dlg(new Browser(frame));
    Browser* browser = frame->get_browser_dlg();
    int ret = frame->load_working_url(browser, url);
    if (ret < 0) {
      cerr << "Error: " << url << " not accessible." << endl;
      return;
    }
    if (ret > 0) {
      browser->Show(true);
      browser->Raise();
    }
  }
}

/******************************************************************
   display source code accordingly 
   note: the same as cnode's.

*******************************************************************/
void FlatTree::OnSourceCode(wxCommandEvent& event) {
  PresenterFrame* frame = myApp->get_frame();
  GUIData* guidata = frame->get_guidata();
  string mname;
  Vertex* obj = GetItemData(right_click_item)->get_obj();
  
  // 1) If subroutines, show nothing.
  if (is_subreg(right_click_item)) return;
  
  SourceCodeDlg* dlg = frame->get_source_dlg();
  if (dlg == NULL) {
    dlg = new SourceCodeDlg(this);
    dlg->Move(frame->GetPosition()-dlg->GetPosition()-wxPoint(400,0));
    frame->set_source_dlg(dlg);
  }
  // 2) show code for a whole module
  if (is_module(right_click_item)) { 
    const Module* m = static_cast<Module*>(obj);
    mname = m->get_name();
    frame->get_source_dlg()->set_module(mname);
    frame->get_source_dlg()->set_line(1);
    string ln = get_line(get_path(mname), get_base(mname), 1);
    if (ln.empty()) return; // user clicked "Cancel"
    display_region_code(0, 0);     
    return;
  }
  // 3) show code for a regular function
  const Region* r = static_cast<Region*>(obj);
  const Module* m = r->get_mod();
  mname = m->get_name();
  frame->get_source_dlg()->set_module(mname);
  frame->get_source_dlg()->set_line(r->get_begn_ln());
  string ln = get_line(get_path(mname), get_base(mname), 1);
  if (ln.empty()) return; // user clicked "Cancel"
  display_region_code(r->get_begn_ln(), r->get_end_ln());  
  return;
}
  
FlatTree::FlatTree(wxWindow* parent) : MidTree(parent) {

}

void FlatTree::OnExpanded(wxTreeEvent& event) {
  wxTreeItemId id = event.GetItem();
  string label;
  PresenterFrame* frame = myApp->get_frame();
  Vertex* obj = GetItemData(id)->get_obj();
  GUIData* guidata = frame->get_guidata();

  if (IsSelected(id)) guidata->xpnd_sel_middle();
  if (is_module(id)) {
    const Module* m = static_cast<Module*>(obj);
    label = guidata->get_mod_label(CUBE_EXCL, m);
    SetItemText(id, label, guidata->get_mod_color(CUBE_EXCL, m));
  } else if (is_region(id)) {
    const Region* r = static_cast<Region*>(obj);
    label = guidata->get_reg_label(CUBE_EXCL, r);
    SetItemText(id, label, guidata->get_reg_color(CUBE_EXCL, r));
  } else {
    cerr << "Warning: subroutine is being expaned!" << endl;
    const Region* r = static_cast<Region*>(obj);
    label = guidata->get_sreg_label(r);
    SetItemText(id, label, guidata->get_sreg_color(r));
  }
  
  if (IsSelected(id)) frame->get_loc_tree()->refresh();
  
}

void FlatTree::OnCollapsed(wxTreeEvent& event) {
  wxTreeItemId id = event.GetItem();
  string label;
  PresenterFrame* frame = myApp->get_frame();
  Vertex* obj = GetItemData(id)->get_obj();
  GUIData* guidata = frame->get_guidata();
  /*
  int ct_aggr_mode = guidata->get_ct_aggr_mode();
  if (is_module(id) && 
      guidata->get_midtree_type() == MID_MODULE_TREE &&
      ct_aggr_mode == CUBE_CT_AGGR_MAX) {
    Expand(id); // for MAX mode, we can't collapse modules!
    return;
    }*/

  if (IsSelected(id)) guidata->clps_sel_middle();
  if (is_module(id)) {
    const Module* m = static_cast<Module*>(obj);
    label = guidata->get_mod_label(CUBE_INCL, m);
    SetItemText(id, label, guidata->get_mod_color(CUBE_INCL, m));
  } else if (is_region(id)) {
    const Region* r = static_cast<Region*>(obj);
    label = guidata->get_reg_label(CUBE_INCL, r);
    SetItemText(id, label, guidata->get_reg_color(CUBE_INCL, r));
  } else {
    cerr << "Warning: subroutine is being collapsed" << endl;
    const Region* r = static_cast<Region*>(obj);
    label = guidata->get_sreg_label(r);
    SetItemText(id, label, guidata->get_sreg_color(r));
  }
  
  if (IsSelected(id)) frame->get_loc_tree()->refresh();
}


void FlatTree::OnSelChanged(wxTreeEvent& event) {
  wxTreeItemId id = GetSelection(); // newly selected node.
  PresenterFrame* frame = myApp->get_frame();
  Vertex* obj = GetItemData(id)->get_obj();
  if (obj == NULL) {
    cerr << "Warning: flat tree node selected has a NULL obj" << endl;  
    return;
  }
  GUIData* guidata = frame->get_guidata();
  if (is_module(id)) {
    const Module* m = static_cast<Module*>(obj);
    guidata->select_module(m);
  } else if (is_region(id)) {
    const Region* r = static_cast<Region*>(obj);
    guidata->select_region(r);
  } else if (is_subreg(id)) {
    const Region* r = static_cast<Region*>(obj);
    guidata->select_subregion(r);
  } else {
    cerr << "Warning: can't identify flat tree node type" << endl;
    return;
  }
  IsExpanded(id) ? guidata->xpnd_sel_middle() : guidata->clps_sel_middle();
  if (frame->get_loc_tree() != NULL)
    frame->get_loc_tree()->refresh();

}

void FlatTree::OnRightClick(wxTreeEvent& event) {
  right_click_item = HitTest(event.GetPoint());
  
  if (!is_region(right_click_item)) { 
    // modules don't have "loc", "Online Descr", "Info" menu item.
    // subregions has nothing.
    popup_menu->Enable(MID_POP_ID_Loc, false);
    popup_menu->Enable(MID_POP_ID_URL, false);
    popup_menu->Enable(MID_POP_ID_Des, false);
    if (is_subreg(right_click_item)) popup_menu->Enable(MID_POP_ID_Code, false);
    PopupMenu(popup_menu, event.GetPoint());
    if (is_subreg(right_click_item)) popup_menu->Enable(MID_POP_ID_Code, true);
    popup_menu->Enable(MID_POP_ID_Loc, true);
    popup_menu->Enable(MID_POP_ID_URL, true);
    popup_menu->Enable(MID_POP_ID_Des, true);
  } else { // hey, I'm a region!
    PresenterFrame* frame = myApp->get_frame();
    Vertex* obj = GetItemData(right_click_item)->get_obj();
    if (obj == NULL) return;

    Region* rn       = static_cast<Region*>(obj);
    int bln          = rn->get_begn_ln();
    int eln          = rn->get_end_ln();
    bool disable_src = (bln < 0 && eln < 0);
    string url       = rn->get_url();
    string des       = rn->get_descr();
    
    if (url.empty()) popup_menu->Enable(MID_POP_ID_URL,  false);
    if (des.empty()) popup_menu->Enable(MID_POP_ID_Des,  false);
    if (disable_src) popup_menu->Enable(MID_POP_ID_Code, false);
    PopupMenu(popup_menu, event.GetPoint());
    if (url.empty()) popup_menu->Enable(MID_POP_ID_URL,  true);
    if (des.empty()) popup_menu->Enable(MID_POP_ID_Des,  true);
    if (disable_src) popup_menu->Enable(MID_POP_ID_Code, true);
 
  }
  
  /* else {
    PresenterFrame* frame = myApp->get_frame();
    Vertex* obj = GetItemData(right_click_item)->get_obj();
    if (obj == NULL) return;
    Region* rn = static_cast<Region*>(obj);
    string url = rn->get_url();
    if(!url.empty()) {//dynamically add menu item
      string text = "Goto URL";
      popup_menu->Append(MID_POP_ID_URL, text.c_str());
    }
    PopupMenu(popup_menu, event.GetPoint());
    if(!url.empty()) {//dynamically remove menu item
      popup_menu->Delete(MID_POP_ID_URL);
    }
    }*/
}

void FlatTree::refresh () {
  if (GetChildrenCount(GetVRootItem(), false) == 0) return;

  traverse(GetVRootItem());
#ifdef CUBE_DEBUG
  cout << "flat tree is refreshed" << endl;
#endif
  PresenterFrame* frame = myApp->get_frame();
  if (frame->get_loc_tree() != NULL)
    frame->get_loc_tree()->refresh();
}

void FlatTree::traverse(wxTreeItemId root) {
  long cookie;
  int mode;
  PresenterFrame* frame = myApp->get_frame();
  GUIData* guidata = frame->get_guidata();

  Vertex* obj = GetItemData(root)->get_obj();
  /*
  int ct_aggr_mode = guidata->get_ct_aggr_mode();
  if (obj != NULL && is_module(root) && 
      guidata->get_midtree_type() == MID_MODULE_TREE &&
      ct_aggr_mode == CUBE_CT_AGGR_MAX) {
    Expand(root); // for MAX mode, we can't collapse modules!
    }*/
  
  mode = IsExpanded(root) ? CUBE_EXCL : CUBE_INCL;

  if (obj != NULL) {
    if(is_module(root)) {
      // cout << "module tree" << endl;
      const Module* m = static_cast<Module*>(obj);
      SetItemText(root, guidata->get_mod_label(mode, m), 
		  guidata->get_mod_color(mode, m));
    } else if(is_region(root)) {
      const Region* r = static_cast<Region*>(obj);
      SetItemText(root, guidata->get_reg_label(mode, r), 
		  guidata->get_reg_color(mode, r));
    } else {
      const Region* r = static_cast<Region*>(obj);
      SetItemText(root, guidata->get_sreg_label(r), 
		  guidata->get_sreg_color(r));
    }

  }
  if (GetChildrenCount(root) > 0) {
    wxTreeItemId child = GetFirstChild(root, cookie); 
    traverse(child);
    while (child = GetNextChild(root, cookie)) {
      traverse(child);
    }
  }

}

