/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/*******************************
				
	PresenterApp.cpp	
					
********************************/
#include <iostream>
#include <fstream>

#include "wx/image.h"
#include "wx/fs_inet.h"
#include "wx/filesys.h"

#include "PresenterApp.h"
#include "PresenterFrame.h"
#include "SpectrumPane.h"
#include "Cube.h"

using namespace std;
using namespace cube;

const int PresenterApp::square_size;

bool PresenterApp::OnInit() {
  parse_command();
  create_image_list();
  m_process = 0;
  m_thread = 0;
  main_frame = 
    new PresenterFrame(NULL, -1, _T("CUBE "), 
		       wxPoint(0, 0), wxSize(800, 600),
		       wxDEFAULT_FRAME_STYLE |
		       wxNO_FULL_REPAINT_ON_RESIZE |
		       wxHSCROLL | wxVSCROLL); 
  main_frame->CentreOnScreen(wxBOTH);
  main_frame->Show(TRUE);
  SetTopWindow(main_frame);
  if (cur_file.empty()) return true; // no filename in command line
  

  ::wxInitAllImageHandlers();
  // wxImage::AddHandler(new wxJPEGHandler);
  // wxImage::AddHandler(new wxPNGHandler);
  // wxImage::AddHandler(new wxGIFHandler);
  wxFileSystem::AddHandler(new wxInternetFSHandler); //wxHtmlWindow online support

  // cube_home = locate_cube();
  // cube_home = cube_home + ".." + wxFILE_SEP_PATH;
  // decide local mirrors
  // string docpath = get_cube_home() + "doc";
  // find_files(docpath, "*.html", local_mirrors);

  main_frame->load_file(cur_file.c_str());

  return true;
} 

PresenterApp::~PresenterApp() {
  delete images;
}

// only support for a single option.
bool parse_option(string option, string usage) {
  if (option.compare("-h") == 0 ||
      option.compare("--help") == 0) {
    cout << usage << endl;
    exit(0);
  } else if (option.compare("-v") == 0 ||
	     option.compare("--version") == 0) {
    cout << "CUBE   Version 2.0" << endl << endl;
    cout << "Copyright(c) 2004 ICL at UTK" << endl;
    exit(0);
  }
  else if (option[0] == '-') {
    // invalid option
    cout << "Invalid option, type -h for help" << endl;
    exit(0);
  }
  return false; // not an option at all 
}
 
void PresenterApp::parse_command() {
  int argc = wxApp::argc;
  char** argv = wxApp::argv;
  string usage = "Usage: cube [option | filename] \n";
  usage += "  where options are --help(-h), --version(-v)";
  
  switch (argc) {
  case 1: // cube
    break;
  case 2: 
    if (parse_option(argv[1], usage)) {// cube option
      // if a valid option affecting the execution of presenter, do sth.
      
    }
    else { // cube filename
      char* fname = argv[1];
      std::ifstream in(fname);
      if (!in) {
	cerr << "File open error: " << fname << endl;
	exit(1);
      }
      in.close();
      cur_file = fname; // if success
    }
    break;
  default:
    cout << usage << endl;
    exit(1);
  }
  
  return;
}

void PresenterApp::create_image_list() {
  images = new wxImageList(square_size, square_size, true);
  wxMemoryDC mem_dc;
  wxBrush brush;
  wxPen pen;
  wxColour c;
  wxBitmap* box;
  wxColour bg_color = wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWFRAME); 
  wxColour shadow_color = wxSystemSettings::GetColour(wxSYS_COLOUR_3DDKSHADOW);
  wxColour light_color = wxSystemSettings::GetColour(wxSYS_COLOUR_3DHIGHLIGHT);
  
  m_process = 0;
  m_thread  = 0;
  int delta = 4; // padding between color box and labels.

  /* To create raised squares in advance 
   * 0 is background color, 101 is white(means error), 102 is mygrey
   * 1 .. 100 are spectrum colors */
  int i = 0;
  for (i = 0; i <= 102; i++) {
    box = new wxBitmap(square_size+delta, square_size);
    mem_dc.SelectObject(*box);
    mem_dc.SetPen(*wxBLACK_PEN);
    if (i == 102) c = SpectrumPane::severity2color(0.2);
    else c = SpectrumPane::severity2color(i);
    brush.SetColour(c);
    mem_dc.SetBrush(brush);
    mem_dc.DrawRectangle(0, 0, square_size, square_size);

    // raised effect, 0 is just flat.
    if (i > 0) {
      pen.SetColour(shadow_color);
      mem_dc.SetPen(pen);
      mem_dc.DrawLine(0, square_size-2, square_size, square_size-2);
      mem_dc.DrawLine(0, square_size-3, square_size, square_size-3);
      mem_dc.DrawLine(square_size-2, 0, square_size-2, square_size);
      pen.SetColour(light_color);
      mem_dc.SetPen(pen);
      mem_dc.DrawLine(1, 1, square_size-2, 1);
      mem_dc.DrawLine(1, 1, 1, square_size-2);
    }
    // draw right padding part
    brush.SetColour(bg_color);
    pen.SetColour(bg_color);
    mem_dc.SetBrush(brush);
    mem_dc.SetPen(pen);
    mem_dc.DrawRectangle(square_size, 0, delta, square_size);
    images->Add(*box);
  }


  /* To create sunken squares in advance 
   * starting from index of 103, 103+0 ... 103+102 */  
  i = 0;
  for (i = 0; i <= 102; i++) {
    box = new wxBitmap(square_size+delta, square_size);
    mem_dc.SelectObject(*box);
    mem_dc.SetPen(*wxBLACK_PEN);
    if (i == 102) c = SpectrumPane::severity2color(0.2);
    else c = SpectrumPane::severity2color(i);
    brush.SetColour(c);
    mem_dc.SetBrush(brush);
    mem_dc.DrawRectangle(0, 0, square_size, square_size);

    // sunken effect
    if (i > 0) { // if 103+0, just flat.
      pen.SetColour(light_color);
      mem_dc.SetPen(pen);
      mem_dc.DrawLine(0, square_size-1, square_size, square_size-1);
      mem_dc.DrawLine(0, square_size-2, square_size, square_size-2);
      mem_dc.DrawLine(square_size-1, 0, square_size-1, square_size);
      pen.SetColour(shadow_color);
      mem_dc.SetPen(pen);
      mem_dc.DrawLine(1, 1, square_size-2, 1);
      mem_dc.DrawLine(1, 1, 1, square_size-2);
    }
    // draw right padding part
    brush.SetColour(bg_color);
    pen.SetColour(bg_color);
    mem_dc.SetBrush(brush);
    mem_dc.SetPen(pen);
    mem_dc.DrawRectangle(square_size, 0, delta, square_size);
    images->Add(*box);
  }

}

wxImageList* PresenterApp::get_image_list() {
  return images;
}


IMPLEMENT_APP(PresenterApp)

