/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


/*******************************
				
     PresenterFrame.cpp	
					
********************************/

#include <stdlib.h>
#include <unistd.h>

#include <iostream>
#include <fstream>
#include <algorithm>
#include <math.h>

#include <wx/filesys.h>

#include "PresenterFrame.h"
#include "PresenterApp.h"
#include "SpectrumPane.h"

#include "Metric.h"
#include "Cnode.h"
#include "Callsite.h"
#include "Region.h"
#include "Location.h"
#include "Module.h"
#include "Machine.h"
#include "GUIData.h"
#include "Cube.h"
#include "CubeCollect.h"
#include "CubeCached.h"
#include "Space.h"

#include "wx/treectrl.h"
#include "MetricTree.h"
#include "CnodeTree.h"
#include "FlatTree.h"
#include "FlatRegionTree.h"
#include "LocationTree.h"
#include "AggrModeDlg.h"
#include "iclogo.xpm"
#include "Browser.h"
#include "CUBException.h"

#include "cube_install.h"

using namespace std;
using namespace cube;

// declaration of const variable (local linkage).
const int PresenterFrame::PADDING;

BEGIN_EVENT_TABLE(PresenterFrame, wxFrame)
  EVT_SIZE               (PresenterFrame::OnSize)
  EVT_SASH_DRAGGED_RANGE (ID_Metric_Panel, ID_Location_Panel, PresenterFrame::OnSashDrag)
  // EVT_CLOSE   (PresenterFrame::OnExit)
  // EVT_PAINT              (PresenterFrame::OnPaint)
  // For menu items
  EVT_MENU               (ID_Open_File, PresenterFrame::OnOpenFile)
  EVT_MENU               (ID_Close_File, PresenterFrame::OnCloseFile)
  EVT_MENU               (ID_Exit, PresenterFrame::OnExit)
  EVT_MENU               (ID_About, PresenterFrame::OnAbout)
  EVT_MENU               (ID_ModuleTree_View, PresenterFrame::OnModuleTreeView)
  EVT_MENU               (ID_RegionTree_View, PresenterFrame::OnRegionTreeView)
  EVT_MENU               (ID_Calltree_View, PresenterFrame::OnCalltreeView)
  EVT_MENU               (ID_Absolute_View, PresenterFrame::OnAbsoluteView)
  EVT_MENU               (ID_PercentRef_View,  PresenterFrame::OnPercentRefView)
  EVT_MENU               (ID_Percent_View,  PresenterFrame::OnPercentageView)
  EVT_MENU               (ID_Relative_View, PresenterFrame::OnRelativeView)
  EVT_MENU               (ID_Positive_Only, PresenterFrame::OnPositiveOnly)
  EVT_MENU               (ID_Negative_Only, PresenterFrame::OnNegativeOnly)

END_EVENT_TABLE()

/* create and initialize a metric tree */
void  PresenterFrame::build_met_tree() {
  if (gui_data == NULL || gui_data->empty() || met_tree == NULL) return;

  wxTreeItemId firstroot;
  const vector<Metric*>& roots = gui_data->get_met_roots();
  for (int k = 0; k < roots.size(); k++) { // several metric roots are possible
    wxTreeItemId rootId = met_tree->AddRoot("NULL", 101); // temporary label
    Metric* p = roots[k];
    met_tree->SetItemData(rootId, new NodeData(p));
    if (k == 0)  {
      firstroot = rootId;
      // met_tree->SelectItem(rootId);
    }
    /* clone a wxTree from a cube metric tree*/
    vector<Metric*>  fromv[2]; // store ids for only two levels
    vector<wxTreeItemId>        tov[2];
    int depth = 0;
    fromv[depth].push_back(p);       // add root of cube tree
    tov[depth].push_back(rootId);    // add root of wxWindows tree
    
    while(fromv[depth%2].size() > 0) {
      //building tree from level #depth to level #(depth+1) 
      for (int i = 0; i < fromv[depth%2].size(); i++) {
	p = fromv[depth%2][i];
	wxTreeItemId pid = tov[depth%2][i];
	for (int j = 0; j < p->num_children(); j++) {
	  Metric* child = const_cast<Metric*>(p->get_child(j));
	  // store to the alternative vector
	  fromv[(depth+1)%2].push_back(child);
	  tov[(depth+1)%2].
	    push_back(met_tree->AppendItem(pid, "NULL", 101, new NodeData(child)));
	}
      }
      fromv[depth%2].clear();
      tov[depth%2].clear();
      depth++;
    } // end of construction of one tree
  
  } // end of construction of one forest
  met_tree->SelectItem(firstroot);
  // gui_data->select_metric(static_cast<Metric*>
  //			 (met_tree->GetItemData(firstroot)->get_obj()));
  met_tree->refresh();

}

/* rebuild a cnode tree */
void  PresenterFrame::build_cnode_tree() {
  if (gui_data == NULL || gui_data->empty() || cnode_tree == NULL) return;

  wxTreeItemId firstroot;
  const vector<Cnode*>& roots = gui_data->get_cnode_roots();
  if (roots.size() == 0) { 
    cerr << "Error: call trees roots = " << roots.size() << endl;
    return;
  }
  for (int k = 0; k < roots.size(); k++) {
    wxTreeItemId rootId = cnode_tree->AddRoot("NULL", 101); // temporary label
    Cnode* cnode = roots[k]; // get a root node
    cnode_tree->SetItemData(rootId, new NodeData(cnode));
    if (k == 0)  {
      firstroot = rootId;
      // cnode_tree->SelectItem(rootId);
    }
    vector<Cnode*>              fromv[2];        // store ids for only two levels
    vector<wxTreeItemId>        tov[2];
    int depth = 0;
    fromv[depth].push_back(cnode);   // add root of cube tree
    tov[depth].push_back(rootId);    // add root of wxWindows tree
    while(fromv[depth%2].size() > 0) {
      //building tree from depth to depth+1
      for (int i = 0; i < fromv[depth%2].size(); i++) {
	cnode = fromv[depth%2][i];
	wxTreeItemId pid = tov[depth%2][i];
	for (int j = 0; j < cnode->num_children(); j++) {
	  Cnode* child = (Cnode*) cnode->get_child(j);
	  // store to the alternative vector
	  fromv[(depth+1)%2].push_back(child);
	  tov[(depth+1)%2].
	    push_back(cnode_tree->AppendItem(pid, "NULL", 101, new NodeData(child)));
	}
      }
      fromv[depth%2].clear();
      tov[depth%2].clear();
      depth++;
    }
  }

  //const NodeData* tmp = cnode_tree->GetItemData(firstroot);
  //if (tmp == NULL) {
  //  cout << "Error: build cnode tree, null node data" << endl;
  //  return;
  //}
  //gui_data->select_cnode(static_cast<Cnode*>
  //			(cnode_tree->GetItemData(firstroot)->get_obj()));
  cnode_tree->SelectItem(firstroot);
  cnode_tree->refresh();
}

// rebuild a flat module tree
void PresenterFrame::build_module_tree() {
  wxTreeItemId sel_func; // default one initially
  if (gui_data == NULL || gui_data->empty() || module_tree == NULL) return;

  const vector<Module*>& modv = gui_data->get_mod_roots();
  vector<Module*> modv2 = const_cast<vector<Module*>&>(modv);
  sort(modv2.begin(), modv2.end(), GenLess<Module*>()); // reorder modules by name
  for (int i = 0; i < modv.size(); i++) {
    wxTreeItemId pid = module_tree->AddRoot("NULL", 101);
    if (i == 0) {
      sel_func = pid;
      // module_tree->SelectItem(pid);
    }
    Module* m = modv[i];
    module_tree->SetItemData(pid, new NodeData(m));
    m->sort_regn();                                    // reorder regions by name
    // for each module, append its functions as child nodes
    for (int j = 0; j < m->num_regions(); j++) {
      Region* r = const_cast<Region*>(m->get_child(j));
      if(r->get_name()=="virtual caller") continue;
      wxTreeItemId funcid = 
	module_tree->AppendItem(pid, "NULL", 101, new NodeData(r));
      if (r->num_children() > 0) // append subregions to each function
	module_tree->AppendItem(funcid, "Subregions", 101, new NodeData(r));
    } // end of inner for loop
  } // end of outer for loop
  module_tree->SelectItem(sel_func);//trigger sel_changed() event handler automatically.
  module_tree->refresh();

}


void PresenterFrame::build_region_tree() {
  wxTreeItemId sel_func; // default one initially
  if (gui_data == NULL || gui_data->empty() || region_tree == NULL) return;
  const vector<Region*>& regns = gui_data->get_regionv();
  vector<Region*>& regns2 = const_cast<vector<Region*>&>(regns);
  sort(regns2.begin(), regns2.end(), GenLess<Region*>()); // reorder regions by name
  for (int i = 0; i < regns.size(); i++) {
    wxTreeItemId pid = region_tree->AddRoot("NULL", 101);
    if (i == 0) {
      sel_func = pid;
      //region_tree->SelectItem(pid);
    }
    Region* r = regns[i];
    if(r->get_name()=="virtual caller") continue;
    region_tree->SetItemData(pid, new NodeData(r));
    if (r->num_children() > 0) // append subregions to each function
      region_tree->AppendItem(pid, "Subregions", 101,
			      new NodeData(r));
  }
  region_tree->SelectItem(sel_func);//trigger sel_changed() event handler automatically.
  region_tree->refresh();
}


/* create and initialize a location tree */
void  PresenterFrame::build_loc_tree() {
  if (gui_data == NULL || gui_data->empty() || loc_tree == NULL) return;

  const int process_level = 2;
  const vector<Machine*>& roots = gui_data->get_loc_roots();

  for (int k = 0; k < roots.size(); k++) {
    wxTreeItemId rootId = loc_tree->AddRoot("NULL", 101);
    Location* p = roots[k]; 
    loc_tree->SetItemData(rootId, new NodeData(p));
    vector<Location*>           fromv[2]; // store ids for only two levels
    vector<wxTreeItemId>        tov[2];
    int depth = 0;
    fromv[depth].push_back(p);       // add root of cube tree
    tov[depth].push_back(rootId);    // add root of wxWindows tree

    while(fromv[depth%2].size() > 0) {
      // building tree from depth to depth+1
      for (int i = 0; i < fromv[depth%2].size(); i++) {
	if (depth == process_level && gui_data->max_num_thrd() <= 1) 
	  continue; // no threads displayed
	p = fromv[depth%2][i];
	wxTreeItemId pid = tov[depth%2][i];
	// if building smp nodes, reorder its children
	if (depth == process_level-1) {
	  Node* n = static_cast<Node*>(p);
	  n->sort_proc();
	}
	for (int j = 0; j < p->num_children(); j++) {
	  Location* child = (Location*) p->get_child(j) ;
	  fromv[(depth+1)%2].push_back(child);
	  tov[(depth+1)%2].push_back
	    (loc_tree->AppendItem(pid, "NULL", 101, new NodeData(child)));
	}
      }
      fromv[depth%2].clear();
      tov[depth%2].clear();
      depth++; // move down to next level
    }
  }
  loc_tree->refresh();
}

PresenterFrame::
PresenterFrame(wxWindow *parent, 
	       const wxWindowID id, 
	       const wxString& title, 
	       const wxPoint& pos, 
	       const wxSize& size,
	       const long style): wxMDIParentFrame(parent, id, title, pos, size, style) {
  met_tree   = NULL;
  cnode_tree = NULL;
  module_tree  = NULL;
  region_tree = NULL;
  loc_tree   = NULL;
  myApp = static_cast<PresenterApp*>(wxTheApp);
  SetBackgroundColour(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWFRAME));
  logo = new wxBitmap(iclogo);
  gui_data = NULL;

  /* create "File" menu */
  wxMenu* file_menu = new wxMenu;
  file_menu->Append(ID_Open_File, "&Open...");
  file_menu->Append(ID_Close_File, "&Close"); 
  file_menu->AppendSeparator();
  file_menu->Append(ID_Exit, "E&xit" );

  /* create "View" menu */
  view_menu = new wxMenu;

  view_menu->AppendCheckItem(ID_Calltree_View, "Call tree");
  view_menu->Check(ID_Calltree_View, true); // default check
  view_menu->AppendCheckItem(ID_ModuleTree_View, "&Module profile");
  view_menu->AppendCheckItem(ID_RegionTree_View, "&Region profile");
  view_menu->AppendSeparator();

  view_menu->AppendCheckItem(ID_Absolute_View, "&Absolute");
  view_menu->Check(ID_Absolute_View, true); // default check
  view_menu->AppendCheckItem(ID_Percent_View,  "&Percentage");
  view_menu->AppendCheckItem(ID_Relative_View, "R&elative percentage");
  view_menu->AppendCheckItem(ID_PercentRef_View,  "&Comparative percentage ...");
  view_menu->AppendSeparator();
  
  hide_menu = new wxMenu;
  view_menu->Append(ID_HIDE_SUBMENU, "Hide values", hide_menu);
  hide_menu->AppendCheckItem(ID_Positive_Only, "Positive");
  hide_menu->Check(ID_Positive_Only, false);
  hide_menu->AppendCheckItem(ID_Negative_Only, "Negative");
  hide_menu->Check(ID_Negative_Only, false);

  /* create "Help" menu */
  wxMenu* help_menu = new wxMenu;
  help_menu->Append(ID_About, "&About");

  /* create Menu Bar on wxFrame */
  wxMenuBar* menu_bar = new wxMenuBar;
  menu_bar->Append(file_menu, "&File");
  menu_bar->Append(view_menu, "&View");
  menu_bar->Append(help_menu, "&Help");
  SetMenuBar(menu_bar);

  CreateStatusBar(2);
  status_bar = GetStatusBar();
  init_statusbar(); // create n x m for mpi/openMP

  /* create color spectrum pane */
  spectrum_pane = new SpectrumPane(this, 1000);
  int default_w = 800/3; // default window width
  met_panel     = create_metric_pane(default_w, ID_Metric_Panel);
  mid_panel   =   create_prog_pane(default_w, ID_Cnode_Panel);
  loc_panel     = create_sys_pane(default_w, ID_Location_Panel);

}

PresenterFrame::~PresenterFrame() {
  if (gui_data != NULL) delete gui_data;
  // close the alive browser window.
  if (get_browser_dlg() != NULL) {
    get_browser_dlg()->OnOK();
  }
}

/* Open file dialog */
void PresenterFrame::OnOpenFile(wxCommandEvent& event) {
  const wxString path=
    ::wxFileSelector(_T("Select a cube file to load"),
		     _T(""), _T(""), _T(""),
		     _T("cube files (*.cube)|*.cube|All files (*.*)|*.*"),
		     wxOPEN|wxCHANGE_DIR,this);
  if (path.empty())  return;
  load_file(path.c_str());
}

void PresenterFrame::load_file(const char* path) {
  wxCommandEvent tmp;
  ifstream cubefile(path);
  if (!cubefile) 
    wxLogMessage(_T("Open file error: '%s'"), (const wxChar*) path);

  // close the previous file if necessary
  OnCloseFile(tmp);
  string title = "CUBE:  ";
  // only display a file's base name
  title += get_base(path);
  SetTitle(title.c_str());
  Cube* cube = new Cube();
  try {
    cubefile >> *cube;
  } catch (CUBExceptionWarning w) {
    if (w.get_msg().find("version") == 0) {
      wxString msg;
      msg.Printf("CUBE file %s incompatible with CUBE version 2.0!", 
                 w.get_msg().c_str());
      wxMessageBox(msg, "Error", wxICON_ERROR, this);
      return;
    }
  }
  cubefile.close();

  determine_ct_aggr_mode(cube);

  if (gui_data == NULL)
    gui_data = new GUIData(new CubeCached(new CubeCollect(cube)));
  else
    gui_data->set_cube(new CubeCached(new CubeCollect(cube)));
  myApp->set_proc_num(gui_data->num_procs());
  myApp->set_thrd_num(gui_data->max_num_thrd());
  change_label_mn();

  build_met_tree();
  build_cnode_tree();
  build_module_tree();
  build_region_tree();
  build_loc_tree();
  
  // Disable some menu items
  if (cube->get_attr("CUBE_CT_AGGR") == "MAX") {
    view_menu->Enable(ID_ModuleTree_View, false);
    cout << endl;
    cout<<"**************************************************************" << endl;
    cout<<"Note: module-tree view is disabled since you are using MAX mode"<< endl;
    cout<<"**************************************************************" << endl;
  }
  // for flat-tree data, we can't show call-tree.
  if (gui_data->flatree_only_data()) {
    /* commented since we always reset guidata.
    wxCommandEvent dummy;
    if (gui_data->get_midtree_type() == MID_MODULE_TREE) 
      OnModuleTreeView(dummy);
    else OnRegionTreeView(dummy); 
    */
    view_menu->Enable(ID_Calltree_View, false);
    cout << endl;
    cout<<"**************************************************************" << endl;
    cout<<"Note: call-tree view is disabled since your data is a flat-profile."<<endl;
    cout<<"**************************************************************" << endl;
    wxCommandEvent dummy;
    OnRegionTreeView(dummy);
  }

  spectrum_pane->Refresh();
}


/* Close current file */
void PresenterFrame::OnCloseFile(wxCommandEvent& event) {
  // /* removed because wxWidgets will do it by default, I can't do it!
  wxTreeItemId rootId;
  rootId = met_tree->GetVRootItem();
  met_tree->DeleteChildren(rootId);
  if (cnode_tree != NULL) {
    rootId = cnode_tree->GetVRootItem();
    cnode_tree->DeleteChildren(rootId);
  }
  if (module_tree != NULL) {
    rootId = module_tree->GetVRootItem();
     module_tree->DeleteChildren(rootId);
  }
  if (region_tree != NULL) {
    rootId = region_tree->GetVRootItem();
    region_tree->DeleteChildren(rootId);
  }
  rootId = loc_tree->GetVRootItem();
  loc_tree->DeleteChildren(rootId);


  if (gui_data != NULL) { // clean up data
    //if (gui_data->flatree_only_data()) {
    view_menu->Enable(ID_Calltree_View, true);   
    view_menu->Enable(ID_ModuleTree_View, true);
    // reset gui_data:
    wxCommandEvent dummy;
    OnCalltreeView(dummy);
    OnAbsoluteView(dummy);
    hide_menu->Check(ID_Positive_Only, false);
    hide_menu->Check(ID_Negative_Only, false);
    gui_data->reset();    
  }

  myApp->set_proc_num(0);
  myApp->set_thrd_num(0);
  change_label_mn();
  spectrum_pane->Refresh();
  SetTitle(_T("CUBE: no file"));  
}

void PresenterFrame::OnExit(wxCommandEvent& event) {
  // close browser too.
  if (get_browser_dlg() != NULL) {
    get_browser_dlg()->OnOK();
  }
  //Close(true);
  exit(0);
}

void PresenterFrame::OnAbout(wxCommandEvent& WXUNUSED(event))
{
  wxDialog* about = new wxDialog(this, -1, "About CUBE", 
				 wxPoint(0,0), wxSize(450, 180));
  wxStaticBitmap *icon = new wxStaticBitmap(about, -1, *logo);
  wxBoxSizer* topsizer = new wxBoxSizer(wxVERTICAL);
  wxBoxSizer* sizer = new wxBoxSizer(wxHORIZONTAL);
  sizer->Add(icon, 0, wxCENTER|wxLEFT, 30);
  sizer->Add(new wxStaticText
	     (about, -1, 
	      "\nCUBE Version 2.0\n\n(C) 2004 Innovative Computing Laboratory,\n     University of Tennessee at Knoxville.", 
	      wxDefaultPosition, wxSize(300, -1)),
	     0,
	     wxCENTER|wxLEFT, 10);
  topsizer->Add(sizer, 1,  wxCENTER | wxLEFT|wxRIGHT|wxTOP, 10);
  topsizer->Add(new wxButton(about, wxID_OK, _("&OK")), 0, 
		wxCENTER|wxALL, 10); 
  about->SetSizer(topsizer);
  about->Centre(wxBOTH);
  about->ShowModal();
  delete about;

}


void PresenterFrame::OnModuleTreeView  (wxCommandEvent& event) {
  view_menu->Check(ID_ModuleTree_View, true);
  view_menu->Check(ID_Calltree_View,   false);
  view_menu->Check(ID_RegionTree_View, false);
  cnodelabel->SetLabel(_T("Module Profile"));

  if (gui_data == NULL || gui_data->get_midtree_type() == MID_MODULE_TREE) return;
  int old_tree_type = gui_data->get_midtree_type();
  gui_data->set_midtree_type(MID_MODULE_TREE);
  if (gui_data->empty()) return;
  switch_mid_tree(old_tree_type, gui_data->get_midtree_type());
}

void PresenterFrame::OnRegionTreeView  (wxCommandEvent& event) {
  view_menu->Check(ID_RegionTree_View, true);
  view_menu->Check(ID_Calltree_View,   false);
  view_menu->Check(ID_ModuleTree_View, false);
  cnodelabel->SetLabel(_T("Region Profile"));

  if (gui_data == NULL || gui_data->get_midtree_type() == MID_REGION_TREE) return;
  int old_tree_type = gui_data->get_midtree_type();
  gui_data->set_midtree_type(MID_REGION_TREE);
  if (gui_data->empty()) return;
  switch_mid_tree(old_tree_type, gui_data->get_midtree_type());
}

void PresenterFrame::OnCalltreeView  (wxCommandEvent& event) {
  view_menu->Check(ID_Calltree_View,   true);
  view_menu->Check(ID_ModuleTree_View, false);
  view_menu->Check(ID_RegionTree_View, false);
  cnodelabel->SetLabel(_T("Call Tree"));
  
  if (gui_data == NULL || gui_data->get_midtree_type() == MID_CALL_TREE) return;
  int old_tree_type = gui_data->get_midtree_type();
  gui_data->set_midtree_type(MID_CALL_TREE);
  if (gui_data->empty()) return;
  switch_mid_tree(old_tree_type, gui_data->get_midtree_type());
}

// absolute severity value view
void PresenterFrame::OnAbsoluteView  (wxCommandEvent& event) {
  view_menu->Check(ID_Absolute_View, true);
  view_menu->Check(ID_PercentRef_View, false);
  view_menu->Check(ID_Relative_View, false);
  view_menu->Check(ID_Percent_View, false);
  if (gui_data == NULL) return;
  if (gui_data->get_sev_mode() != ABSOLUTE) {
    gui_data->set_sev_mode(ABSOLUTE);
    if (gui_data->empty()) return;
    spectrum_pane->Refresh();
    met_tree->refresh();
    get_mid_tree()->refresh();
    loc_tree->refresh();
  }
}

// reference percentage view
void PresenterFrame::OnPercentRefView  (wxCommandEvent& event) {
  view_menu->Check(ID_PercentRef_View,  true);
  view_menu->Check(ID_Relative_View, false);
  view_menu->Check(ID_Absolute_View, false);
  view_menu->Check(ID_Percent_View, false);
  if (gui_data == NULL) return;

  // if (gui_data->get_sev_mode() != PERCENT) {
    gui_data->set_sev_mode(PERCENT);
    if (gui_data->empty()) return;
    wxString path=
      ::wxFileSelector(_T("Select the reference cube file to load"),
		       _T(""), _T(""), _T(""),
		       _T("cube files (*.cube)|*.cube|All files (*.*)|*.*"),
		       wxOPEN|wxCHANGE_DIR,this);
    if ( !path )  return; //file selection cancelled by user
    ifstream cubefile((const char*)path);
    if (!cubefile) {
      wxLogMessage(_T("Open file error: '%s'"), (const wxChar*) path);
      return;
    }
    Cube* ref_cube = new Cube();
    cubefile >> *ref_cube;
    ref_cube->def_attr("CUBE_CT_AGGR", gui_data->get_attr("CUBE_CT_AGGR"));
    gui_data->set_refcube(new CubeCached(new CubeCollect(ref_cube)));

    // repaint GUI
    spectrum_pane->Refresh();
    met_tree->refresh();
    get_mid_tree()->refresh();
    loc_tree->refresh();
  //} 
}

// percentage view
void PresenterFrame::OnPercentageView  (wxCommandEvent& event) {
  view_menu->Check(ID_Percent_View,  true);
  view_menu->Check(ID_Relative_View, false);
  view_menu->Check(ID_Absolute_View, false);
  view_menu->Check(ID_PercentRef_View, false);
  if (gui_data == NULL) return;

  // if (gui_data->get_sev_mode() != PERCENT ) {
    gui_data->set_sev_mode(PERCENT);
    if (gui_data->empty()) return;
    gui_data->set_refcube(NULL);
    // repaint GUI
    spectrum_pane->Refresh();
    met_tree->refresh();
    get_mid_tree()->refresh();
    loc_tree->refresh();
  // } 
}

// Relative percentage view
void PresenterFrame::OnRelativeView  (wxCommandEvent& event) {
  view_menu->Check(ID_Relative_View, true);
  view_menu->Check(ID_Absolute_View, false);
  view_menu->Check(ID_PercentRef_View,  false);
  view_menu->Check(ID_Percent_View, false);
  if (gui_data == NULL) return;
  if (gui_data->get_sev_mode() != RELATIVE_PERC) {
    gui_data->set_sev_mode(RELATIVE_PERC);
    if (gui_data->empty()) return;
    spectrum_pane->Refresh();
    met_tree->refresh();
    get_mid_tree()->refresh();
    loc_tree->refresh();
  }
}

void PresenterFrame::OnPositiveOnly  (wxCommandEvent& event) {
  if (gui_data == NULL) return;
  gui_data->set_positive(!gui_data->get_positive());
  if (!gui_data->empty()) {
    met_tree->refresh();
    get_mid_tree()->refresh();
    loc_tree->refresh();
  }
}
void PresenterFrame::OnNegativeOnly  (wxCommandEvent& event) {
  if (gui_data == NULL) return;
  gui_data->set_negative(!gui_data->get_negative());
  if (!gui_data->empty()) {
    met_tree->refresh();
    get_mid_tree()->refresh();
    loc_tree->refresh();
  }
}

void PresenterFrame::OnSize(wxSizeEvent& WXUNUSED(event)) {
  int width;
  GetClientSize(&width, NULL);
  width /= 3;
  met_panel->SetDefaultSize(wxSize(width, 1000));
  mid_panel->SetDefaultSize(wxSize(width, 1000));
  loc_panel->SetDefaultSize(wxSize(width, 1000));

  wxLayoutAlgorithm layout;
  layout.LayoutWindow(this, loc_panel);
  /* adjust statusbar's field size */
  int width2[2];
  int w;
  GetClientSize(&w, NULL);
  width2[0] = w/12;
  width2[1] = w - width2[0];
  status_bar->SetFieldsCount(2, width2);  
}

void PresenterFrame::OnSashDrag(wxSashEvent& event) {
  if (event.GetDragStatus() == wxSASH_STATUS_OUT_OF_RANGE)
    return;
  
  switch (event.GetId()) {
  case ID_Metric_Panel:
    met_panel->SetDefaultSize(wxSize(event.GetDragRect().width, 1000));
    break;
  case ID_Cnode_Panel:
    mid_panel->SetDefaultSize(wxSize(event.GetDragRect().width, 1000));
    break;
  }

  wxLayoutAlgorithm layout;
  // layout.LayoutMDIFrame(this);
  layout.LayoutWindow(this, loc_panel);
  // GetClientWindow()->Refresh();
}

/* create a label with 3D style */
wxPanel* PresenterFrame::create_3Dlabel(wxWindow* parent, wxString label) {
  wxBoxSizer* sizer = new wxBoxSizer(wxVERTICAL);
  wxPanel* label_pane = new wxPanel(parent, -1, wxDefaultPosition,
				    wxDefaultSize, wxRAISED_BORDER);
  // a data member
  wxStaticText* text = new wxStaticText(label_pane,
					-1, label, wxPoint(0,0),
					wxDefaultSize, wxALIGN_LEFT);
  if (label.CompareTo("Call Tree") == 0) cnodelabel = text;
  label_pane->SetSize(100, 25); // default height
  label_pane->SetSizer(sizer);
  sizer->Add(text, 1, wxGROW|wxALL, 2);
  return label_pane;
}

wxSashLayoutWindow* 
PresenterFrame::create_metric_pane(int width, int ID) {
  wxSashLayoutWindow* win;
  win = new wxSashLayoutWindow(this, ID,
                               wxDefaultPosition, wxSize(200, 30),
                               wxSW_3D);
  win->SetDefaultSize(wxSize(width, 1000));
  win->SetOrientation(wxLAYOUT_VERTICAL);
  win->SetAlignment(wxLAYOUT_LEFT);
  win->SetSashVisible(wxSASH_RIGHT, TRUE);
  win->SetExtraBorderSize(PADDING);  
  // setup a wxPanel covering it.
  wxBoxSizer* topsizer = new wxBoxSizer(wxVERTICAL);
  wxPanel* pane = new wxPanel(win, -1);
  win->SetSizer(topsizer);
  wxBoxSizer* pane_sizer = new wxBoxSizer(wxVERTICAL);
  pane->SetSizer(pane_sizer);
  wxPanel* label_pane = create_3Dlabel(pane, _T("Performance Metrics"));
  pane_sizer->Add(label_pane, 0, wxGROW|wxALL, 0);
  // add metric tree
  met_tree = new MetricTree(pane);
  met_tree->SetBackgroundColour(GetBackgroundColour());
  pane_sizer->Add(met_tree,
		  1,
		  wxGROW|wxALL,
		  0);  
  topsizer->Add(pane, 1, wxGROW|wxALL, 0);
  return win;
}

wxSashLayoutWindow* 
PresenterFrame::create_prog_pane(int width, int ID) {
  wxSashLayoutWindow* win;
  win = new wxSashLayoutWindow(this, ID,
                               wxDefaultPosition, wxSize(200, 30),
                               wxSW_3D);
  win->SetDefaultSize(wxSize(width, 1000));
  win->SetOrientation(wxLAYOUT_VERTICAL);
  win->SetAlignment(wxLAYOUT_LEFT);
  win->SetSashVisible(wxSASH_RIGHT, TRUE);
  win->SetExtraBorderSize(PADDING);  
  // setup a wxPanel covering it.
  wxBoxSizer* topsizer = new wxBoxSizer(wxVERTICAL);
  middle_pane = new wxPanel(win, -1);
  win->SetSizer(topsizer);
  wxBoxSizer* pane_sizer = new wxBoxSizer(wxVERTICAL);
  middle_pane->SetSizer(pane_sizer);
  wxPanel* label_pane = create_3Dlabel(middle_pane, _T("Call Tree"));
  pane_sizer->Add(label_pane, 0, wxGROW|wxALL, 0);
  // create and add call node tree
  cnode_tree = new CnodeTree(middle_pane);
  cnode_tree->SetBackgroundColour(GetBackgroundColour());
  pane_sizer->Add(cnode_tree,
		  1,
		  wxGROW|wxALL,
		  0);  
  topsizer->Add(middle_pane, 1, wxGROW|wxALL, 0);
  return win;
}

wxSashLayoutWindow* 
PresenterFrame::create_sys_pane(int width, int ID) {
  wxSashLayoutWindow* win;
  win = new wxSashLayoutWindow(this, ID,
                               wxDefaultPosition, wxSize(200, 30),
                               wxSW_3D);
  win->SetDefaultSize(wxSize(width, 1000));
  win->SetOrientation(wxLAYOUT_VERTICAL);
  win->SetAlignment(wxLAYOUT_LEFT);
  win->SetExtraBorderSize(PADDING);  
  // setup a wxPanel covering it.
  wxBoxSizer* topsizer = new wxBoxSizer(wxVERTICAL);
  wxPanel* pane = new wxPanel(win, -1);
  win->SetSizer(topsizer);
  wxBoxSizer* pane_sizer = new wxBoxSizer(wxVERTICAL);
  pane->SetSizer(pane_sizer);
  wxPanel* label_pane = create_3Dlabel(pane, _T("System Tree"));
  pane_sizer->Add(label_pane, 0, wxGROW|wxALL, 0);
  // gap between label and tree
  // wxPanel* gap = new wxPanel(pane, -1);
  // gap->SetSize(100, 10);
  // pane_sizer->Add(gap, 0, wxGROW|wxALL, 0);
  // add location tree
  loc_tree = new LocationTree(pane);
  loc_tree->SetBackgroundColour(GetBackgroundColour());
  pane_sizer->Add(loc_tree,
		  1,
		  wxGROW|wxALL,
		  0);  
  topsizer->Add(pane, 1, wxGROW|wxALL, 0);
  return win;
}

//display call tree or float tree in the middle pane.
//called only by Menu event handler.
void PresenterFrame::switch_mid_tree(int old_type, int new_type) {
  if (gui_data == NULL || gui_data->empty()) return;
  if (middle_pane == NULL) return;
  wxSizer* pane_sizer;

  //1) remove old tree
  switch(old_type) {
  case MID_CALL_TREE:
    pane_sizer = const_cast<wxSizer*>(cnode_tree->GetContainingSizer());
    if (pane_sizer == NULL) {
      cerr << "Error: containing sizer NULL(call_tree's)." << endl;
      return;
    }
    pane_sizer->Remove(cnode_tree);
    cnode_tree->Hide();
    break;
  case MID_MODULE_TREE:
    pane_sizer = const_cast<wxSizer*>(module_tree->GetContainingSizer());
    if (pane_sizer == NULL) {
      cerr << "Error: containing sizer NULL(module_tree's)." << endl;
      return;
    }
    pane_sizer->Remove(module_tree);
    module_tree->Hide();
    break;
  case MID_REGION_TREE:
    pane_sizer = const_cast<wxSizer*>(region_tree->GetContainingSizer());
    if (pane_sizer == NULL) {
      cerr << "Error: containing sizer NULL(region_tree's)." << endl;
      return;
    }
    pane_sizer->Remove(region_tree);
    region_tree->Hide();
    break;
  default:
    cerr << "Warning: to remove an unknown type of middle tree" << endl;
  }
  
  //2) mount new tree
  switch(new_type) {
  case MID_CALL_TREE:
    if (cnode_tree == NULL) {
      cnode_tree = new CnodeTree(middle_pane);
      cnode_tree->SetBackgroundColour(GetBackgroundColour());
      pane_sizer->Add(cnode_tree, 1, wxGROW|wxALL, 0);
      build_cnode_tree();
    } else {
      pane_sizer->Add(cnode_tree, 1, wxGROW|wxALL, 0);
    }
    cnode_tree->refresh();
    cnode_tree->Show();
    break;
  case MID_MODULE_TREE:
    if (module_tree == NULL) {
      module_tree = new FlatTree(middle_pane);
      module_tree->SetBackgroundColour(GetBackgroundColour());
      pane_sizer->Add(module_tree, 1, wxGROW|wxALL, 0);
      build_module_tree();
    } else {
      pane_sizer->Add(module_tree, 1, wxGROW|wxALL, 0);
    }
    module_tree->refresh();
    module_tree->Show();
    break;
  case MID_REGION_TREE:
    if (region_tree == NULL) {
      region_tree = new FlatRegionTree(middle_pane);
      region_tree->SetBackgroundColour(GetBackgroundColour());
      pane_sizer->Add(region_tree, 1, wxGROW|wxALL, 0);
      build_region_tree();
    } else {
      pane_sizer->Add(region_tree, 1, wxGROW|wxALL, 0);
    }
    region_tree->refresh();
    region_tree->Show();
    break;
  default:
    cerr << "Warning: to mount an unknown type of middle tree" << endl;
  }
  
  pane_sizer->Layout();
}

/*
//remove&destroy a call-tree, then rebuild a flat-tree.
void PresenterFrame::switch_cnode2flat() {
  wxSizer* pane_sizer = 
    const_cast<wxSizer*>(cnode_tree->GetContainingSizer());
  if (pane_sizer == NULL) {
    cout << "Error: get containing sizer(->flat)." << endl;
    return;
  }
  
  pane_sizer->Remove(cnode_tree);
  cnode_tree->Hide();
  if (module_tree == NULL) {
    module_tree = new FlatTree(middle_pane);
    module_tree->SetBackgroundColour(GetBackgroundColour());
    pane_sizer->Add(module_tree, 1, wxGROW|wxALL, 0);
    build_module_tree();
  } else {
    pane_sizer->Add(module_tree, 1, wxGROW|wxALL, 0);
  }
  module_tree->refresh();
  module_tree->Show();
  
  pane_sizer->Layout();  

}

//remove&destroy a flat-tree, then rebuild a cnode_tree
void PresenterFrame::switch_flat2cnode() {
  wxSizer* pane_sizer = 
    const_cast<wxSizer*>(module_tree->GetContainingSizer());
  if (pane_sizer == NULL) {
    cout << "Error: get containing sizer(->cnode)." << endl;
    return;
  }
  
  pane_sizer->Remove(module_tree);
  module_tree->Hide();
  if (cnode_tree == NULL) {
    cnode_tree = new CnodeTree(middle_pane);
    cnode_tree->SetBackgroundColour(GetBackgroundColour());
    pane_sizer->Add(cnode_tree, 1, wxGROW|wxALL, 0);
    build_cnode_tree();
  } else {
    pane_sizer->Add(cnode_tree, 1, wxGROW|wxALL, 0);
  }
  cnode_tree->refresh();
  cnode_tree->Show();

  pane_sizer->Layout();  
}
*/

void PresenterFrame::OnPaint(wxPaintEvent& event) {

}

void PresenterFrame::init_statusbar() {
  wxString text;
  text.Printf("%d x %d", myApp->get_proc_num(), myApp->get_thrd_num());
  label_mn = new wxStaticText(status_bar, -1, text,
			      wxPoint(5,5), wxDefaultSize, wxALIGN_CENTRE);
  /* adjust statusbar's field size */
  int width[2];
  int w;
  GetClientSize(&w, NULL);
  width[0] = w/12;
  width[1] = w - width[0];
  status_bar->SetFieldsCount(2, width);  
}

void PresenterFrame::change_label_mn() {
  wxString text;
  text.Printf("%d x %d", myApp->get_proc_num(), myApp->get_thrd_num());
  label_mn->SetLabel(text);
}

void PresenterFrame::determine_ct_aggr_mode(Cube* cb) {
  if (cb->get_space()->get_prog_roots().size() == 1) {
    cb->def_attr("CUBE_CT_AGGR", "SUM");
    return;
  }

  if (!cb->get_attr("CUBE_CT_AGGR").empty()) return;

  // pop up a dialog to ask
  AggrModeDlg* dlg = new AggrModeDlg(this);
  dlg->Centre(wxBOTH);
  dlg->ShowModal();
  int n = dlg->get_selection(); // obtain user's choice
  if (n == 0) cb->def_attr("CUBE_CT_AGGR", "MAX");
  if (n == 1) cb->def_attr("CUBE_CT_AGGR", "SUM");
  delete dlg;
}

MidTree* PresenterFrame::get_mid_tree() const {
  int t = gui_data->get_midtree_type();
  switch(t) {
  case MID_CALL_TREE:
    return cnode_tree;
    break;
  case MID_MODULE_TREE:
    return module_tree;
    break;
  case MID_REGION_TREE:
    return region_tree; 
    break;
  default:
    cerr << "Error: unknow middle tree type" << endl;
    return NULL;
  }

}


bool PresenterFrame::try_all_mirrors(vector<string>& mirrors, string tail) {
  string url;
  vector<string>::iterator i;
  if (mirrors.size() == 0) return false;
  for (i = mirrors.begin(); i != mirrors.end(); i++) {
    url = *i;
    url = trunc_space(url);
    if (url.size() == 0 ) return false;
    if (url[url.length()-1] != '/' && tail[0] != '/') url += "/";
    url += tail;
    int pos = url.find_last_of('#');
    if (pos == string::npos) {
      cout << "Loading " << url << " ...  ";
      if (browser->LoadPage(url.c_str())) {
	cout << "success." << endl;
	break;
      } 
      else cout << "failure." << endl;
    } else {
      string anchor = url.substr(pos, url.length()-pos);
      string base = url.substr(0, pos);
      cout << "Loading " << url << " ...  ";
      if (browser->LoadPage(base.c_str())) {
	browser->Show(true);
	if (browser->LoadPage(anchor.c_str())) {
	  cout << "success." << endl;
	  break;
	}
      } 
      cout << "failure." << endl;
    }
  }
  
  if (i == mirrors.end()) {
    return false;
  } else {
    if (i != mirrors.begin()) {
      string tmp = *i;
      mirrors.erase(i);
      mirrors.insert(mirrors.begin(), tmp);
    }
    return true;
  }   
  
}

// OUTPUT: -1->error, 0->empty, 1->success.
int PresenterFrame::load_working_url(Browser* browser, string url) {
  string protocol;

  if (gui_data == NULL || gui_data->empty() || browser == NULL) return 0;
  if (url.empty()) { 
    wxMessageBox("URL empty!", _T("Warning"), wxICON_INFORMATION, this);
    return 0;
  }

  Cube* cu = const_cast<Cube*>(gui_data->get_cube());
  vector<string>& mirrors =  cu->get_mirrors();
  string target = trunc_space(url);
 
  if (target.length() >= 4) {
    protocol = target.substr(0, 4);
    if (protocol == "http" || protocol == "file") {
      if (browser->LoadPage(target.c_str())) return 1;
      else return -1;
    }
  }
  string pattern = "@mirror@"; 
  if (target.length() >= pattern.length()) {
    protocol = target.substr(0, pattern.length());
    if (protocol == pattern) {
      string tail = target.substr(pattern.length(), target.length()-pattern.length());
      if (try_all_mirrors(mirrors, tail)) return 1;
      vector<string> local_mirror;
      local_mirror.push_back(DOCDIR);
      if (try_all_mirrors(local_mirror, tail)) return 1;
    }
  }

  return -1;

}


namespace cube {
  string get_base(const string name) {
    int pos = name.find_last_of(wxFILE_SEP_PATH);
    if (pos != string::npos)
      return name.substr(pos+1, name.length()); 
    else 
      return name;
  }
  string get_path(const string name) {
    int pos = name.find_last_of(wxFILE_SEP_PATH);
    if (pos != string::npos) {
      return name.substr(0, pos);
    } 
    else 
      return "";
  }
  // spaces on both ends are removed
  string trunc_space(std::string str) {
    int s = str.find_first_not_of(' ');
    if (s == string::npos) return "";
    int e = str.find_last_not_of(' ');
    return str.substr(s, e-s+1);
  }


  /*
  std::string locate_cube() {
    char *path, *dir;    
    path = getenv("PATH");
    if (path != NULL) {
      path = strdup(path);
      for (dir = strtok(path, ":"); dir != NULL; dir = strtok(NULL, ":")) {
	string candidate = dir;
	candidate += wxFILE_SEP_PATH;
	string tmp = candidate + "cube-config";
	// cout << "trying " << tmp << endl;
	if (access(tmp.c_str(), X_OK) == 0) {
	  // cout << "executable is " << tmp << endl;
	  return candidate;
	}
      }
      free(path);
    }
    return "";
  }

  // Each path is converted to a url of file://
  void find_files(string docpath, string pattern, vector<string>& ret) {
    wxFileSystem fs;
    wxString fn;
    wxFileName wxfn;
    
    fs.ChangePathTo(docpath.c_str(), true);
    fn = fs.FindFirst(pattern.c_str(), wxFILE);
    while (!fn.IsEmpty()) {
      wxfn = wxFileName(docpath.c_str(), fn);
      ret.push_back(wxFileSystem::FileNameToURL(wxfn).c_str());
      fn = fs.FindNext();
    }
  }
  */  
 
}


