/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


/***************************************
					
     PresenterFrame class
          
***************************************/


#ifndef CUBE_PRESENTER_FRAME_H
#define CUBE_PRESENTER_FRAME_H

#include "wx/wxprec.h"
#include "wx/wx.h"
#include "wx/mdi.h"
#include "wx/toolbar.h"
#include "wx/laywin.h"
#include "wx/sizer.h"
#include "wx/listctrl.h"

#include <string>
#include <vector>

#define PRECISION         1
#define ID_Metric_Panel   1002
#define ID_Cnode_Panel    1003
#define ID_Location_Panel 1004

// #define CUBE_DEBUG

namespace cube {

  enum {
    // File Menu
    ID_Open_File = 1,
    ID_Close_File,
    // View Menu
    ID_Calltree_View,
    ID_ModuleTree_View,
    ID_RegionTree_View,
    ID_Absolute_View, // absolute severity view
    ID_Percent_View,  // absolute percentage view or 
    ID_PercentRef_View, // to show a percentage relative to a reference cube file.
    ID_Relative_View, // relative percentage view
    ID_HIDE_SUBMENU,
    ID_Positive_Only,
    ID_Negative_Only,
    // Others
    ID_Exit,
    ID_About
  };

  class SourceCodeDlg;
  class Browser;
  class Cube;
  class CubeCached;
  class GUIData;
  class SpectrumPane;
  class MetricTree;
  class CnodeTree;
  class MidTree;
  class FlatTree;
  class FlatRegionTree;
  class LocationTree;
  class PresenterApp;

  class PresenterFrame: public wxMDIParentFrame {    
  public:
    
    PresenterFrame (wxWindow *parent, const wxWindowID id, const wxString& title,
		    const wxPoint& pos, const wxSize& size, const long style);
    ~PresenterFrame();
    
    SpectrumPane* get_spectrum_pane() { return spectrum_pane; }
    
    void build_met_tree();
    void build_cnode_tree();
    void build_loc_tree();    
    void build_module_tree();
    void build_region_tree();

    SourceCodeDlg* get_source_dlg()        { return dlg; }
    void set_source_dlg(SourceCodeDlg* d)  { dlg = d; }
    Browser* get_browser_dlg()             { return browser; }
    void set_browser_dlg(Browser* bw)      { browser = bw; }

    /* Menu Event Handlers */
    void OnOpenFile      (wxCommandEvent& event);
    void OnCloseFile     (wxCommandEvent& event);
    void OnExit          (wxCommandEvent& event);
    void OnAbout         (wxCommandEvent& event);
    void OnModuleTreeView(wxCommandEvent& event);
    void OnRegionTreeView(wxCommandEvent& event);
    void OnCalltreeView  (wxCommandEvent& event);
    void OnAbsoluteView  (wxCommandEvent& event);
    void OnRelativeView  (wxCommandEvent& event);
    void OnPercentRefView(wxCommandEvent& event);
    void OnPercentageView(wxCommandEvent& event);
    void OnPositiveOnly  (wxCommandEvent& event);
    void OnNegativeOnly  (wxCommandEvent& event);
    
    /* GUI Event Handlers */
    void OnSize     (wxSizeEvent& event);
    void OnSashDrag (wxSashEvent& event);
    void OnPaint    (wxPaintEvent& event);

    void change_label_mn();    // repaint mxn label on status bar.    
    void init_statusbar ();

    void load_file(const char* path);

    GUIData* get_guidata() { return gui_data; }
    void     set_cube(Cube* cb);

    LocationTree*   get_loc_tree()   const    { return loc_tree; }
    MetricTree*     get_met_tree()   const    { return met_tree; }
    MidTree*        get_mid_tree()   const;
    // CnodeTree*      get_cnode_tree()   const  { return cnode_tree; }
    // FlatTree*       get_module_tree()  const  { return module_tree; }
    // FlatRegionTree* get_region_tree()  const  { return region_tree;}

    int             get_num_procs()  const;
    int             get_max_thrdnum()const;  
    
    int load_working_url (Browser* browser, std::string url);
    
    DECLARE_EVENT_TABLE()
      
    private:

    // private  methods
    wxPanel*            create_3Dlabel (wxWindow* parent, wxString label);
    wxSashLayoutWindow* create_metric_pane (int default_width, int ID);
    wxSashLayoutWindow* create_prog_pane (int default_width, int ID);
    wxSashLayoutWindow* create_sys_pane (int default_width, int ID);

    // users switch views between call-tree and flat-tree.
    void switch_mid_tree(int old_type, int new_type);

    void determine_ct_aggr_mode(Cube* cb);

    bool try_all_mirrors  (std::vector<std::string>& mirrors, std::string tail);

    SourceCodeDlg* dlg;
    Browser*       browser;
    wxBitmap*      logo; // ICL logo .xpm
    
    /* data memebers */
    wxMenu *view_menu, *hide_menu;
    SpectrumPane       *spectrum_pane;
    wxSashLayoutWindow *met_panel, *mid_panel, *loc_panel;
    wxPanel                        *middle_pane;
    wxStaticText*       cnodelabel; // the title in the middle pane

    MetricTree*         met_tree;
    CnodeTree*          cnode_tree;
    FlatTree*           module_tree;
    FlatRegionTree*     region_tree;
    LocationTree*       loc_tree;

    wxStatusBar*        status_bar;

    GUIData* gui_data;

    PresenterApp* myApp;
    wxStaticText* label_mn; // mxn label on the left status bar
    static const int PADDING = 0;
  };
  
  std::string get_base(const std::string name);
  std::string get_path(const std::string name);

  std::string trunc_space(std::string str);
  std::string locate_cube(); // cube home directory
  void find_files(std::string docpath, std::string pattern,
		  std::vector<std::string>& files);
}

#endif
