/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/************************************************

             GUIData.cpp


 ************************************************/

#include <iostream>
#include <sstream>
#include <math.h>

#include "CubeCached.h"
#include "CubeCollect.h"
#include "Cube.h"
#include "Space.h"
#include "GUIData.h"
#include "Metric.h"
#include "Cnode.h"
#include "Module.h"
#include "Region.h"
#include "Location.h"

using namespace std;
using namespace cube;

GUIData::GUIData (CubeCached* cubec) {
  this->cube = cubec;
  ref_cube = NULL;
  midtree_type = MID_CALL_TREE;
  sev_mode = ABSOLUTE;
  pos_only = neg_only = true;
  sel_mid_type = -1;
  sel_left = NULL;
  sel_ctree_mid = NULL;
  sel_flat_mid = NULL;
  sel_left_mode = sel_ctree_mid_mode = sel_flat_mid_mode = CUBE_INCL;
}

void GUIData::reset() {
  set_cube(NULL);
  set_refcube(NULL);
  midtree_type = MID_CALL_TREE;
  sev_mode = ABSOLUTE;
  pos_only = neg_only = true;
  sel_mid_type = -1;
  sel_left = NULL;
  sel_ctree_mid = NULL;
  sel_flat_mid = NULL;
  sel_left_mode = sel_ctree_mid_mode = sel_flat_mid_mode = CUBE_INCL;
}

void GUIData::set_cube(CubeCached* cubec) {
  if (this->cube != NULL) delete this->cube;
  this->cube = cubec;
  ref_cube = NULL;
  sel_mid_type = -1;
  sel_left = NULL;
  sel_ctree_mid = NULL;
  sel_flat_mid = NULL;
  sel_left_mode = sel_ctree_mid_mode = sel_flat_mid_mode = CUBE_INCL;
}

void GUIData::set_refcube(CubeCached* ref) {
  if (ref_cube != NULL) delete ref_cube;
  ref_cube = ref;
}

void GUIData::set_midtree_type(int type) {
  midtree_type = type;
}

int  GUIData::get_midtree_type() const {
  return midtree_type;
}

void GUIData::set_sev_mode(int m) {
  sev_mode = m;
}

void GUIData::set_positive(bool pos) {
  pos_only = pos;
}

void GUIData::set_negative(bool neg) {
  neg_only = neg;
}

void GUIData::set_ref_cube(CubeCached* cubec) {
  ref_cube = cubec;
}
   
void GUIData::select_metric(const Metric* met) {
  sel_left = const_cast<Metric*>(met);
}

void GUIData::select_cnode(const Cnode*  cnode) {
  if(midtree_type != MID_CALL_TREE) {
    // cout << "Warning: select cnode in flat tree!" << endl;
  }
  sel_ctree_mid = const_cast<Cnode*>(cnode);
}

void GUIData::select_module(const Module* mod) {
  if(midtree_type != MID_MODULE_TREE) {
    // cout << "Warning: select module in call-path tree!" << endl;
  }
  sel_flat_mid = const_cast<Module*>(mod);
  sel_mid_type = MID_MODULE;
}

void GUIData::select_region(const Region* reg) {
  if(midtree_type == MID_CALL_TREE) {
    // cout << "Warning: select region in call-path tree!" << endl;
  }
  sel_flat_mid = const_cast<Region*>(reg);
  sel_mid_type = MID_REGION; 
}

void GUIData::select_subregion(const Region* reg) {
  //if(midtree_type == MID_CALL_TREE) 
  //  cout << "Warning: select subregions in call-path tree!" << endl;  
  sel_flat_mid = const_cast<Region*>(reg);
  sel_mid_type = MID_SUBREGION;
  sel_flat_mid_mode = CUBE_EXCL;
}
    
void GUIData::xpnd_sel_left() {
  sel_left_mode = CUBE_EXCL;
}

void GUIData::xpnd_sel_middle() {
  if(midtree_type == MID_CALL_TREE)
    sel_ctree_mid_mode = CUBE_EXCL;
  else
    sel_flat_mid_mode = CUBE_EXCL;
}

void GUIData::clps_sel_left() {
  sel_left_mode = CUBE_INCL;
}

void GUIData::clps_sel_middle() {
  if(midtree_type == MID_CALL_TREE)
    sel_ctree_mid_mode = CUBE_INCL;
  else
    sel_flat_mid_mode = CUBE_INCL;
}

/****************************************************************         
                 Rules to display labels:
  
Applies only to the absolute mode:
Floating Point
  (1) Take the maximum
  (2) If 0.10 <= maximum < 100 don't do anything
  (3) Write it in scientific notation: a.bc E XY 
  (4) Scale all values in the tree with respect to E XY
  (5) Wite a.b E XY on the scale
Integer
  (1) Take the maximum
  (2) If 0 <= maximum < 999 don't do anything
  (3) Write it in scientific notation: abc E XY
  (4) Scale all values in the tree with respect to E XY
  (5) Wite abc E XY on the scale
- (realtive) percentage: one digit after decimal point
- absolute non-scientific decimal: one digit after decimal point
- absolute non-scientific integer: no decimal point
- absolute scientific: three significant digits 
******************************************************************/

// for max >= 1
int get_exponent(double max) {
  if (max < 0) max = -max;
  double val = floor(max);
  int k = 0;
  while (val > 0) {
    val /= 10;
    val = floor(val); // only integer part wanted
    k++;
  } 
  return k;
}   

// for 0<max<1
int get_exponent2(double max) {
  if (max < 0) max = -max;
  double val = max;
  int k = 0;
  while (val != 0 && val < 1) {
    val *= 10;
    k++;
  } 
  return k;
}   

double GUIData::get_max(const Metric* use_sel_met)  const {
  if (ref_cube == NULL || sev_mode != PERCENT) {
    if (use_sel_met == NULL) {
      return cube->get_max(sel_left);
    } else 
      return cube->get_max(use_sel_met);
    // to get a label for a metric which is not in the same tree
    // as of the selected metric.
  } else {
    const Space* space = ref_cube->get_cube()->get_cube()->get_space();
    if (use_sel_met == NULL) {
      string name = sel_left->get_name();
      const Metric* m = space->find_met(name);
      if (m == NULL) return 0;
      return ref_cube->get_max(m);
    } else {
      string name = use_sel_met->get_name();
      const Metric* m = space->find_met(name);
      if (m == NULL) return 0;
      return ref_cube->get_max(m);
    }
  }
}

double GUIData::sev2disp(double sev, double maxsev, double refsev) const{
  if(maxsev <= 1.0e-5) return 0; // ignored val
  if(sev_mode == ABSOLUTE) {
    return sev;
  }
  else if(sev_mode == PERCENT){
    return sev/maxsev*100.0;
  } else {
    if (fabs(refsev) <= 1.0e-5) refsev = maxsev;
    return sev/refsev*100.0;
  }
}


// used to create any entity's label in absolute mode(complex!)
string GUIData::gen_abs_label(double sev, string name, string uom, double maxsev) const {
  ostringstream out;
  bool isInteger = false;
  int exponent = get_exponent(maxsev);
  if (uom.compare("occ") == 0) isInteger = true;

  // cout << "sev: " << sev << "maxsev: " << maxsev << endl;
      
  if (!isInteger) {
    if (maxsev >= 100.0) {
      out.precision(2);     // a.bc E xx
      out.setf(ios::fixed); // omit the E xx part 
      for (int m = 1; m < exponent; m++) sev /= 10.0;
      out << sev << " " << name << " ";
      return out.str();
    } else if (maxsev < 0.1) { // scale up for very smalle values
      out.precision(2);     // a.bc E xx
      out.setf(ios::fixed); // omit the E xx part 
      exponent = get_exponent2(maxsev);
      for (int m = 0; m < exponent; m++) sev *= 10.0;
      out << sev << " " << name << " ";
      return out.str();
    }
  }
  if (isInteger && maxsev >= 1000) {
    out.precision(2);     // a.bc E xx
    out.setf(ios::fixed); // omit the E xx part     
    for (int m = 1; m < exponent; m++) sev /= 10.0;
    out << sev << " " << name << " ";
    return out.str();
  } 
  if (isInteger) {
    int z = (int) sev;
    out << z << " " << name << " ";
    return out.str();
  }
  // if sev not integer and maxsev<100.0, do as follows
  out.precision(2);
  out.setf(ios::fixed);
  out << sev << " " << name << " ";
  return out.str();

}

string GUIData::get_met_label  (int mode, const Metric* met) const {
  double disp_v;
  double max_sev;
  double sev;  
  ostringstream out;

  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};

  modes[CUBE_MET] = mode;
  sev = cube->get_vcsev(modes, met);
  max_sev = get_max(met);
  
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, met->get_name(), met->get_uom(), max_sev);
  }
  
  disp_v = sev2disp(sev, max_sev, 0);
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << met->get_name() << " ";
  return out.str();
}

string GUIData::get_cnode_label(int mode, const Cnode*  cnode) const {
  double disp_v;
  double max_sev;
  double sev, refsev;  
  ostringstream out;
  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = mode;
  sev = cube->get_vcsev(modes, sel_left, cnode);
  max_sev = get_max(sel_left);

  string name = cnode->get_csite()->get_callee()->get_name();
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, name, sel_left->get_uom(), max_sev);
  }
  
  refsev = cube->get_vcsev(modes, sel_left);
  disp_v = sev2disp(sev, max_sev, refsev);
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << name << " ";
  return out.str();
}

string GUIData::get_loc_label  (int mode, const Location* loc)  const {
  if (midtree_type == MID_CALL_TREE) return get_loc4cnode_label(mode, loc);
  else {
    switch(sel_mid_type) {
    case MID_MODULE:
      return get_loc4module_label(mode, loc);
    case MID_REGION:
      return get_loc4reg_label(mode, loc);
    case MID_SUBREGION:
      return get_loc4sreg_label(mode, loc);
    }
    cerr << "Warning: unknown selected middle node type:" << sel_mid_type << endl;
    return "N/A";
  }
}

string GUIData::get_loc4cnode_label  (int mode, const Location* loc)  const {
  double disp_v;
  double max_sev;
  double sev, refsev;  
  ostringstream out;
  Cnode* cnode;
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};

  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = sel_ctree_mid_mode;
  modes[CUBE_LOC] = mode;
  cnode = sel_ctree_mid;
  sev = cube->get_vcsev(modes, sel_left, cnode, loc);
  max_sev = get_max(sel_left);

  string name = loc->get_name();
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, name, sel_left->get_uom(), max_sev);
  }
  
  refsev = cube->get_vcsev(modes, sel_left, cnode);
  disp_v = sev2disp(sev, max_sev, refsev);
 
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << name << " ";
  return out.str();
}

string GUIData::get_loc4module_label  (int mode, const Location* loc)  const {
  double disp_v;
  double max_sev;
  double sev, refsev;  
  ostringstream out;
  Module* mod;
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};

  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = sel_flat_mid_mode;
  modes[CUBE_LOC] = mode;
  mod = static_cast<Module*>(sel_flat_mid);
  sev = cube->get_vmsev(modes, sel_left, mod, loc);
  max_sev = get_max(sel_left);

  string name = loc->get_name();
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, name, sel_left->get_uom(), max_sev);
  }
  
  refsev = cube->get_vmsev(modes, sel_left, mod);
  disp_v = sev2disp(sev, max_sev, refsev);
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << name << " ";
  return out.str();
}

string GUIData::get_loc4reg_label  (int mode, const Location* loc)  const {
  double disp_v;
  double max_sev;
  double sev, refsev;  
  ostringstream out;
  Region* reg;
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};

  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = sel_flat_mid_mode;
  modes[CUBE_LOC] = mode;
  reg = static_cast<Region*>(sel_flat_mid);
  sev = cube->get_vrsev(modes, sel_left, reg, loc);
  max_sev = get_max(sel_left);

  string name = loc->get_name();
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, name, sel_left->get_uom(), max_sev);
  }
  
  refsev = cube->get_vrsev(modes, sel_left, reg);
  disp_v = sev2disp(sev, max_sev, refsev);
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << name << " ";
  return out.str();
}

string GUIData::get_loc4sreg_label  (int mode, const Location* loc)  const {
  double disp_v;
  double max_sev;
  double sev, refsev;  
  ostringstream out;
  Region* reg;
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};

  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_LOC] = mode;
  reg = static_cast<Region*>(sel_flat_mid);
  sev = cube->get_vrsev(modes, sel_left, reg, loc);
  refsev = cube->get_vrsev(modes, sel_left, reg);
  modes[CUBE_PROG] = CUBE_EXCL;
  sev -= cube->get_vrsev(modes, sel_left, reg, loc);
  refsev -= cube->get_vrsev(modes, sel_left, reg);
  max_sev = get_max(sel_left);

  string name = loc->get_name();
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, name, sel_left->get_uom(), max_sev);
  }
  
  disp_v = sev2disp(sev, max_sev, refsev);
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << name << " ";
  return out.str();
}

string GUIData::get_mod_label  (int mode, const Module* mod) const {
  double disp_v;
  double max_sev;
  double sev, refsev;  
  ostringstream out;
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = mode;

  sev = cube->get_vmsev(modes, sel_left, mod);
  max_sev = get_max(sel_left);

  string name = mod->get_name();
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, name, sel_left->get_uom(), max_sev);
  }
  refsev = cube->get_vcsev(modes, sel_left);
  disp_v = sev2disp(sev, max_sev, refsev);
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << name << " ";
  return out.str();
  
}

string GUIData::get_reg_label  (int mode, const Region* reg) const {
  double disp_v;
  double max_sev;
  double sev, refsev;  
  ostringstream out;
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = mode;
  
  sev = cube->get_vrsev(modes, sel_left, reg);
  max_sev = get_max(sel_left);

  string name = reg->get_name();
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, name, sel_left->get_uom(), max_sev);
  }
  refsev = cube->get_vcsev(modes, sel_left);
  disp_v = sev2disp(sev, max_sev, refsev);
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << name << " ";
  return out.str();
}

string GUIData::get_sreg_label (const Region* reg) const {
  double disp_v;
  double max_sev;
  double sev, refsev;  
  ostringstream out;
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  
  modes[CUBE_MET] = sel_left_mode;
  sev = cube->get_vrsev(modes, sel_left, reg);
  modes[CUBE_PROG] = CUBE_EXCL;
  sev -= cube->get_vrsev(modes, sel_left, reg);
  max_sev = get_max(sel_left);

  string name = "Subroutines";
  if(sev_mode == ABSOLUTE) {
    return gen_abs_label(sev, name, sel_left->get_uom(), max_sev);
  }
  refsev = cube->get_vcsev(modes, sel_left);
  disp_v = sev2disp(sev, max_sev, refsev);
  out.precision(1);
  out.setf(ios::fixed);
  out << disp_v << " " << name << " ";
  return out.str();
}

// maximum of focused metric tree
string GUIData::get_max_label() const {
  double maxsev = get_max(sel_left);
  // if (maxsev == 0) return "n/a";
  return get_max_label(maxsev);
}
string GUIData::get_halfmax_label() const {
  double maxsev = get_max(sel_left)/2.0;
  // if (maxsev == 0) return "n/a";
  return get_max_label(maxsev);
}
// maximum of focused metric tree
string GUIData::get_max_label(double maxsev) const {
  ostringstream out;
  bool isInteger = false;
  string uom;

  if (sel_left == NULL) return "n/a"; // error protection.
  if (maxsev == 0) return "0.00";

  uom = sel_left->get_uom();
  
  if (uom.compare("occ") == 0) isInteger = true;

  if (!isInteger) {
    if (maxsev >= 100.0 || maxsev < 0.1) {
      out.precision(2); // a.bc E xx
      out.setf(ios::scientific);      
      out << maxsev << " ";
      return out.str();
    } 
  }
  if (isInteger && maxsev >= 1000) {
    out.precision(2); // a.bc E xx
    out.setf(ios::scientific); 
    out << maxsev << " ";
    return out.str();
  } 
  if (isInteger) {
    int z = (int) maxsev;
    out << z  << " ";
    return out.str();
  }
  out.precision(2);
  out.setf(ios::fixed); 
  out << maxsev << " ";
  return out.str();
}

int GUIData::color_index(double per) const {
  bool negative = false;
  const int NegStart = 103;

  if (per > 0) {
    if (!pos_only) return 0;
  } else if (per < 0) {
    if (!neg_only) return 0;
  } else return 0; // exact "zero"

  if (per < 0) {
    negative = true;
    per = -per; // always positive now.
  }
  
  if (per > 0 && per < 0.5) {
    if (!negative) return 102; // dark grey color
    return NegStart + 102; //sunken square
  }
  if (per >= 0.5 && per < 1.0) {
    if (!negative) return 1;
    return NegStart + 1;
  }

  int i = (int) per;
  if (i > 102) {
    i = 101;
    cerr<< "Warning: Color out of range" << endl;
  }
  if (!negative) return i;
  return NegStart + i;
}

double GUIData::sev2perc(double sev, double maxsev, double refsev) const{
  if (maxsev == 0) return 0;
  if (maxsev <= 1.0e-5) return 0.1; // ignore very small perc
  if (sev_mode == PERCENT || sev_mode == ABSOLUTE){
    return sev/maxsev*100.0;
  }
  if (fabs(refsev) <= 1.0e-5) refsev = maxsev;
  return sev/fabs(refsev)*100.0;
}

int GUIData::get_met_color(int mode, const Metric* met) const {
  double max_sev;
  double sev, perc;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};

  modes[CUBE_MET] = mode;
  sev = cube->get_vcsev(modes, met);
  max_sev = get_max(met);

  perc = sev2perc(sev, max_sev, 0);
  return color_index(perc);
}

int GUIData::get_cnode_color(int mode, const Cnode*  cnode) const {
  double max_sev;
  double sev, perc, refsev;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = mode;
  sev = cube->get_vcsev(modes, sel_left, cnode);
  refsev = cube->get_vcsev(modes, sel_left);
  max_sev = get_max(sel_left);
  
  perc = sev2perc(sev, max_sev, refsev);
  return color_index(perc);
}

int GUIData::get_mod_color(int mode, const Module* mod) const {
  double max_sev;
  double sev, perc, refsev;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = mode;
  sev = cube->get_vmsev(modes, sel_left, mod);
  refsev = cube->get_vcsev(modes, sel_left);
  max_sev = get_max(sel_left);
  
  perc = sev2perc(sev, max_sev, refsev);
  return color_index(perc);
}

int GUIData::get_reg_color(int mode, const Region* reg) const {
  double max_sev;
  double sev, perc, refsev;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = mode;
  sev = cube->get_vrsev(modes, sel_left, reg);
  refsev = cube->get_vcsev(modes, sel_left);
  max_sev = get_max(sel_left);
  
  perc = sev2perc(sev, max_sev, refsev);
  return color_index(perc);
}

int GUIData::get_sreg_color(const Region* reg) const {
  double max_sev;
  double sev, perc, refsev;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  
  modes[CUBE_MET] = sel_left_mode;
  sev = cube->get_vrsev(modes, sel_left, reg);
  modes[CUBE_PROG] = CUBE_EXCL;
  sev -= cube->get_vrsev(modes, sel_left, reg);
  refsev = cube->get_vcsev(modes, sel_left);
  max_sev = get_max(sel_left);
  
  perc = sev2perc(sev, max_sev, refsev);
  return color_index(perc);
}

int GUIData::get_loc_color(int mode, const Location* loc) const {
  if(midtree_type == MID_CALL_TREE) return get_loc4cnode_color(mode, loc);
  else {
    switch(sel_mid_type) {
    case MID_MODULE:
      return get_loc4module_color(mode, loc);
    case MID_REGION:
      return get_loc4reg_color(mode, loc);
    case MID_SUBREGION:
      return get_loc4sreg_color(mode, loc);
    }
    cerr << "Warning: unknown selected middle node type" << endl;
    return 0;
  }
}

int GUIData::get_loc4cnode_color(int mode, const Location* loc) const {
  double max_sev;
  double sev, perc, refsev;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  Cnode* cnode;
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = sel_ctree_mid_mode;
  modes[CUBE_LOC] = mode;
  cnode = sel_ctree_mid;

  sev = cube->get_vcsev(modes, sel_left, cnode, loc);
  max_sev = get_max(sel_left);
  refsev = cube->get_vcsev(modes, sel_left, cnode);
  
  perc = sev2perc(sev, max_sev, refsev);
  return color_index(perc);
}

int GUIData::get_loc4module_color(int mode, const Location* loc) const {
  double max_sev;
  double sev, perc, refsev;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  Module* mod;
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = sel_flat_mid_mode;
  modes[CUBE_LOC] = mode;
  mod = static_cast<Module*>(sel_flat_mid);

  sev = cube->get_vmsev(modes, sel_left, mod, loc);
  max_sev = get_max(sel_left);
  refsev = cube->get_vmsev(modes, sel_left, mod);
  
  perc = sev2perc(sev, max_sev, refsev);
  return color_index(perc);
}

int GUIData::get_loc4reg_color(int mode, const Location* loc) const {
  double max_sev;
  double sev, perc, refsev;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  Region* reg;
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_PROG] = sel_flat_mid_mode;
  modes[CUBE_LOC] = mode;
  reg = static_cast<Region*>(sel_flat_mid);

  sev = cube->get_vrsev(modes, sel_left, reg, loc);
  max_sev = get_max(sel_left);
  refsev = cube->get_vrsev(modes, sel_left, reg);
  
  perc = sev2perc(sev, max_sev, refsev);
  return color_index(perc);
}

int GUIData::get_loc4sreg_color(int mode, const Location* loc) const {
  double max_sev;
  double sev, perc, refsev;  
  int modes[3] = {CUBE_INCL, CUBE_INCL, CUBE_INCL};
  Region* reg;
  
  modes[CUBE_MET] = sel_left_mode;
  modes[CUBE_LOC] = mode;
  reg = static_cast<Region*>(sel_flat_mid);

  sev = cube->get_vrsev(modes, sel_left, reg, loc);
  refsev = cube->get_vrsev(modes, sel_left, reg);
  modes[CUBE_PROG] = CUBE_EXCL;
  sev -= cube->get_vrsev(modes, sel_left, reg, loc);
  refsev -= cube->get_vrsev(modes, sel_left, reg);
  max_sev = get_max(sel_left);
  
  perc = sev2perc(sev, max_sev, refsev);
  return color_index(perc);
}

const vector<Metric*>& GUIData::get_met_roots() {
  const Cube* cu = cube->get_cube()->get_cube();
  return cu->get_space()->get_met_roots();
}

const vector<Module*>& GUIData::get_mod_roots() {
  const Cube* cu = cube->get_cube()->get_cube();
  return cu->get_space()->get_modv();
}

const vector<Cnode*>&  GUIData::get_cnode_roots() {
  const Cube* cu = cube->get_cube()->get_cube();
  return cu->get_space()->get_prog_roots();
}

const vector<Machine*>& GUIData::get_loc_roots() {
  const Cube* cu = cube->get_cube()->get_cube();
  return cu->get_space()->get_sys_roots();
}

const vector<Region*>& GUIData::get_regionv() {
  const Cube* cu = cube->get_cube()->get_cube();
  return cu->get_space()->get_regnv();
}

string GUIData::get_attr(string key){
  Cube* cu = cube->get_cube()->get_cube();
  return cu->get_attr(key);
}

int GUIData::num_procs() {
  const Cube* cu = cube->get_cube()->get_cube();
  return cu->get_space()->num_procs();
}

int GUIData::max_num_thrd() {
  const Cube* cu = cube->get_cube()->get_cube();
  return cu->get_space()->max_num_thrd();
}

bool GUIData::flatree_only_data() {
  if (empty()) return false;
  const Cube* cu = cube->get_cube()->get_cube();
  const Space* space = cu->get_space();
  return (space->get_cnodev().size() == space->get_prog_roots().size());
}

const Cube* GUIData::get_cube() {
  if (empty()) return NULL;
  return cube->get_cube()->get_cube();
}
 
// debug usage: mode to string
string m2str(int i) {
  if (i == CUBE_INCL) return "incl";
  else return "excl";
}

// debug usage
void GUIData::print_info() {
  Metric* m = sel_left;
  Cnode* cn;
  Module* mod;
  Region* r;
  cout << "met('" << m->get_name() << "'," << m2str(sel_left_mode) << ") ";
  if(midtree_type == MID_CALL_TREE) {
    cn = sel_ctree_mid;
    cout << "cnode('" << cn->get_csite()->get_callee()->get_name() << "'," << 
      m2str(sel_ctree_mid_mode) <<")";
  } else
    switch(sel_mid_type) {
    case MID_MODULE:
      mod = static_cast<Module*>(sel_flat_mid);
      cout << "module('" << mod->get_name() << "'," << m2str(sel_flat_mid_mode) <<")";
      break;
    case MID_REGION:
      r = static_cast<Region*>(sel_flat_mid);
      cout << "region('" << r->get_name() << "'," << m2str(sel_flat_mid_mode) << ")";
      break;
    case MID_SUBREGION:
      cout <<"subregions";
    }

  cout << endl;
}
