/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


#ifndef CUBE_GUIDATA_H
#define CUBE_GUIDATA_H   

/*
 *----------------------------------------------------------------------------
 *
 * class GUIData
 *  
 * Represents the GUI's states together with cube data set
 *
 *----------------------------------------------------------------------------
 */


#include <string>
#include <vector>

#define ABSOLUTE         0
#define PERCENT          1
#define RELATIVE_PERC    2 


#define MID_MODULE       1
#define MID_REGION       2
#define MID_SUBREGION    3

// tree type displayed in the middle panel
#define MID_CALL_TREE    1
#define MID_MODULE_TREE  2
#define MID_REGION_TREE  3

namespace cube {
  class CubeCached;
  class Metric;
  class Cnode;
  class Module;
  class Region;
  class Location;
  class Vertex;
  class Machine;

  class GUIData {
    
  public:
    
    GUIData (CubeCached* cube);
    ~GUIData( ) {}

    const Cube* get_cube();     // only used by frame->load_working_url()

    void set_cube(CubeCached*); // replaced with a new cube file

    CubeCached* get_refcube() { return ref_cube; }
    void set_refcube(CubeCached*);

    bool empty()                { return cube == NULL; } 
    void reset();               // reset all state info.
    bool flatree_only_data();   // Is user data flat-tree only type.
    
    void select_metric          (const Metric* met);
    void select_cnode           (const Cnode*  cnode);
    void select_module          (const Module* mod);
    void select_region          (const Region* reg);
    void select_subregion       (const Region* reg);

    void print_info(); // debug useage

    void xpnd_sel_left          ();
    void xpnd_sel_middle        ();
    void clps_sel_left          ();
    void clps_sel_middle        ();

    void set_midtree_type(int type);       // shows calltree or function tree
    void set_sev_mode(int m);        // shows severity in which view mode
    void set_positive(bool pos);     // displays positive nodes only
    void set_negative(bool neg);     // displays negative nodes only

    int  get_midtree_type() const ;
    int  get_sev_mode()   const { return sev_mode; }        
    bool get_positive()   const { return pos_only; }   
    bool get_negative()   const { return neg_only; }    

    void set_ref_cube(CubeCached* cube); // reference cube in Percentage mode

    // label for a tree node
    std::string get_met_label  (int mode, const Metric* met)   const;
    std::string get_cnode_label(int mode, const Cnode*  cnode) const;
    std::string get_loc_label  (int mode, const Location* loc) const;
    std::string get_mod_label  (int mode, const Module* mod)   const;
    std::string get_reg_label  (int mode, const Region* reg)   const;
    std::string get_sreg_label (const Region* reg)             const;
    
    // maximum of focused metric tree (displayed in spectrumpane)
    std::string get_max_label  ()                              const;
    std::string get_halfmax_label()                            const;

    // color index for a tree node
    int get_met_color          (int mode, const Metric* met)   const;
    int get_cnode_color        (int mode, const Cnode*  cnode) const;
    int get_mod_color          (int mode, const Module* mod)   const;
    int get_reg_color          (int mode, const Region* reg)   const;
    int get_sreg_color         (const Region* reg)             const;
    int get_loc_color          (int mode, const Location* loc) const;

    // access cube's data member
    const std::vector<Metric*>&   get_met_roots();
    const std::vector<Module*>&   get_mod_roots();
    const std::vector<Cnode*>&    get_cnode_roots();
    const std::vector<Machine*>&  get_loc_roots();
    const std::vector<Region*>&   get_regionv();
    std::string                   get_attr(std::string key);

    int num_procs();
    int max_num_thrd();
    
  private:

    double get_max (const Metric* use_sel_met = NULL)  const;
    std::string get_max_label(double max) const;
    double sev2disp(double sev, double maxsev, double refsev) const;
    std::string gen_abs_label(double sev, std::string name, std::string uom,
			      double maxsev) const;
    
    std::string get_loc4cnode_label (int mode, const Location* loc) const;
    std::string get_loc4module_label(int mode, const Location* loc) const;
    std::string get_loc4reg_label   (int mode, const Location* loc) const;
    std::string get_loc4sreg_label  (int mode, const Location* loc) const;

    int get_loc4cnode_color         (int mode, const Location* loc) const;
    int get_loc4module_color        (int mode, const Location* loc) const;
    int get_loc4reg_color           (int mode, const Location* loc) const;
    int get_loc4sreg_color          (int mode, const Location* loc) const;

    double sev2perc(double sev, double maxsev, double refsev) const;
    int color_index(double percent) const;
    
    CubeCached*   cube;           // cube data displayed in GUI
    CubeCached*   ref_cube;       // reference cube data to compute perc

    int           midtree_type;   // display which type of tree
    int           sev_mode;       // sev in absolute, perc, or relative perc
    bool          pos_only;       // displays positive val only
    bool          neg_only;       // displays negative val only

    Metric*       sel_left;       // selected metric id in the 1st dimension
    int           sel_left_mode;  // selected metric state in the 1st dimension
    Cnode*        sel_ctree_mid;  // selected call-node in the middle dimension
    int           sel_ctree_mid_mode;
    Vertex*       sel_flat_mid;   // selected flat-entity in the middle dimension
    int           sel_flat_mid_mode; 
    int           sel_mid_type;   // which class is the middle entity
  };
  
}

#endif






