/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/*******************************************


	Parser.cpp


********************************************/

#include <iostream>
#include <sstream>

#include "Parser.h"
#include "CUBException.h"
#include "Metric.h"
#include "Module.h"
#include "Region.h"
#include "Callsite.h"
#include "Cnode.h"
#include "Machine.h"
#include "Node.h"
#include "Process.h"
#include "Thread.h"
#include "Cube.h"

using namespace std;
using namespace cube;

/* Constructor */
Parser::Parser(Cube& cube) {
  this->cube = &cube;
  last_vertex  = NULL;
  state = -1;
  thrd_id = 0;
  error_flag = 0;

  memset(&handler, 0, sizeof(handler));

  /* assign event handlers */
  handler.startElement = startElement;
  handler.endElement   = endElement;
  handler.characters   = characters;
  handler.cdataBlock   = cdataBlock;

  ctxt = ::xmlCreatePushParserCtxt(&handler,this, 0, 0, 0);
  if (ctxt == NULL) throw CUBException("ParserCtxt created error");
  /* allocate parsing buffer */
  buffer = (char*) malloc(BUFFSIZE);
  if (buffer == NULL) throw CUBException("malloc error");

}

/* Destructor */
Parser::~Parser() { 
  xmlFreeParserCtxt(ctxt);
  free(buffer);
  builder_memo.clear();
}

/* Parse user's given istream and build a Cube object in memory*/
int Parser::do_parse(istream& in) {
  bool ret; // parser's return valule.

  while (in.good()){
    int i = 0;
    char ch;
    while(in.read(&ch, 1) && ch != '>') {
      if(((int)ch) == 0x0D) ch = ' '; // shut outs DOS eol.
      buffer[i++] = ch;
      /* special handler for CDATA section */
      if (ch == '[' && i >= 9) {
	string str1(buffer+i-9, 9);
	if (str1 == "<![CDATA[") {
	  bool done = false;
	  while(!done && in.read(&ch, 1)) {// continue reading out the whole block
	    if(((int)ch) == 0x0D) ch = ' '; // shut outs DOS eol.
	    buffer[i++] = ch;
	    if (ch == '>' && i >= 12) {
	      string str2(buffer+i-3, 3);
	      if (str2 == "]]>") {
		done = true;
	      }
	    }
	  }
	}
      }// end of special handler
    }

    if (ch == '>') buffer[i] = '>';
    buffer[++i] = 0;
    
    if (i >= BUFFSIZE) {
      cerr << "Warning: Parsing a file with a big element" << endl;
    }
    //cout << "what parser reads: " << buffer;    
    ret = ::xmlParseChunk(ctxt, buffer, i, 0); // 0 indicates success.
    if (error_flag < 0) break; // FATAL PARSER ERROR HAPPEND!
    if (ret != false) {
      cerr << "Can't parse this chunk: " << buffer << endl;
      throw CUBException("Parser reading XML error!"); 
    }   
  }  
  if (error_flag == 0) {
    ::xmlParseChunk(ctxt, buffer, 0, 1);
  }
  return error_flag;
}

/* Parser changes its state when entering an element */
void Parser::enter_state_change(ItemType& item) {
  if (item.first == "cube") {
    state = PARSE_CUBE;
  }
  else if (item.first == "di" || item.first == "attr") {
    state = PARSE_DI;
  }
  if (item.first == "murl") {
    state = PARSE_MIRROR;
  }
  else if(item.first == "metric") {
    state = PARSE_METRIC;      
  } 
  else if (item.first == "module") {
    state = PARSE_MODULE;
  }
  else if (item.first == "region") {
    state = PARSE_REGION;
  }
  else if (item.first == "csite") {
    state = PARSE_CSITE;
  }
  else if (item.first == "cnode") {
    state = PARSE_CNODE;
  }
  else if (item.first == "machine") {
    state = PARSE_MACHINE;   
  }
  else if (item.first == "node") {
    state = PARSE_NODE;
  }
  else if (item.first == "process") {
    state = PARSE_PROCESS;
  }
  else if (item.first == "thread") {
    state = PARSE_THREAD;
  }
  else if (item.first == "severity") {
    state = PARSE_SEVERITY;
  }

  return;
}

/* Parser changes its state when leaving an element */
void Parser::exit_state_change(ItemType& current) {
  if (current.first == "cnode")   leave_cnode();
  else if (current.first == "node")    state = PARSE_MACHINE;
  else if (current.first == "process") state = PARSE_NODE;
  else if (current.first == "thread")  state = PARSE_PROCESS;

  return;
}

/* Based on current state, parser calls its corresponding handler. 
   This method is called in element_end_event handler.
*/
void Parser::dispatch_state_handler(ItemType& parent, ItemType& current )
{
  try {
    switch(state) {
    case PARSE_DI:
      parse_di(parent, current);
    case PARSE_MIRROR:
      parse_mirror(parent, current);
    case PARSE_METRIC:
      parse_met(parent, current);
      break;
    case PARSE_MODULE:
      parse_module(parent, current);
      break;
    case PARSE_REGION:
      parse_region(parent, current);
    break;
    case PARSE_CSITE:
      parse_csite(parent, current);
      break;
    case PARSE_MACHINE:
      parse_machine(parent, current);
      break;
    case PARSE_NODE:
      parse_node(parent, current);
      break;
    case PARSE_PROCESS:
      parse_process(parent, current);
      break;
    case PARSE_THREAD:
      parse_thread(parent, current);
      break;
    case PARSE_SEVERITY:
      parse_row(parent, current);
    }
  } catch(CUBException e) {
    cerr << e;
  }
}

/* be called whenever entering an element */
void Parser::start_element(const xmlChar* name, const xmlChar ** atts) {
  char** attributes = (char**) atts;
  ItemType item;
  /* back up the element */
  item.first = (char*) name;
 
  /* debug by walking the tree */
  // cout << "Meeting " << name;
  for(int i = 0; attributes != NULL && attributes[i] != NULL; i++) {
    item.second.push_back(string(attributes[i]));
  }

  /* push stack */
  track_stack.push(item);
  /* state changes */
  enter_state_change(item);

  /* Special cases, parse tags at the entry point*/
  if (state == PARSE_CNODE) parse_cnode(item); 
  if (state == PARSE_CUBE)  parse_cube_version(item);

  return;
}

/* being called whenever leaving an element */
void Parser::end_element(const xmlChar* name) {
  ItemType parent;
  ItemType current;

  current = track_stack.top();
  track_stack.pop();
  if (track_stack.empty()) return; /* parser is done */
  parent = track_stack.top();
  
  // cout << ":parse " << parent.first << " " << current.first << endl;
  /* call different handlers */
  dispatch_state_handler(parent, current);
  
  /* exit the element */
  data = "";
  exit_state_change(current);

}

/* SAX events order: start_element->characters->end_element */
void Parser::data_content(string content) {
  int i;
  char ch;
  /* skip the string consisting of only spaces, tabs and newlines. */
  for (i = 0; i < content.size(); i++) {
    ch = content.at(i);
    if (ch != ' ' && ch != '\n' && ch != '\t') break;
  }
  if (i >= content.size()) {
    return; // no content in string
  }
  data = content;
  return;
}

void Parser::cdata_block(string block) {
  data = block;  
}

int Parser::get_attr_id(ItemType item) {
  int i;
  if (item.second.size() < 2) throw CUBException("attribute id not found");
  i = atoi(item.second[1].c_str());
  return i;
}

void Parser::builder_memo_check(int n) {
  if (builder_memo.size() != n) {
    throw CUBException("Parser error: incomplete elements for an entity\n");
  }
}

void Parser::parse_cube_version (ItemType& current) {
  version = current.second[1];
  if (version.empty()) {
    throw CUBException("Parser error: unknown cube file version!");
  }
  if (version.find("2.0") == string::npos) {
    error_flag = -1;
  }
  state = -1;
}


void Parser::parse_di(ItemType& parent, ItemType& current) {
  if (current.first == "di" || current.first == "attr") {
    cube->def_attr(current.second[1], current.second[3]);
    // Try to detect user-defined ct-aggr-mode here.
    if (current.second[1] == "CUBE Call Tree Aggregation Mode") {
      if (current.second[3] == "SUM") 
	cube->def_attr("CUBE_CT_AGGR", "SUM");
      else if (current.second[3] == "MAX")
	cube->def_attr("CUBE_CT_AGGR", "MAX");
    }
  }
}

void Parser::parse_mirror(ItemType& parent, ItemType& current) {
  if (current.first == "murl") {
    cube->def_mirror(data);
  }
}

void Parser::parse_met(ItemType& parent, ItemType& current) {
  if( parent.first  == "metric" && 
     (current.first == "name"   || 
      current.first == "uom"    ||
      current.first == "url"    ||
      current.first == "descr") ) {
    builder_memo.push_back(data); // collecting a metric's info.
    // only if reaching the tag of </descr>, build a metric object
    if (current.first == "descr") {
      builder_memo_check(4);
      if (last_vertex == NULL) {
	last_vertex = 
	  cube->def_met(builder_memo[0], builder_memo[1], builder_memo[2], 
			builder_memo[3], NULL);
      } 
      else {
	Metric* p = static_cast<Metric*>(const_cast<Vertex*>(last_vertex));
	last_vertex = 
	  cube->def_met(builder_memo[0], builder_memo[1], builder_memo[2],
			builder_memo[3], p);
      }
      met_mapping[get_attr_id(parent)] = 
	static_cast<Metric*>(const_cast<Vertex*>(last_vertex));
      builder_memo.clear(); // after creation of a new metric
    }
  }
  /* after finishing a metric element, last_vertex
     moves one level up pointing to its parent metric */
  if (current.first == "metric") {
    if (last_vertex != NULL) {
      const Metric* p = static_cast<Metric*>(const_cast<Vertex*>(last_vertex));
      last_vertex = p->get_parent();
    }    
  }
}

void Parser::parse_module(ItemType& parent, ItemType& current) {
  if (parent.first == "module" && current.first == "name" ) {
    builder_memo.push_back(data);
    if (current.first == "name") {
      builder_memo_check(1);
      int id = get_attr_id(parent);
      mod_mapping[id] = cube->def_module(builder_memo[0]);
    }
    builder_memo.clear(); // after creation of a new Module
  }
}

void Parser::parse_region(ItemType& parent, ItemType& current) {
  if (parent.first == "region"  && 
      (current.first == "name"  || 
       current.first == "mod"   ||
       current.first == "begin" ||
       current.first == "end"   || 
       current.first == "url"   ||
       current.first == "descr")) {
    // cout << data << endl;
    builder_memo.push_back(data);
    /* when at the end of <descr>, new a Region object */
    if (current.first == "descr") {
      int id, modid, begin, end;
      builder_memo_check(6);
      if (builder_memo[1].empty() || builder_memo[2].empty() || 
	  builder_memo[3].empty()) {
	throw CUBException("Parser error: region lines or mod id undefined");
      }
      modid = atoi(builder_memo[1].c_str());
      const Module* m = mod_mapping[modid];
      begin = atoi(builder_memo[2].c_str());
      end = atoi(builder_memo[3].c_str());
      id = get_attr_id(parent);
      regn_mapping[id] = 
	cube->def_region(builder_memo[0], begin, end, builder_memo[4], builder_memo[5], m);
      builder_memo.clear(); // after creation of a new Region
    }
  }
}

void Parser::parse_csite(ItemType& parent, ItemType& current) {
  if (parent.first  == "csite"   && 
      (current.first == "line"   ||
       current.first == "mod" ||
       current.first == "callee") ) {
    builder_memo.push_back(data);
    if (current.first == "callee") {
      int id, mid, line, callee;
      builder_memo_check(3);
      if (builder_memo[1].empty()) {
	throw CUBException("Parser error: csite's callee empty");
      }	
      line = atoi(builder_memo[0].c_str());
      mid = atoi(builder_memo[1].c_str());      
      callee = atoi(builder_memo[2].c_str());
      id = get_attr_id(parent);
      csite_mapping[id] = cube->
	def_csite(mod_mapping[mid], line, regn_mapping[callee]);
      // cout << id << "->" << csite_mapping[id] << endl;
      builder_memo.clear(); // after creation of a new csite
    }
  }
}

/* being called from element_start event handler */
void Parser::parse_cnode(ItemType& current) {
  int id;
  string csite = current.second[3]; // csiteId="xxx"

  if (csite.empty()) {
    throw CUBException("Parser error: csiteId undefined");
  }
  int csid = atoi(csite.c_str());
  id = get_attr_id(current);
  if (last_vertex == NULL) {
    if (csite_mapping[csid] == NULL) {
      throw CUBException("Parser error: can't find csite id");
    }
    last_vertex = cube->def_cnode(csite_mapping[csid], NULL);
  } 
  else {
    Cnode* p = (Cnode*) last_vertex;
    last_vertex = cube->def_cnode(csite_mapping[csid], p);
  }
  cnode_mapping[id] = static_cast<Cnode*>(const_cast<Vertex*>(last_vertex));
}

void Parser::leave_cnode() {
  if (last_vertex != NULL) {
    const Cnode* cnode = static_cast<Cnode*>(const_cast<Vertex*>(last_vertex));
    last_vertex = cnode->get_parent();
  }
  if (last_vertex == NULL) state = -1; // not in PARSE_CNODE any more
}

void Parser::parse_machine(ItemType& parent, ItemType& current) {
  if (parent.first == "machine" && current.first == "name" ) {
    builder_memo.push_back(data);
    builder_memo_check(1);
    int id = get_attr_id(parent);
    loc_mapping[id] = cube->def_mach(builder_memo[0]);
    last_vertex = loc_mapping[id];
    builder_memo.clear();
  }
}

void Parser::parse_node(ItemType& parent, ItemType& current) {
  if (parent.first == "node" && current.first == "name") {
    builder_memo.push_back(data);
    builder_memo_check(1);

    if (last_vertex != NULL) { // should have a parent
      const Machine* m = static_cast<Machine*>(const_cast<Vertex*>(last_vertex));
      int id = get_attr_id(parent);
      loc_mapping[id] = cube->def_node(builder_memo[0], m);
      last_vertex = loc_mapping[id];
    } else throw CUBException("Parser error: SMP Node's parent = NULL");
    builder_memo.clear();
  }
  if (current.first == "node") {
    const Node* n = static_cast<Node*>(const_cast<Vertex*>(last_vertex));
    last_vertex = n->get_mach();
  }
}

void Parser::parse_process(ItemType& parent, ItemType& current) {
  if (parent.first == "process" && 
      (current.first == "name"  ||
       current.first == "rank") ) {
    builder_memo.push_back(data);
 
    if (current.first == "rank") {
      builder_memo_check(2);
      if (last_vertex != NULL) { // should have a parent
	const Node* node = static_cast<Node*>(const_cast<Vertex*>(last_vertex));
	int id = get_attr_id(parent); // process id
	int rank = atoi(builder_memo[1].c_str());
	string name = builder_memo[0];
	if (name.empty()) {
	  name = "Process " + builder_memo[1];
	}
	loc_mapping[id] = cube->def_proc(name, rank, node);
	last_vertex = loc_mapping[id];
      } else throw CUBException("Parser: process's parent = NULL");
      builder_memo.clear();
    }
  }  
  if (current.first == "process") {
    const Process* p = static_cast<Process*>(const_cast<Vertex*>(last_vertex));
    last_vertex = p->get_node();
  }
}

void Parser::parse_thread(ItemType& parent, ItemType& current) {
  if (parent.first == "thread" && 
      (current.first == "name" ||
       current.first == "rank" ) ) {
    builder_memo.push_back(data);
    
    if (current.first == "rank") {
      builder_memo_check(2);
      if (last_vertex != NULL) { // should have a parent
	const Process* p = static_cast<Process*>
	  (const_cast<Vertex*>(last_vertex));
	int id = get_attr_id(parent);
	string name = builder_memo[0];
	int rank = atoi(builder_memo[1].c_str());
	if (name.empty()) {
	  name = "Thread " + builder_memo[1];
	}
	loc_mapping[id] = thrd_mapping[thrd_id++] = cube->def_thrd(name, rank, p);
	last_vertex = loc_mapping[id];
      }
      else throw CUBException("Parser: thread's parent = NULL");
      builder_memo.clear();
    }
  }
  
  if (current.first == "thread") {
    const Thread* t = static_cast<Thread*>(const_cast<Vertex*>(last_vertex));
    last_vertex = t->get_proc();
  }
}

/* parse severity values */
void Parser::parse_row(ItemType& parent, ItemType& current) {
  double val = 0.0;
  int thrd_index = 0;

  if (current.first != "row") return;
  int met_id  = get_attr_id(parent);
  int cnode_id = get_attr_id(current);
  istringstream in(data);
  in.setf(ios::fixed, ios::floatfield);

  while(in >> val) {
    cube->set_sev(met_mapping[met_id], 
		  cnode_mapping[cnode_id], 
		  thrd_mapping[thrd_index], val);
    /*DEBUG
      if (met_id == 4 && cnode_id == 31) {
      cout << data.c_str();
      }*/
    thrd_index++;
  }
  return;
}

namespace cube {
  // switch to the parser's event-handlers.
  void startElement(void* parser, const xmlChar *name, const xmlChar **atts) {
    // cout << "**start element: " << (char*)name << endl;
    ((Parser*)(parser))->start_element(name, atts);
  }
  
  void endElement(void* parser, const xmlChar *name) {
    // cout << "**end element: " << (char*)name << endl;
    ((Parser*)(parser))->end_element(name);
  }
  
  void characters(void* parser, const xmlChar* text, int length) {
    string content((char*)text, length);
    // cout << "**characters: " << content;
    ((Parser*)(parser))->data_content(content);    
  }

  void cdataBlock   (void* parser, const xmlChar* text, int length) {
    string block((char*)text, length);
    // cout << "**cdata: " << block;
    ((Parser*)(parser))->cdata_block(block);
  }
}
