#ifndef CUBE_XMLPARSER_H
#define CUBE_XMLPARSER_H

/*
 *----------------------------------------------------------------------------
 *
 * class Parser
 *
 *----------------------------------------------------------------------------
 */

#include <iostream>
#include <string>
#include <stack>
#include <vector>
#include <utility>
#include <map>
  
#include "libxml/parser.h"
#include "Vertex.h"

#define BUFFSIZE      1024*1024

//Parser's DFA states
#define PARSE_DI          0
#define PARSE_METRIC      1
#define PARSE_MODULE      2
#define PARSE_REGION      3
#define PARSE_CSITE       4
#define PARSE_CNODE       5
#define PARSE_MACHINE     6
#define PARSE_NODE        7
#define PARSE_PROCESS     8
#define PARSE_THREAD      9
#define PARSE_SEVERITY    10
#define PARSE_MIRROR      11
#define PARSE_CUBE        12

namespace cube {

  class Cube;
  class Metric;
  class Module;
  class Region;
  class Callsite;
  class Cnode;
  class Location;
  class Thread;

  typedef std::pair<std::string, std::vector<std::string> > ItemType;
  
  class Parser {
  public:
    
    Parser      (Cube& cube);
    ~Parser();

    int  do_parse      (std::istream& in); 
    
    void start_element (const xmlChar* name, const xmlChar ** atts);
    void end_element   (const xmlChar* name);
    void data_content  (std::string content);
    void cdata_block   (std::string block);

    std::string get_version() { return version; };

  private:
    void parse_cube_version (ItemType& current);
    void parse_di           (ItemType& parent, ItemType& current);
    void parse_mirror       (ItemType& parent, ItemType& current);
    void parse_met          (ItemType& parent, ItemType& current);
    void parse_module       (ItemType& parent, ItemType& current); 
    void parse_region       (ItemType& parent, ItemType& current); 
    void parse_csite        (ItemType& parent, ItemType& current); 
    void parse_cnode        (ItemType& current); 
    void leave_cnode        ();
    void parse_machine      (ItemType& parent, ItemType& current); 
    void parse_node         (ItemType& parent, ItemType& current); 
    void parse_process      (ItemType& parent, ItemType& current); 
    void parse_thread       (ItemType& parent, ItemType& current);  
    void parse_row          (ItemType& parent, ItemType& current);

    void enter_state_change     (ItemType& cur);
    void exit_state_change      (ItemType& cur);
    void dispatch_state_handler (ItemType& parent, ItemType& current);

    int  get_attr_id(ItemType item);
    void builder_memo_check(int n); // error checking
    
    Cube*              cube;
    const Vertex*      last_vertex; // parent node kept for creating next object.
 
    std::string version;
    int error_flag;                 // 0 if correct

    /* required by libxml2 */
    xmlSAXHandler      handler;
    xmlParserCtxtPtr   ctxt;
    char*              buffer;


    /* for walking xml documents */
    int state;
    std::stack<ItemType, std::vector<ItemType> > track_stack;
    std::string   data;     // data within elements
    // std::string   prevdata; // the immediately previous data

    /* builder_memo is used to record a CUBE object's information located
       at previously walked elements. It contains those elements' contents.*/ 
    std::vector<std::string>  builder_memo;

    int thrd_id; // allocate id for thread dynamically
    
    // record the mappings: entity id -> entity pointer
    std::map<int, const Metric*>   met_mapping;
    std::map<int, const Module*>   mod_mapping;
    std::map<int, const Region*>   regn_mapping;
    std::map<int, const Callsite*> csite_mapping;
    std::map<int, const Cnode*>    cnode_mapping;
    std::map<int, const Location*> loc_mapping;
    std::map<int, const Thread*>   thrd_mapping;

  };

  /* dispatch specific parser's event-handler methods */
  void startElement (void* parser, const xmlChar *name, const xmlChar **atts);
  void endElement   (void* parser, const xmlChar *name);
  void characters   (void* parser, const xmlChar* text, int length);
  void cdataBlock   (void* parser, const xmlChar* text, int len);
}

#endif


