/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/************************************************
             Space.cpp

 ************************************************/

#include <iostream>
#include <fstream>
#include <string>

#include <set>
#include <algorithm>

#include "CUBException.h"
#include "Metric.h"
#include "Module.h"
#include "Region.h"
#include "Callsite.h"
#include "Cnode.h"
#include "Machine.h"
#include "Node.h"
#include "Process.h"
#include "Thread.h"

#include "Space.h"

#include "Cube.h"
#include "Severity.h"

using namespace std;
using namespace cube;

Space::Space() {
  top_region = NULL;
}

Space::~Space() {
  if (top_region != NULL) delete(top_region);
}

void Space::def_attr(string key, string value) {
  attrs[key] = value;
}

string Space::get_attr(std::string key) {
  return attrs[key];
}


void Space::def_mirror(std::string url) {
  mirror_urls.push_back(url);
}

vector<string>&  Space::get_mirrors(){
  return mirror_urls;
}

// metric dimension

// A url may be file:// or http://, or "@mirror@...".
Metric* Space::def_met(string name, string uom, string url, string descr, 
		       Metric* parent) {
  Metric* met = NULL;

  /* could read in a file's contents
  char buffer[1024];
  if (!descr.empty() && (descr[0] == '/' || descr[1] == ':')) {
    // If it is a absolute path.
    ifstream in (descr.c_str());
    if (!in) {
      cout << "Error: can't open metric description file " << descr << endl;
      in.close();
    } else {
      descr = "";
      while(in.good()) {
	in.getline(buffer, 1024);
	descr += buffer;
	descr += "\n";
	// cout << descr;
      }
      in.close();
    }
  }
  */
  // create a new metric
  met = new Metric(name, uom, url, descr, parent);
  // whether it's a root
  if(parent == NULL) root_metv.push_back(met);
  else parent->add_child(met);
  metv.push_back(met); 
  return met;    
}

// program dimension
Module* Space::def_module(string name) {
  Module* mod = NULL;

  mod = new Module(name);
  modv.push_back(mod);
  return mod;
}

Region* Space::def_region(string name, long begln, long endln, 
			  string url, string descr, Module* mod) {
  Region* reg = NULL;

  /* could read in a file's contents
  char buffer[1024];
  if (!descr.empty() && (descr[0] == '/' || descr[1] == ':')) {
    // If it is a absolute path.
    ifstream in (descr.c_str());
    if (!in) {
      cout << "Error: can't open region description file " << descr << endl;
      in.close();
    } else {
      descr = "";
      while(in.good()) {
	in.getline(buffer, 1024);
	descr += buffer;
	descr += "\n";
	// cout << descr;
      }
      in.close();
    }
  }
  */

  reg = new Region(name, begln, endln, url, descr, mod);
  regionv.push_back(reg);
  mod->add_region(reg);
  return reg;
}

Callsite* Space::def_csite(Module* mod, int line, Region* callee) {
  Callsite* csite;
  csite = new Callsite(mod, line, callee);
  csitev.push_back(csite);
  return csite;
}

Cnode* Space::def_cnode(Callsite* csite, Cnode* parent) {
  Cnode* cnode;
  Region* caller;

  cnode = new Cnode(csite, parent);
  csite->add_cnode(cnode);
  if (parent  == NULL) {
    root_cnodev.push_back(cnode); // root node
    if(top_region == NULL) { // create the virtual caller
      Module* m = const_cast<Module*>(csite->get_mod());
      top_region = new Region("virtual caller", 0, 0, "", "not existent", m);
      //caller = def_region("virtual caller", 0, 0, "not existing", m);
      m->add_region(top_region);
    }
    top_region->add_csite(csite);
    csite->set_caller(top_region);
  }
  else {
    parent->add_child(cnode);
    const Callsite* psite = const_cast<Cnode*>(parent)->get_csite();
    caller = const_cast<Region*>(psite->get_callee()); // but current one's caller
    csite->set_caller(caller);
    caller->add_csite(csite);
  }
  cnodev.push_back(cnode);
  return cnode;
}

Cnode* Space::clone_cnode(const Cnode* origin, Cnode* parent) {
  return def_cnode(copy_if_notfound(origin->get_csite()), parent);  
}

Module* Space::copy_if_notfound(const Module* mod) {
  ent_equal<Module*> equal(const_cast<Module*>(mod));
  vector<Module*>::iterator mi = find_if(modv.begin(), modv.end(), equal);
  if (mi == modv.end()) {
    return def_module(mod->get_name());
  } else {
    return *mi;
  }
}

Region* Space::copy_if_notfound(const Region* r) {
  ent_equal<Region*> equal(const_cast<Region*>(r));
  vector<Region*>::iterator ri = find_if(regionv.begin(), regionv.end(), equal);
  if (ri == regionv.end()) {
    return def_region(r->get_name(), r->get_begn_ln(), r->get_end_ln(), r->get_url(),
		      r->get_descr(), copy_if_notfound(r->get_mod()));
  } else {
    return *ri;
  }
}
    
Callsite* Space::copy_if_notfound(const Callsite* csite) {
  ent_equal<Callsite*> equal(const_cast<Callsite*>(csite));
  vector<Callsite*>::iterator ci = find_if(csitev.begin(), csitev.end(), equal);
  if (ci == csitev.end()) {
    return def_csite(copy_if_notfound(csite->get_mod()),
		     csite->get_line(),
		     copy_if_notfound(csite->get_callee()));
  } else {
    return *ci;
  }
}

// system dimension
Machine* Space::def_mach(string name) {
  Machine* m;

  m = new Machine(name);
  machv.push_back(m);
  return m;
}

Node*  Space::def_node(string name, Machine* mach) {
  Node* node;
  
  node = new Node(name, mach);
  mach->add_node(node);
  nodev.push_back(node);
  return node;
}

Process* Space::def_proc(string name, int rank, Node* node) {
  Process* proc;

  proc = new Process(name, rank, node);
  node->add_proc(proc);
  procv.push_back(proc);
  return proc;
}

Thread* Space::def_thrd(string name, int rank, Process* proc) {
  Thread* thrd;

  thrd = new Thread(name, rank, proc);
  proc->add_thrd(thrd);
  thrdv.push_back(thrd);
  return thrd;
}

int Space::max_num_thrd() const {
  int ret = 0;
  for (int i = 0; i < procv.size(); i++) {
    const Process* p = procv[i];
    if (p->num_thrds() > ret) ret = p->num_thrds();
  }
  return ret;
}

const Metric* Space::find_met(string name) const {
  for (int i = 0; i < metv.size(); i++) {
    if (metv[i]->get_name() == name) 
      return metv[i];
  }
  return NULL;
}


// locally defined function for operator==()
bool compare_sys_forest(const Space& lhs, const Space& rhs) {
  const vector<Process*>& lhs_procs = lhs.get_procv();
  const vector<Process*>& rhs_procs = rhs.get_procv();
  const vector<Thread*>&  lhs_thrds = lhs.get_thrdv();
  const vector<Thread*>&  rhs_thrds = rhs.get_thrdv();

  if (lhs_procs.size() != rhs_procs.size() ||
      lhs_thrds.size() != rhs_thrds.size()) {
    return false;
  }
  // based on comparing pairs (procid, thrdid)
  for (int i = 0; i < lhs_thrds.size(); i++) {
    if(!exists(lhs_thrds[i], rhs_thrds)) return false;
  }
  return true;
}

bool operator== (const Space& lhs, const Space& rhs) {
  if (!compare_forest<Metric>(lhs.get_met_roots(), rhs.get_met_roots())) {
    cout << "metric dimension different" << endl;
    return false;
  }
  if (!compare_forest<Cnode>(lhs.get_prog_roots(), rhs.get_prog_roots())) {
    cout << "program dimension different" << endl;
    return false;
  }
  if (!compare_sys_forest(lhs, rhs)) {
    cout << "system dimension different" << endl;
    return false;
  }  
  return true;
}

// mount a whole tree to this space's metric tree
void Space::
copy_tree(Metric* attach_pt, const vector<Metric*> rmt_mets, map<Metric*, Metric*>& mapping) {
  Metric* m;
  for (int i = 0; i < rmt_mets.size(); i++) {
    m = rmt_mets[i];
    Metric* newmet = def_met(m->get_name(), m->get_uom(), 
			     m->get_url(), m->get_descr(), attach_pt);    
    mapping[newmet] = m;
    copy_tree(newmet, m->get_children(), mapping);
  }
}

// mount to a cnode tree
void Space::
copy_tree(Cnode* attach_pt, const vector<Cnode*> rmt_cnodes, map<Cnode*, Cnode*>& mapping) {
  Cnode* cn;
  for (int i = 0; i < rmt_cnodes.size(); i++) {
    cn = rmt_cnodes[i];
    Cnode* newcnode = clone_cnode(cn, attach_pt);
    mapping[newcnode] = cn;
    copy_tree(newcnode, cn->get_children(), mapping);
  }

}

namespace cube {

Space* space_merge(const Space* lhs, const Space* rhs, 
		   SpaceMapping& lhsm, SpaceMapping& rhsm) {
  
  Space* newspace = new Space();

  merge_met_forest(NULL, lhs->get_met_roots(), rhs->get_met_roots(),
		   lhsm.metm, rhsm.metm, newspace);
  
  merge_prog_forest(NULL, lhs->get_prog_roots(), rhs->get_prog_roots(),
		    lhsm.cnodem, rhsm.cnodem, newspace);
  
  merge_sys_tree(lhs, rhs, lhsm.sysm, rhsm.sysm, newspace);
  
  vector<string> mirrors = const_cast<Space*>(lhs)->get_mirrors();
  for (int i = 0; i < mirrors.size(); i++) {
    newspace->def_mirror(mirrors[i]);
  }

  return newspace;
}

void merge_met_forest(const Metric* newupper, const vector<Metric*> lhs, 
		      const vector<Metric*> rhs, 
		      map<Metric*, Metric*>& lhsm, map<Metric*, Metric*>& rhsm,
		      Space* newspace) {
  int i;
  Metric *m, *newm;
  vector<Metric*> union_mets;

  // compute the union in the current level.
  for (i = 0; i < lhs.size(); i++) {
    m = lhs[i];
    newm = newspace->def_met(m->get_name(), m->get_uom(), m->get_url(), 
			     m->get_descr(), const_cast<Metric*>(newupper));
    union_mets.push_back(newm);
    lhsm[newm] = m;
  }
  for (i = 0; i < rhs.size(); i++) {
    m = exists(rhs[i], union_mets);
    if (m == NULL) {
      m = rhs[i];
      newm =  newspace->def_met(m->get_name(), m->get_uom(), m->get_url(), 
				m->get_descr(), const_cast<Metric*>(newupper));
      union_mets.push_back(newm);
      rhsm[newm] = m;
      newspace->copy_tree(const_cast<Metric*>(newupper), m->get_children(), rhsm);
    } else {
      rhsm[m] = rhs[i];
      merge_met_forest(m, lhsm[m]->get_children(), rhsm[m]->get_children(), 
		       lhsm, rhsm, newspace);
    }
  }

}


void merge_prog_forest(const Cnode* newupper, 
		       const vector<Cnode*> lhs, const vector<Cnode*> rhs, 
		       map<Cnode*, Cnode*>& lhsm, map<Cnode*, Cnode*>& rhsm,
		       Space* newspace) {

  int i;
  Cnode *newcn, *cn;
  vector<Cnode*> union_cnodes;
  
  for (i = 0; i < lhs.size(); i++) {
    cn = lhs[i];
    newcn = newspace->clone_cnode(cn, const_cast<Cnode*>(newupper));
    union_cnodes.push_back(newcn);
    lhsm[newcn] = cn;
  }

  for (i = 0; i < rhs.size(); i++) {
    cn = exists(rhs[i], union_cnodes);
    if (cn == NULL) {
      cn = rhs[i];
      newcn = newspace->clone_cnode(cn, const_cast<Cnode*>(newupper));
      union_cnodes.push_back(newcn);
      rhsm[newcn] = cn;
      newspace->copy_tree(const_cast<Cnode*>(newupper), cn->get_children(), rhsm);
    } else {
      rhsm[cn] = rhs[i];
      merge_prog_forest(cn, lhsm[cn]->get_children(), rhsm[cn]->get_children(), 
		       lhsm, rhsm, newspace);
    }
  }
}

// locally defined function for is_compatible()
bool subset(const Node* A, const Node* B) {
  set<int> Pa, Pb, diff;
  int i;
  // cout << "set A: ";
  for (i = 0; i < A->num_children(); i++) {
    Pa.insert(A->get_child(i)->get_rank());
    // cout << A->get_child(i)->get_rank() << " ";
  }
  // cout << endl;
  // cout << "set B: ";
  for (i = 0; i < B->num_children(); i++) {
    Pb.insert(B->get_child(i)->get_rank());
    // cout << B->get_child(i)->get_rank() << " ";
  }
  //cout << endl;
  //Create an insert_iterator for result
  insert_iterator<set<int> >  diff_ins(diff, diff.begin());
  set_difference(Pa.begin(), Pa.end(), 
		 Pb.begin(), Pb.end(), diff_ins);
  //cout << diff.empty() << endl;
  return diff.empty();
}

// locally defined function for merget_sys_tree()

// FOR equal # of nodes
bool is_compatible_for_equal(const vector<Node*>& lhs, vector<Node*> rhs,
			     map<Node*, Node*>& relation) {
  vector<Node*>::iterator j;
  // lhs <= rhs
  for (int i = 0; i < lhs.size(); i++) {
    bool found = false;
    for (j = rhs.begin(); j != rhs.end(); j++) {
      if (subset(lhs[i], *j)) {
	relation[*j] = lhs[i]; // equal size, pick the the RHS!
	found = true;
	rhs.erase(j);
	break;
      } 
    }
    if (!found) {
      return false;
    }
  }
  return true;
}
bool is_compatible(const vector<Node*>& lhs, vector<Node*> rhs,
		   map<Node*, Node*>& relation, bool& bigger_left) {
  vector<Node*>::iterator j;
  
  if (lhs.size() == rhs.size()) {
    cout << "Equal node num: comparing 1st time" << endl;
    if(is_compatible_for_equal(lhs, rhs, relation)) {
      bigger_left = false;
      return true;
    } else {
      cout << "Equal node num: comparing 2nd time" << endl;
      relation.clear();
      if (is_compatible_for_equal(rhs, lhs, relation)) {
	bigger_left = true;
	return true;
      }
      else return false;
    }
  } else {
    // if not equal node #
    if (lhs.size() > rhs.size() ) {
      bigger_left = true;
      bool tmp;
      return is_compatible(rhs, lhs, relation, tmp);
    }
    bigger_left = false;
    for (int i = 0; i < lhs.size(); i++) {
      bool found = false;
      for (j = rhs.begin(); j != rhs.end(); j++) {
	if (subset(lhs[i], *j)) {
	  relation[*j] = lhs[i]; // >= subset relation record
	  // cout << (*j)->get_name() << "-->" << lhs[i]->get_name() << endl;
	  found = true;
	  rhs.erase(j);
	  break;
	} 
      }
      if (!found/*j == rhs.end()*/) {
	return false;
      }
    }
    return true;
  }
}

// precondition: compatible trees as input. lhs is the SUPERSET.
void 
merge_compa_sys_tree(const vector<Node*>& lhs, const vector<Node*>& rhs, 
		     map<Location*, Location*>& lhsm, map<Location*, Location*>& rhsm,
		     map<Node*, Node*> relation, Space* newspace) {
  int i, j;
  Node* node;
  Process* proc;
  // choose bigger one.
  //if (lhs.size() < rhs.size()) 
  //  return merge_compa_sys_tree(rhs, lhs, rhsm, lhsm, relation, newspace);
  Machine* mach = newspace->def_mach(lhs[0]->get_mach()->get_name());
  lhsm[mach] = const_cast<Machine*>(lhs[0]->get_mach());
  rhsm[mach] = const_cast<Machine*>(rhs[0]->get_mach());
  // clone upper 3 levels
  for (i = 0; i < lhs.size(); i++) {
    Node* lhs_node = lhs[i];
    node = newspace->def_node(lhs_node->get_name(), mach);
    lhsm[node] = lhs_node;
    rhsm[node] = relation[lhs_node];
    for (j = 0; j < lhs_node->num_children(); j++) {
      Process* lhs_proc = const_cast<Process*>(lhs_node->get_child(j));
      proc = newspace->def_proc(lhs_proc->get_name(),
				lhs_proc->get_rank(),node);
      lhsm[proc] = lhs_proc;
      if (relation[lhs_node] != NULL) {
	rhsm[proc] = exists(lhs_proc, relation[lhs_node]->get_children());
      }
    }
  }
  
}

void
merge_thrd_level(map<Location*, Location*>& lhsm, map<Location*, Location*>& rhsm,
		 Space* newspace) {
  int i, j;
  Process *lhsp, *rhsp;
  Thread* nt;
  const vector<Process*>& procv = newspace->get_procv();

  for (i = 0; i < procv.size(); i++) {
    lhsp = static_cast<Process*>(lhsm[procv[i]]);
    rhsp = static_cast<Process*>(rhsm[procv[i]]);
    // ret = LHS
    for (j = 0; lhsp != NULL && j < lhsp->num_children(); j++) {
      nt = newspace->def_thrd(lhsp->get_child(j)->get_name(), 
			      lhsp->get_child(j)->get_rank(), procv[i]);
      lhsm[nt] = const_cast<Thread*>(lhsp->get_child(j));
    }
    // ret += RHS - LHS
    for (j = 0; rhsp != NULL && j < rhsp->num_children(); j++) {
      Thread* rhs_thrd = const_cast<Thread*>(rhsp->get_child(j));
      Thread* t = exists(rhs_thrd, procv[i]->get_children());
      if (t == NULL) { 
	Thread* nt = newspace->def_thrd(rhs_thrd->get_name(),
					rhs_thrd->get_rank(), procv[i]);
	rhsm[nt] = rhs_thrd;
      } else {
	rhsm[t] = rhs_thrd;
      }
    }
    // cout << "merged a process" << endl;
  }
  
}

void 
merge_sys_tree(const Space* lhs, const Space* rhs, 
	       map<Location*, Location*>& lhsm, 
	       map<Location*, Location*>& rhsm,
	       Space* newspace) {  
  int i;
  map<Node*, Node*> rel;
  const vector<Machine*>& lhs_machs = lhs->get_sys_roots();
  const vector<Machine*>& rhs_machs = rhs->get_sys_roots();
  bool bigger_left = true;

  if (lhs_machs.size() != 1 || rhs_machs.size() != 1) {
    cerr << "Error: can't merge system dimensions with multiple machines";
    exit(1);
  }
  
  if (lhs_machs.size() == 1 && rhs_machs.size() == 1 &&
      is_compatible(lhs_machs[0]->get_children(), 
		    rhs_machs[0]->get_children(), rel, bigger_left) ) {
    cout << "System tree compatible: using the bigger one. Left or not? " 
	 << bigger_left << endl;
    // clone the bigger tree for upper 3 levels
    if(bigger_left) {
      merge_compa_sys_tree(lhs_machs[0]->get_children(), 
			   rhs_machs[0]->get_children(),
			   lhsm, rhsm, rel, newspace);
    } else {
      merge_compa_sys_tree(rhs_machs[0]->get_children(), 
			   lhs_machs[0]->get_children(),
			   rhsm, lhsm, rel, newspace);
    }
  } else {
    // collapse
    cout << "System tree incompatible: collapsing it" << endl;
    Machine* mach = newspace->def_mach("Machine");
    Node* node = newspace->def_node("Node", mach);
    Process* newp;
    const vector<Process*>& lhs_procs = lhs->get_procv();
    const vector<Process*>& rhs_procs = rhs->get_procv();
    for (i = 0; i < lhs_procs.size(); i++) {
      newp = newspace->
	def_proc(lhs_procs[i]->get_name(), lhs_procs[i]->get_rank(), node);
      lhsm[newp] = lhs_procs[i];
    }
    for (i = 0; i < rhs_procs.size(); i++) {
      Process* p = 
	exists(const_cast<Process*>(rhs_procs[i]), newspace->get_procv());
      if (p == NULL) {
	newp = newspace->def_proc(rhs_procs[i]->get_name(),
				  rhs_procs[i]->get_rank(), node);
	rhsm[newp] = rhs_procs[i];
      } else {
	rhsm[p] = rhs_procs[i];
      }
    }
  }
  cout << "Merging thread level" << endl;

  if (bigger_left) 
    merge_thrd_level(lhsm, rhsm, newspace); 
  else 
    merge_thrd_level(rhsm, lhsm, newspace); 

  cout << "Merging thread level done" << endl;
  
 }
  
}
