/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#ifndef CUBE_SPACE_H
#define CUBE_SPACE_H   

/*
 *----------------------------------------------------------------------------
 *
 * class Space
 *
 *----------------------------------------------------------------------------
 */

#include <string>
#include <iostream>
#include <map>
#include <vector>
#include <functional>

namespace cube {

  class Callsite;
  class Cnode;
  class Location;
  class Machine;
  class Metric;
  class Module;
  class Node;
  class Process;
  class Region;
  class Thread;

  
  struct SpaceMapping {
    std::map<Metric*, Metric*>      metm;
    std::map<Cnode*, Cnode*>        cnodem;
    std::map<Location*, Location*>  sysm;
  };
  
  class Space  {
  public:
    
    Space();
    ~Space();

    void         def_attr (std::string key, std::string value);
    std::string  get_attr (std::string key);
    const std::map<std::string, std::string>&  get_attrs() const { return attrs; }

    void                       def_mirror(std::string url);
    std::vector<std::string>&  get_mirrors();

    // metric dimension
    Metric*   def_met(std::string name, std::string uom, std::string url, 
		      std::string descr, Metric* parent);     
    // program dimension
    Module*   def_module(std::string name);
    Region*   def_region(std::string name, long begln, long endln, 
			 std::string url, std::string descr, Module* mod);
    Callsite* def_csite(Module* mod, int line, Region* callee);
    Cnode*    def_cnode(Callsite* csite, Cnode* parent);
    Cnode*    clone_cnode(const Cnode* origin, Cnode* parent);

    // system dimension
    Machine*  def_mach(std::string name);
    Node*     def_node(std::string name, Machine* mach);
    Process*  def_proc(std::string name, int rank, Node* node);
    Thread*   def_thrd(std::string name, int rank, Process* proc);
    
    // get the roots of each dimension
    const std::vector<Metric*>&  get_met_roots()  const { return root_metv; };
    const std::vector<Cnode*>&   get_prog_roots() const { return root_cnodev; };
    const std::vector<Machine*>& get_sys_roots()  const { return machv; };
    
    // get vector of entities
    const std::vector<Metric*>&   get_metv()      const { return metv; }
    const std::vector<Module*>&   get_modv()      const { return modv; }
    const std::vector<Region*>&   get_regnv()     const { return regionv; }
    const std::vector<Callsite*>& get_csitev()    const { return csitev; }
    const std::vector<Process*>&  get_procv()     const { return procv; }
    const std::vector<Thread*>&   get_thrdv()     const { return thrdv; }
    const std::vector<Cnode*>&    get_cnodev()    const { return cnodev; }

    int num_procs()    const { return procv.size(); }
    int max_num_thrd() const ;
    
    // search for a metric by name
    const Metric* find_met(std::string name) const;

    void copy_tree(Metric* attach_pt, const std::vector<Metric*> rmt_mets,
		   std::map<Metric*, Metric*>&); // metric tree
    void copy_tree(Cnode*  attach_pt, const std::vector<Cnode*> rmt_cnodes,
		   std::map<Cnode*, Cnode*>&); // prog tree
    
  private:
    
    std::vector<Metric*>  root_metv;
    std::vector<Cnode*>   root_cnodev;
    std::map<std::string, std::string> attrs; // <key, value> attributes
    std::vector<std::string> mirror_urls;

    Region* top_region; // a virtual region
    // machv contains roots of the system dimension
    
    std::vector<Metric*>   metv;    /* all metrics */  
    std::vector<Module*>   modv;    /* all modules */
    std::vector<Region*>   regionv; /* all regions */
    std::vector<Callsite*> csitev;  /* all call sites */
    std::vector<Cnode*>    cnodev;  /* all call-tree nodes */
    std::vector<Machine*>  machv;   /* all machines */
    std::vector<Node*>     nodev;   /* all nodes */
    std::vector<Process*>  procv;   /* all processes */
    std::vector<Thread*>   thrdv;   /* all threads */

    
    Module*   copy_if_notfound(const Module*);
    Region*   copy_if_notfound(const Region*);
    Callsite* copy_if_notfound(const Callsite*);
  
  };
  
  // non-member functions
  bool operator== (const Space& lhs, const Space& rhs);
 
  Space* space_merge(const Space* lhs, const Space* rhs, 
		     SpaceMapping& lhsm, SpaceMapping& rhsm);

  void merge_met_forest(const Metric* newupper, 
			const std::vector<Metric*> lhs, const std::vector<Metric*> rhs, 
			std::map<Metric*, Metric*>& lhsm, std::map<Metric*, Metric*>& rhsm,
			Space* newspace);

  void merge_prog_forest(const Cnode* newupper, 
			 const std::vector<Cnode*> lhs, const std::vector<Cnode*> rhs, 
			 std::map<Cnode*, Cnode*>& lhsm, std::map<Cnode*, Cnode*>& rhsm,
			 Space* newspace);
  
  void merge_sys_tree (const Space* lhs, const Space* rhs, 
		       std::map<Location*, Location*>& lhsm, 
		       std::map<Location*, Location*>& rhsm,
		       Space* newspace);

  
  
  /* Template Definitions of compare_forest(), used by metric&call trees*/

  template<class T>
  bool compare_forest(const std::vector<T*>& lhs, const std::vector<T*>& rhs) {
    T* copy;
    bool ret;
    int i;    
    for (i = 0; i < lhs.size(); i++) {// internal node
      copy = exists(lhs[i], rhs);
      if (copy == NULL) return false;
      ret = compare_forest<T>(lhs[i]->get_children(), copy->get_children());
      if (ret == false) return false; 
    }
    if (i > 0 && i == lhs.size()) return true;

    // lhs = leaf node (i == 0)
    if (rhs.size() != 0) return false;
    else return true;
  }

  
  template<class T>
  T* exists(T* target, const std::vector<T*> vec) { 
    for (int i = 0; i < vec.size(); i++) {
      if (*target == *vec[i]) return vec[i];
    }
    return NULL;
  }

  //unary predicate
  template<class T>
  class  ent_equal : public std::unary_function<T, bool> {
  private:
    T own;
  public :
    explicit ent_equal(T ent) { own = ent; }
    bool operator() (const T ent) const {
      return *ent == *own;
    }
  };

} // end of namespace


#endif
