/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <fstream>
#include <iostream>
#include <map>
#include <vector>
#include <string>

#include "Cube.h"
#include "TauProfile.h"

using namespace std;
using namespace cube;

map<const TauMetric*, const Metric*>  tmet2cmet;
map<const TauRegion*, const Region*>  treg2creg;
map<const TauCpath*, const Cnode*>    tcpath2ccnode;
map<const TauLoc*, const Thread*>     tloc2cthrd;

// call tree structure
map<const TauCpath*, const TauCpath*>          tcpath2parent;
map<const TauCpath*, vector<const TauCpath*> > tcpath2childv;


const TauCpath* get_parent_cpath(const TauProfile* tauprof, const TauCpath* cpath)
{
  const TauCpath* result = NULL;

  // is root?
  if ( cpath->size() == 1 )
    return NULL;
  for ( int i = 0; i < tauprof->get_ncpaths(); i++ )
    {
      if ( tauprof->get_cpath(i)->size() + 1 !=  cpath->size() )
	continue;
       
      for ( int j = 0; j < cpath->size(); j++ )
	  if ( j == cpath->size() - 1 )
	      result = tauprof->get_cpath(i);
	  else if ( *(*cpath)[j] != *(*tauprof->get_cpath(i))[j] )
	      break;
    }
  if ( result == NULL )
    {
      cerr << endl;
      cerr << "ERROR: Incomplete TAU call tree." << endl;
      cerr << "Only 1-level, 2-level or full call-path profiles supported." << endl;
      cerr << *cpath << endl;
      exit(EXIT_FAILURE);
    }
  return result;
}

void create_metric_dim(const TauProfile* tauprof, Cube* cube)
{
  for ( int i = 0; i < tauprof->get_nmets(); i++ )
    {
      string uom;
      if ( tauprof->get_met(i)->get_name() == "TIME" )
	uom = "sec";
      else
	uom = "occ";
      tmet2cmet[tauprof->get_met(i)] = cube->def_met(tauprof->get_met(i)->get_name(), uom, "", "", NULL);
    }
}

void create_call_tree(const TauCpath* cpath, Cube* cube)
{
  const TauRegion* callee = (*cpath)[cpath->size() - 1];
  tcpath2ccnode[cpath] = cube->def_cnode(treg2creg[callee], tcpath2ccnode[tcpath2parent[cpath]]); 
  cerr << "Call-tree node created" << endl;

  for ( int i = 0; i < tcpath2childv[cpath].size(); i++ )
    create_call_tree(tcpath2childv[cpath][i], cube);
}

void create_program_dim(const TauProfile* tauprof, Cube* cube)
{
  // identify parent and child nodes
  for ( int i = 0; i < tauprof->get_ncpaths(); i++ )
    {
      const TauCpath* child  = tauprof->get_cpath(i);
      const TauCpath* parent = get_parent_cpath(tauprof, child);

      tcpath2parent[child] = parent;
      tcpath2childv[parent].push_back(child); // tcpath2childv[NULL] = root vector
    }
  
  // create pseudo module
  const Module* pseudo_mod = cube->def_module("UNKNOWN");
  
  // create regions
  for ( int i = 0; i < tauprof->get_nregions(); i++ )
    treg2creg[tauprof->get_region(i)] = cube->def_region(tauprof->get_region(i)->get_name(), -1, -1, "", "", pseudo_mod);
  
  // create call tree
  tcpath2ccnode[NULL] = NULL; 

  // determine whether the profile is a true call-path profile
  bool is_flat = true;
  for ( int i = 0; i < tcpath2childv[NULL].size(); i++ )
    if ( ! tcpath2childv[tcpath2childv[NULL][i]].empty() )
      is_flat = false;

  vector<const TauCpath*> rootv;

  // if true call-path profile then eliminate all roots without children
  for ( int i = 0; i < tcpath2childv[NULL].size(); i++ )
    if ( is_flat )
	rootv.push_back(tcpath2childv[NULL][0]);
    else if ( ! tcpath2childv[tcpath2childv[NULL][i]].empty() )
	  rootv.push_back(tcpath2childv[NULL][i]);

  for ( int i = 0; i < rootv.size(); i++ )
    create_call_tree(rootv[i], cube);
}

void create_system_dim(const TauProfile* tauprof, Cube* cube)
{
  // create machine with one node
  const Node* node = cube->def_node("Node", cube->def_mach("Machine"));

  // create processes & threads
  map<int, const Process*> proc_id2cproc;

  for ( int i = 0; i < tauprof->get_nlocs(); i++ )
    {
      const TauLoc* loc = tauprof->get_loc(i);
      
      long proc_id = loc->get_proc_id();
      long thrd_id = loc->get_thrd_id();


      // create process if needed
      const Process* cproc = NULL;
      map<int, const Process*>::const_iterator pit = proc_id2cproc.find(proc_id);
      if ( pit != proc_id2cproc.end() )
	{
	  cproc = pit->second;
	}
      else
	{
	  cproc = cube->def_proc("", proc_id, node);
	  proc_id2cproc[proc_id] = cproc;
	}
      // create thread if needed
      tloc2cthrd[loc] = cube->def_thrd("", thrd_id, cproc);
    }
}

void enter_severity(const TauProfile* tauprof, Cube* cube)
{
  for ( int i = 0; i < tauprof->get_nmets(); i++ )
    for ( int j = 0; j < tauprof->get_ncpaths(); j++ )
      for ( int k = 0; k < tauprof->get_nlocs(); k++ )
	{
	  const TauMetric* met = tauprof->get_met(i);
	  const TauCpath* cpath = tauprof->get_cpath(j);
	  const TauLoc* loc = tauprof->get_loc(k);
	  const TauData* data = tauprof->get_data(met, cpath, loc);

	  double value;
	  
	  if ( data == NULL ) 
	    value = 0.0;
	  else if ( ! tcpath2childv[cpath].empty() )
	    // take inclusive value if call path has children
	    value = data->get_excl();
	  else 
	    // take inclusive value if call path is a leaf
	    value = data->get_incl();
	    
	  cube->set_sev(tmet2cmet[met],
			tcpath2ccnode[cpath],
			tloc2cthrd[loc],
			value);
	}
}


int main(int argc, char* argv[])
{
  string tau_profile_dir;
  string cube_file;
  /*
  // ckeck # of arguments 
  if ( argc > 3 )
    {
      cerr << "Wrong # of arguments\nUsage: tau2cube [<tau-profile-dir>] [<cube-file>]" << endl;
      exit(EXIT_FAILURE);
    }
  else if ( argc == 1 )
    {
      tau_profile_dir = ".";
      cube_file = "a.cube";
    }
  else if ( argc == 2 )
    {
      tau_profile_dir = argv[1];
      cube_file = "a.cube";
    }
  else 
    {
      tau_profile_dir = argv[1];
      cube_file = argv[2];
    }
  */
  if ( argc > 4 ) { 
    cerr << "Usage: tau2cube [<tau-profile-dir>] [-o <cube-file>]" << endl;
    exit(EXIT_FAILURE);
  }
  vector<string> inputs;
  string output;
  string cur;
  bool out_flag = false;
  // analyze the input command 
  for (int i = 1; i < argc; i++) {
    cur = argv[i];
    if (out_flag) {
      output = cur;
      out_flag = false;
      continue;
    }
    if (cur != "-o") {
      inputs.push_back(cur);
    } else {
      out_flag = true;
    }
  }
  
  if (inputs.size() == 0) tau_profile_dir = ".";
  else tau_profile_dir = inputs[0];
  if (!output.empty()) cube_file = output;
  else cube_file = "a.cube";
  
  // create TAU profile
  cerr << "Parsing TAU profile...          " << endl;
  TauProfile tauprof(tau_profile_dir.c_str());
  cerr << "Parsing TAU profile...          ";
  cerr << "done." << endl;

  // create CUBE profile
  cerr << "Creating CUBE profile...        ";
  Cube cube;

  // create metric dimension
  create_metric_dim(&tauprof, &cube);
  // create program dimension
  create_program_dim(&tauprof, &cube);
  // create system dimension
  create_system_dim(&tauprof, &cube);
  // enter severity data
  enter_severity(&tauprof, &cube); 

  // set aggregation mode
  cube.def_attr("CUBE_AGGR", "MAX");

  // write CUBE file
  ofstream file(cube_file.c_str());
  if ( !file )
   {
     cerr << endl << "ERROR: Cannot open\"" << cube_file << "\"" << endl;
     exit(EXIT_FAILURE);
   }
  file << cube;
  file.close();
  cerr << "done." << endl;
}
