/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <iostream>
#include <vector>

#include "elg_rw.h"

#include "Alock_rep.h"
#include "SeqTrace.h"
#include "Communicator.h"
#include "Error.h"
#include "Enter_rep.h"
#include "Exit_rep.h"
#include "Fork_rep.h"
#include "Join_rep.h"
#include "Location.h"
#include "Machine.h"
#include "MPICExit_rep.h"
#include "Node.h"
#include "OMPCExit_rep.h"
#include "Process.h"
#include "Recv_rep.h"
#include "Rlock_rep.h"
#include "Send_rep.h"
#include "Thread.h"
#include "Cartesian.h"

using namespace earl;
using namespace std;

#define MAXSTR 4096

SeqTrace::SeqTrace(string path) : grid(new Grid())
{ 
  if (!(file = ElgIn_open(path.c_str())))
    throw RuntimeError(string("Cannot open EPILOG trace file \"") + path + "\".");
  
  ElgRec* rec;
  bool done = false;
  
  while(!done)
    {
      if (!(rec = ElgIn_read_record(file)))
	throw FatalError("Invalid EPILOG trace file.");
      
      switch(ElgRec_get_type(rec))
	{ 
	case ELG_STRING :
	  {
	    long strid;
	    int cntc = 0;
	    char str[MAXSTR];
	    char* strptr = str;
	    vector<ElgRec*> cntvec;
	    
	    strid = ElgRec_read_ui4(rec);
	    cntc  = ElgRec_read_ui1(rec);
	    sprintf(strptr, "%.250s", ElgRec_read_string(rec));
	    strptr += 250;
	    
	    /* Are there any continuation rercords? */       
	    while(cntc)
	      {
		ElgRec* cntrec;
		
		if (!(cntrec = ElgIn_read_record(file)))
		  throw FatalError("Invalid EPILOG trace file.");
		sprintf(strptr, "%.255s", ElgRec_read_string(cntrec));
		ElgRec_free(cntrec);
		strptr += 255;
		cntc--;
	      }
	    strm[strid] = str;
	    break;
	  }
	case ELG_MACHINE :                
	  {
	    long mid, mnid;
	    elg_ui4 tmp;

	    mid   = ElgRec_read_ui4(rec);
	    // skip nodec
	    ElgRec_read_ui4(rec);
	    tmp = ElgRec_read_ui4(rec);
	    if ( tmp == ELG_NO_ID )
	      mnid = -1;
	    else
	      mnid = tmp;

	    grid->create_mach(mid, this->get_str(mnid));

	    break;
	  }
	case ELG_NODE :                   
	  {
	    long nid, mid, cpuc, nnid; 
	    double cr;
	    elg_ui4 tmp;

	    nid  = ElgRec_read_ui4(rec);
	    mid  = ElgRec_read_ui4(rec);
	    cpuc = ElgRec_read_ui4(rec);
	    tmp = ElgRec_read_ui4(rec);
	    if ( tmp == ELG_NO_ID )
	      nnid = -1;
	    else
	      nnid = tmp;
	    cr   = ElgRec_read_d8(rec);

	    grid->create_node(mid, nid, this->get_str(nnid), cpuc, cr);
	    break;
	  }
	case ELG_PROCESS :                
	  {
	    long pid, pnid; 
	    
	    pid  = ElgRec_read_ui4(rec);
	    pnid = ElgRec_read_ui4(rec);

	    grid->create_proc(pid, this->get_str(pnid));
	    break;
	  }
	case ELG_THREAD :                 
	  {
	    long tid, pid, tnid; 
	    
	    tid  = ElgRec_read_ui4(rec);
	    pid  = ElgRec_read_ui4(rec);
	    tnid = ElgRec_read_ui4(rec);

	    grid->create_thrd(pid, tid, this->get_str(tnid));
	    break;
	  }
	case ELG_LOCATION :              
	  {
	    long lid, mid, nid, tid, pid; 

	    lid  = ElgRec_read_ui4(rec);
	    mid  = ElgRec_read_ui4(rec);
	    nid  = ElgRec_read_ui4(rec);
	    pid  = ElgRec_read_ui4(rec);
	    tid  = ElgRec_read_ui4(rec);

	    grid->create_loc(lid, mid, nid, pid, tid); 
	    break;
	  }
	case ELG_FILE :                   
	  {
	    long fid, fnid; 
	    
	    fid  = ElgRec_read_ui4(rec);
	    fnid = ElgRec_read_ui4(rec);

	    filem[fid] = this->get_str(fnid);
	    break;
	  }
	case ELG_REGION :                 
	  {
	    long rid, rnid, fid, begln, endln, rdid, rtype;
	    string rtype_str;
	    elg_ui4 tmp;
	    
	    rid   = ElgRec_read_ui4(rec);
	    rnid  = ElgRec_read_ui4(rec);

	    tmp   = ElgRec_read_ui4(rec);
	    if ( tmp == ELG_NO_ID )
	      fid = -1;
	    else
	      fid = tmp;

	    tmp = ElgRec_read_ui4(rec);
	    if ( tmp == ELG_NO_LNO )
	      begln  = -1;
	    else
	      begln = tmp;

	    tmp = ElgRec_read_ui4(rec);
	    if ( tmp == ELG_NO_LNO )
	      endln  = -1;
	    else
	      endln = tmp;

	    tmp = ElgRec_read_ui4(rec);
	    if ( tmp == ELG_NO_ID )
	      rdid  = -1;
	    else
	      rdid = tmp;

	    rtype = ElgRec_read_ui1(rec);
	    
	    switch(rtype)
	      {
	      case ELG_UNKNOWN :    
		rtype_str = "UNKNOWN";
		break;
	      case ELG_FUNCTION :              
		rtype_str = "FUNCTION";
		break;
	      case ELG_LOOP :                  
		rtype_str = "LOOP";
		break;
	      case ELG_USER_REGION :                  
		rtype_str = "USER_REGION";
		break;
	      case ELG_OMP_PARALLEL :          
		rtype_str = "OMP_PARALLEL";
		break;
	      case ELG_OMP_LOOP :             
		rtype_str = "OMP_LOOP";
		break;
	      case ELG_OMP_SECTIONS :          
		rtype_str = "OMP_SECTIONS";
		break;
	      case ELG_OMP_SECTION :          
		rtype_str = "OMP_SECTION";
		break;
	      case ELG_OMP_WORKSHARE :          
		rtype_str = "OMP_WORKSHARE";
		break;
	      case ELG_OMP_SINGLE :           
		rtype_str = "OMP_SINGLE";
		break;
	      case ELG_OMP_MASTER :           
		rtype_str = "OMP_MASTER";
		break;
	      case ELG_OMP_CRITICAL :         
		rtype_str = "OMP_CRITICAL";
		break;
	      case ELG_OMP_ATOMIC :         
		rtype_str = "OMP_ATOMIC";
		break;
	      case ELG_OMP_BARRIER :          		
		rtype_str = "OMP_BARRIER";
		break;
	      case ELG_OMP_IBARRIER :          		
		rtype_str = "OMP_IBARRIER";
		break;
	      case ELG_OMP_FLUSH :          		
		rtype_str = "OMP_FLUSH";
		break;
	      case ELG_OMP_CRITICAL_SBLOCK :          		
		rtype_str = "OMP_CRITICAL_SBLOCK";
		break;
	      case ELG_OMP_SINGLE_SBLOCK :          		
		rtype_str = "OMP_SINGLE_SBLOCK";
		break;
	      default:
		throw FatalError("No such region type.");
	      }

	    regm[rid] = new Region(rid, this->get_str(rnid), this->get_file(fid), 
				      begln, endln, this->get_str(rdid), rtype_str);

	    break;
	  }
	case ELG_CALL_SITE :                 
	  {
	    long csid, fid, lno, erid;
	    
	    csid  = ElgRec_read_ui4(rec);
	    fid   = ElgRec_read_ui4(rec);
	    lno   = ElgRec_read_ui4(rec);
	    erid  = ElgRec_read_ui4(rec);
	    
	    csitem[csid] = new Callsite(csid, this->get_file(fid), lno, this->get_reg(erid));
	    break;
	  }
	case ELG_METRIC :                 
	  {
	    long metid, metnid, metdid, metdtype, metmode, metiv;
	    elg_ui4 tmp;

	    metid    = ElgRec_read_ui4(rec);
	    metnid   = ElgRec_read_ui4(rec);

	    tmp = ElgRec_read_ui4(rec);
	    if ( tmp == ELG_NO_ID )
	      metdid  = -1;
	    else
	      metdid = tmp;

	    metdtype = ElgRec_read_ui1(rec);
	    metmode  = ElgRec_read_ui1(rec);
	    metiv    = ElgRec_read_ui1(rec);
	    
	    // data-type check
	    if ( metdtype != ELG_INTEGER )
	      throw FatalError("This metric data type is not supported yet.");
	    
	    // mode check
	    if ( metmode != ELG_COUNTER )
	      throw FatalError("This metric mode is not supported yet.");
	    
	    // interval check
	    if ( metiv != ELG_START )
	      throw FatalError("This metric interval semantics is not supported yet.");
	    
	    metm[metid] = new Metric(metid, 
				     this->get_str(metnid), 
				     this->get_str(metdid),
				     "INTEGER", "COUNTER", "START");
	    
	    break;
	  }
	case ELG_MPI_COMM :              
	  {
	    long grpc, cid;
	    vector<long> groupv;
	    
	    cid  = ElgRec_read_ui4(rec);	  
	    grpc = ElgRec_read_ui4(rec);
	    
	    for (int i = 0; i < grpc; i++)
	      {
		elg_ui1 byte = ElgRec_read_ui1(rec);
		for (int j = 0; j < 8; j++)
		  if (byte & (1 << j))
		    groupv.push_back(i*8 + j);
	      }
	    grid->create_com(cid, groupv);
	    break;
	  }
	  /* topology specific definition records */   
	case ELG_CART_TOPOLOGY :
	  {
	    long          id, ndims;
	    vector<long>  dimv;
	    vector<bool>  periodv;
	    
	    id = ElgRec_read_ui4(rec);	   
	    ndims = ElgRec_read_ui1(rec);
	    	    
	    /* push the dimensions into dimv */	    
	    for (int i = 0; i < ndims ; i++)
	      {	
		dimv.push_back( ElgRec_read_ui4(rec));
	      }
	    
	    /* push the periods into periodv */
	    for (int i = 0; i < ndims ; i++)
	      {		 
		periodv.push_back(ElgRec_read_ui1(rec));	
	      }

	    /* create the topology object */
	    grid->create_cart(id, ndims, dimv, periodv);
	   
	    break;
	  }
	case ELG_CART_COORDS :
	  {
	    long               id, lid, ndims;	    
	    std::vector<long>  coordv;
	    Location*          loc;
	    Cartesian*          cart;
	    
	    id = ElgRec_read_ui4(rec);	   
	    lid = ElgRec_read_ui4(rec);	   
	    ndims = ElgRec_read_ui1(rec);
	    
	    /* push the coordinates in coordv */
	    for (int i = 0; i < ndims ; i++)
	      {	
		coordv.push_back(ElgRec_read_ui4(rec));
	      }  

	    /* get the Location corresponding to the lid */
	    loc = grid->get_loc(lid);

	    /* get the Cartesian corresponding to topid */
	    cart = grid->get_cart(id);

	    /* now fill in the loc2coords map of cartptr */
	    cart->set_coords(loc, coordv);
	    
	    break;
	  }
	case ELG_NUM_EVENTS :              
	  {
	    nevents = ElgRec_read_ui4(rec);
	    break;
	  }
	case ELG_LAST_DEF :              
	  {
	    done = true;
	    break;
	  }
	}
      ElgRec_free(rec);
    }
  // create regions for dump and log off

  dump   = new Region(regm.size(), "DUMP", "EPILOG", -1, -1, "Trace file dumped.", "EPILOG");
  regm[regm.size()] = dump;

  logoff = new Region(regm.size(), "LOG OFF", "EPILOG", -1, -1, "Event logging switched off.", "EPILOG");
  regm[regm.size()] = logoff;

  // create pseudo file "EPILOG"

  filem[filem.size()] = "EPILOG";
}


SeqTrace::~SeqTrace() 
{ 
  ElgIn_close(file);

  map<long, Region*>::const_iterator rit = regm.begin();
  while (rit != regm.end())
    {
      delete rit->second;
      rit++;
    }
  map<long, Callsite*>::const_iterator csit = csitem.begin();
  while (csit != csitem.end())
    {
      delete csit->second;
      csit++;
    }
  map<long, Metric*>::const_iterator metit = metm.begin();
  while (metit != metm.end())
    {
      delete metit->second;
      metit++;
    }
  delete grid;
}

Event 
SeqTrace::next_event(State& state, EventBuffer* buffer) 
{ 
  Event event;

  event = buffer->retrieve(state.get_pos() + 1);
  if ( !event.null() )
    {
      this->skip_event();
      event.trans(state);
      return event;
    }

  ElgRec* rec;
  bool done = false;
  
  while(!done)
    {
      if (!(rec = ElgIn_read_record(file)))
	return NULL;
      
      if (ElgRec_is_event(rec))
	  break;
      ElgRec_free(rec);
    }

  switch(ElgRec_get_type(rec))
    { 
    case ELG_ENTER :
      {
	long lid, rid;
	double time;
	vector<elg_ui8> metv;

	lid  = ElgRec_read_ui4(rec);
	time = ElgRec_read_d8(rec);
	rid  = ElgRec_read_ui4(rec);

	for (int i = 0 ; i < metm.size(); i++)
	  metv.push_back(ElgRec_read_ui8(rec));

	ElgRec_free(rec);

	event = new Enter_rep(state, 
			      buffer, 
			      grid->get_loc(lid), 
			      time, 
			      this->get_reg(rid), 
			      metm, 
			      metv,
			      NULL);
	break;
      }
    case ELG_ENTER_CS :
      {
	long lid, csid;
	double time;
	vector<elg_ui8> metv;

	lid  = ElgRec_read_ui4(rec);
	time = ElgRec_read_d8(rec);
	csid = ElgRec_read_ui4(rec);

	for (int i = 0 ; i < metm.size(); i++)
	  metv.push_back(ElgRec_read_ui8(rec));

	ElgRec_free(rec);

	Callsite* csite = this->get_csite(csid);

	event = new Enter_rep(state, 
			      buffer, 
			      grid->get_loc(lid), 
			      time, 
			      csite->get_callee(), 
			      metm, 
			      metv,
			      csite);
	break;
      }
    case ELG_EXIT :
      {
	long lid;
	double time;
	vector<elg_ui8> metv;

	lid  = ElgRec_read_ui4(rec);
	time = ElgRec_read_d8(rec);

	for (int i = 0 ; i < metm.size(); i++)
	  metv.push_back(ElgRec_read_ui8(rec));

	ElgRec_free(rec);

	event = new Exit_rep(state, 
			     buffer, 
			     grid->get_loc(lid), 
			     time, 
			     metm, 
			     metv);
	break;
      }
    case ELG_MPI_SEND :
      {
	long lid, tag, cid, len, dlid;
	double time;

	lid  = ElgRec_read_ui4(rec);
	time = ElgRec_read_d8(rec);
	dlid = ElgRec_read_ui4(rec);
	cid  = ElgRec_read_ui4(rec);
	tag  = ElgRec_read_ui4(rec);
	len  = ElgRec_read_ui4(rec);
	
	ElgRec_free(rec);
	
	event = new Send_rep(state, 
			     buffer, 
			     grid->get_loc(lid), 
			     time, 
			     grid->get_com(cid),
			     tag,
			     len,
			     grid->get_loc(dlid));
	break;
      }
    case ELG_MPI_RECV :
      {
	long lid, tag, cid, slid;
	double time;

	lid  = ElgRec_read_ui4(rec);
	time = ElgRec_read_d8(rec);
	slid = ElgRec_read_ui4(rec);
	cid  = ElgRec_read_ui4(rec);
	tag  = ElgRec_read_ui4(rec);

	ElgRec_free(rec);

	event = new Recv_rep(state, 
			     buffer, 
			     grid->get_loc(lid), 
			     time, 
			     grid->get_com(cid),
			     tag,
			     grid->get_loc(slid));

	break;
      }
    case ELG_MPI_COLLEXIT :
      {
	long lid, rtid, cid, sent, recvd;
	double time;
	vector<elg_ui8> metv;

	lid   = ElgRec_read_ui4(rec);
	time  = ElgRec_read_d8(rec);

	for (int i = 0 ; i < metm.size(); i++)
	  metv.push_back(ElgRec_read_ui8(rec));

	rtid  = ElgRec_read_ui4(rec);
	cid   = ElgRec_read_ui4(rec);
	sent  = ElgRec_read_ui4(rec);
	recvd = ElgRec_read_ui4(rec);

	ElgRec_free(rec);

	/* some collective operations don't have a root location */
	Location* root;
	if (rtid == ELG_NO_ID)
	  root = NULL;
	else
	  root = grid->get_loc(rtid);

	event = new MPICExit_rep(state, 
				 buffer, 
				 grid->get_loc(lid), 
				 time, 
				 metm, 
				 metv,
				 root,
				 grid->get_com(cid),
				 sent,
				 recvd);
	break;
      }
    case ELG_OMP_FORK : 
      {
	long lid;
	double time;
	
	lid   = ElgRec_read_ui4(rec);
	time  = ElgRec_read_d8(rec);
	
	ElgRec_free(rec);

	event = new Fork_rep(state, 
			     buffer, 
			     grid->get_loc(lid), 
			     time);	
	break;
      }
    case ELG_OMP_JOIN : 
      {
	long lid;
	double time;
	
	lid   = ElgRec_read_ui4(rec);
	time  = ElgRec_read_d8(rec);

	ElgRec_free(rec);
	
	event = new Join_rep(state, 
			     buffer, 
			     grid->get_loc(lid), 
			     time);	
	break;
      }
    case ELG_OMP_ALOCK : 
      {
	long lid, lkid;
	double time;

	lid   = ElgRec_read_ui4(rec);
	time  = ElgRec_read_d8(rec);
	lkid  = ElgRec_read_ui4(rec);
	
	ElgRec_free(rec);

	event = new Alock_rep(state, 
			      buffer, 
			      grid->get_loc(lid), 
			      time,
			      lkid);	
 	break;
     }
    case ELG_OMP_RLOCK :
      {
	long lid, lkid;
	double time;

	lid   = ElgRec_read_ui4(rec);
	time  = ElgRec_read_d8(rec);
	lkid  = ElgRec_read_ui4(rec);

	ElgRec_free(rec);

	event = new Rlock_rep(state, 
			      buffer, 
			      grid->get_loc(lid), 
			      time,
			      lkid);	
 	break;
     }
    case ELG_OMP_COLLEXIT :
      {
	long lid;
	double time;
	vector<elg_ui8> metv;
	
	lid  = ElgRec_read_ui4(rec);
	time = ElgRec_read_d8(rec);

	for (int i = 0 ; i < metm.size(); i++)
	  metv.push_back(ElgRec_read_ui8(rec));

	ElgRec_free(rec);

	event = new OMPCExit_rep(state, 
				 buffer, 
				 grid->get_loc(lid), 
				 time, 
				 metm, 
				 metv);
	break;
      }
     case ELG_LOG_OFF :
       {
	 long lid;
	 double time;
	 vector<elg_ui8> metv;
	 
	 lid  = ElgRec_read_ui4(rec);
	 time = ElgRec_read_d8(rec);
	 
	 for (int i = 0 ; i < metm.size(); i++)
	   metv.push_back(ElgRec_read_ui8(rec));

	 ElgRec_free(rec);
	 
	 event = new Enter_rep(state, 
			       buffer, 
			       grid->get_loc(lid), 
			       time, 
			       logoff, 
			       metm, 
			       metv,
			       NULL);
    	  break;
       }
     case ELG_LOG_ON :
       {
	 long lid;
	 double time;
	 vector<elg_ui8> metv;
	 
	 lid  = ElgRec_read_ui4(rec);
	 time = ElgRec_read_d8(rec);
	 
	 for (int i = 0 ; i < metm.size(); i++)
	   metv.push_back(ElgRec_read_ui8(rec));
	 
	 ElgRec_free(rec);
	 
	 event = new Exit_rep(state, 
			      buffer, 
			      grid->get_loc(lid), 
			      time, 
			      metm, 
			      metv);
    	  break;
       }
     case ELG_ENTER_DUMP :
       {
	long lid;
	double time;
	vector<elg_ui8> metv;

	lid  = ElgRec_read_ui4(rec);
	time = ElgRec_read_d8(rec);

	for (int i = 0 ; i < metm.size(); i++)
	  metv.push_back(ElgRec_read_ui8(rec));

	ElgRec_free(rec);

	event = new Enter_rep(state, 
			      buffer, 
			      grid->get_loc(lid), 
			      time, 
			      dump, 
			      metm, 
			      metv,
			      NULL);
    	  break;
       }
     case ELG_EXIT_DUMP :
       {
	 long lid;
	 double time;
	 vector<elg_ui8> metv;
	 
	 lid  = ElgRec_read_ui4(rec);
	 time = ElgRec_read_d8(rec);
	 
	 for (int i = 0 ; i < metm.size(); i++)
	   metv.push_back(ElgRec_read_ui8(rec));
	 
	 ElgRec_free(rec);
	 
	 event = new Exit_rep(state, 
			      buffer, 
			      grid->get_loc(lid), 
			      time, 
			      metm, 
			      metv);
    	  break;
       }
    default :
      {
	ElgRec_free(rec);
	throw FatalError("Error in SeqTrace::next_event().");
      }
    }
  event.trans(state);
  return event;
}
  
bool 
SeqTrace::skip_event() 
{ 
  if ( ElgIn_skip_record(file) != EOF )
    return true;
  else
    return false;
}

void       
SeqTrace::set_offset(const fpos_t* offset) 
{ 
  ElgIn_setpos(file, offset);
}
 
fpos_t
SeqTrace::get_offset() 
{ 
  fpos_t offset;

  ElgIn_getpos(file, &offset);
  return offset;
}
  
string   
SeqTrace::get_str(long strid) const
{
  map<long, string>::const_iterator it = strm.find(strid);
  
  if ( strid == -1 )
    return "UNKNOWN";

  if (it == strm.end())
    throw RuntimeError("Error in SeqTrace::get_str().");
  else
    return it->second;
}


string   
SeqTrace::get_file(long file_id) const 
{ 
  map<long, string>::const_iterator it = filem.find(file_id);

  if ( file_id == -1 )
    return "UNKNOWN";

  if (it == filem.end())
    throw RuntimeError("Error in SeqTrace::get_file().");
  else
    return it->second;
}
  
Region*       
SeqTrace::get_reg(long reg_id) const 
{ 
  map<long, Region*>::const_iterator it = regm.find(reg_id);

  if (it == regm.end())
    throw RuntimeError("Error in SeqTrace::get_reg().");
  else
    return it->second;
}
  
Callsite*     
SeqTrace::get_csite(long csite_id) const 
{ 
  map<long, Callsite*>::const_iterator it = csitem.find(csite_id);

  if (it == csitem.end())
    throw RuntimeError("Error in SeqTrace::get_csite().");
  else
    return it->second;
}

Metric*       
SeqTrace::get_met(long met_id) const 
{ 
  map<long, Metric*>::const_iterator it = metm.find(met_id);

  if (it == metm.end())
    throw RuntimeError("Error in SeqTrace::get_met().");
  else
    return it->second;
}
























