/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <stdlib.h>
#include <string.h>
#if (defined (ELG_OMPI) || defined (ELG_OMP))
  #include "opari_omp.h"
#endif
#include "elg_trc.h"
#define EPILOG
#include "elg_user.h"

#define ELG_User_start_U  ELG_USER_START
#define ELG_User_end_U    ELG_USER_END
#define ELG_User_start_L  elg_user_start
#define ELG_User_end_L    elg_user_end
#include "pomp_fwrapper_def.h"

/*
 *-----------------------------------------------------------------------------
 *
 *  EPILOG Library (Event Processing, Investigating, and Logging)
 *
 *  - Generic Compiler Interface
 * 
 *-----------------------------------------------------------------------------
 */

/*
 *-----------------------------------------------------------------------------
 * Simple hash table to map function names to region identifier
 *-----------------------------------------------------------------------------
 */

typedef struct HN {
  long id;            /* hash code (address of function name) */
  elg_ui4 elgid;      /* associated EPILOG region identifier  */
  struct HN* next;
} HashNode;

#define HASH_MAX 1021

static int gen_init = 1;       /* is initialization needed? */
static char *gen_main_id = 0;  /* name of main function */

static HashNode* htab[HASH_MAX];

/*
 * Stores EPILOG region identifier `e' under hash code `h'
 */

static void hash_put(long h, elg_ui4 e) {
  long id = h % HASH_MAX;
  HashNode *add = malloc(sizeof(HashNode));
  add->id = h;
  add->elgid = e;
  add->next = htab[id];
  htab[id] = add;
}

/*
 * Lookup hash code `h'
 * Returns EPILOG region identifier if already stored, otherwise ELG_NO_ID
 */

static elg_ui4 hash_get(long h) {
  long id = h % HASH_MAX;
  HashNode *curr = htab[id];
  while ( curr ) {
    if ( curr->id == h ) {
      return curr->elgid;
    }
    curr = curr->next;
  }
  return ELG_NO_ID;
}

/*
 * Register new region with EPILOG system
 */

static elg_ui4 register_region(char *name, char* file, int lno) {
  elg_ui4 rid;
  elg_ui4 fid;

  /* -- register file and region with EPILOG and store region identifier -- */
  fid = elg_def_file(file);
  rid = elg_def_region(name, fid, lno, ELG_NO_LNO, "USR", ELG_FUNCTION);
  hash_put((long) name, rid);
  return rid;
}

/*
 * This function is called at the entry of each function
 * C/C++ version
 */

void ELG_User_start(char* name, char *file, int lno) {
  elg_ui4 rid;

  /* -- if not yet initialized, initialize EPILOG -- */
  if ( gen_init ) {
    gen_init = 0;
    gen_main_id = name;
    elg_open();
  }

  /* -- get region identifier -- */
  if ( (rid = hash_get((long) name)) == ELG_NO_ID ) {
    /* -- region entered the first time, register region -- */
#   if defined (ELG_OMPI) || defined (ELG_OMP)
    if (omp_in_parallel()) {
#     pragma omp critical (elg_comp_generic_1)
      {
        if ( (rid = hash_get((long) name)) == ELG_NO_ID ) {
          rid = register_region(name, file, lno);
        }
      }
    } else {
      rid = register_region(name, file, lno);
    }
#   else
    rid = register_region(name, file, lno);
#   endif
  }

  /* -- write EPILOG enter record -- */
  elg_enter(rid);
}

/*
 * This function is called at the exit of each function
 * C/C++ version
 */

void ELG_User_end(char *name) {
  /* -- write EPILOG exit record -- */
  elg_exit();

  /* -- if main function, finalize EPILOG -- */
  if ( name == gen_main_id ) elg_close();
}

/*
 * This function is called at the entry of each function
 * Fortran version
 */

static char fnambuf[128];
static char ffilbuf[1024];

void FSUB(ELG_User_start)(char* name, char *file, int *lno, int nl, int fl) {
  elg_ui4 rid;

  /* -- convert Fortran to C strings -- */
  int namlen = ( nl < 128 ) ? nl : 127;
  int fillen = ( fl < 1024 ) ? fl : 1023;
  strncpy(fnambuf, name, namlen);
  fnambuf[namlen] = '\0';
  strncpy(ffilbuf, file, fillen);
  fnambuf[fillen] = '\0';

  /* -- if not yet initialized, initialize EPILOG -- */
  if ( gen_init ) {
    gen_init = 0;
    gen_main_id = name;
    elg_open();
  }

  /* -- get region identifier -- */
  if ( (rid = hash_get((long) name)) == ELG_NO_ID ) {
    /* -- region entered the first time, register region -- */
#   if defined (ELG_OMPI) || defined (ELG_OMP)
    if (omp_in_parallel()) {
#     pragma omp critical (elg_comp_generic_1)
      {
        if ( (rid = hash_get((long) name)) == ELG_NO_ID ) {
          rid = register_region(fnambuf, ffilbuf, *lno);
        }
      }
    } else {
      rid = register_region(fnambuf, ffilbuf, *lno);
    }
#   else
    rid = register_region(fnambuf, ffilbuf, *lno);
#   endif
  }

  /* -- write EPILOG enter record -- */
  elg_enter(rid);
}

/*
 * This function is called at the exit of each function
 * Fortran version
 */

void FSUB(ELG_User_end)(char *name, int nl) {
  /* -- write EPILOG exit record -- */
  elg_exit();

  /* -- if main function, finalize EPILOG -- */
  if ( name == gen_main_id ) elg_close();
}
