/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "elg_mpireq.h"
#include "elg_mpicom.h"
#include "elg_trc.h"

#include <stdio.h>
#include <stdlib.h>

/* 
 *-----------------------------------------------------------------------------
 *
 * Request management
 *
 *-----------------------------------------------------------------------------
 */

#define ELG_REQBLK_SIZE 10

struct ElgRequestBlock {
  struct ElgRequest req[ELG_REQBLK_SIZE];
  struct ElgRequestBlock *next;
  struct ElgRequestBlock *prev;
};

static struct ElgRequestBlock *head_block = 0;
static struct ElgRequestBlock *last_block = 0;
static struct ElgRequest *lastreq = 0;
static int lastidx = ELG_REQBLK_SIZE;

void elg_request_finalize()
{
  struct ElgRequestBlock *block;

  /* free request blocks */

  while (head_block) {
    block = head_block;
    head_block = head_block->next;
    free(block);
  }
}

void elg_request_create(MPI_Request request, 
			unsigned flags,
			int tag, 
			int dest, 
			int bytes, 
			MPI_Comm comm)
{
  struct ElgRequestBlock *new_block;

  lastidx++;
  if (lastidx >= ELG_REQBLK_SIZE) 
    {
      if (head_block == 0 ) 
	{
	  /* first time: allocate and initialize first block */
	  new_block = malloc(sizeof(struct ElgRequestBlock));
	  new_block->next = 0;
	  new_block->prev = 0;
	  head_block = last_block = new_block;
	} 
      else if (last_block == 0 ) 
	{
	  /* request list empty: re-initialize */
	  last_block = head_block;
	} 
      else 
	{
	  if (last_block->next == 0 ) 
	    {
	      /* request list full: expand */
	      new_block = malloc(sizeof(struct ElgRequestBlock));
	      new_block->next = 0;
	      new_block->prev = last_block;
	      last_block->next = new_block;
	    }
	  /* use next available block */
	  last_block = last_block->next;
	}
      lastreq = &(last_block->req[0]);
      lastidx  = 0;
    } 
  else 
    {
      lastreq++;
    }
  /* store request information */
  lastreq->request = request;
  lastreq->flags   = ERF_NONE;
  lastreq->flags  |= flags;
  lastreq->dest    = dest;
  lastreq->bytes   = bytes;
  lastreq->comm    = comm;
  lastreq->tag     = tag;
}

struct ElgRequest* elg_request_get(MPI_Request request)
{
  int i;
  struct ElgRequestBlock *block;
  struct ElgRequest *curr;

  /* list empty */
  if (!lastreq) return 0;

  /* search all requests in all blocks */
  block = head_block;
  while (block) {
    curr = &(block->req[0]);
    for (i = 0; i < ELG_REQBLK_SIZE; ++i) 
      {
	/* found? */
	if (curr->request == request) 
	  return curr;

	/* end of list? */
	if (curr == lastreq)
	  return 0;

	curr++;
      }
    block = block->next;
  }
  return 0;
}

void elg_request_free(struct ElgRequest* req)
{
  /* delete request by copying last request in place of req */ 
  if (!lastreq) {
    elg_error_msg("INTERNAL ERROR in request handling - no last request");
  }
  *req = *lastreq;
  lastreq->flags = ERF_NONE;
  lastreq->request = 0;

  /* adjust pointer to last request */
  lastidx--;
  if (lastidx < 0) 
    {
      /* reached low end of block */
      if (last_block->prev) 
	{
	  /* goto previous block if existing */
	  lastidx = ELG_REQBLK_SIZE-1;
	  lastreq = &(last_block->prev->req[lastidx]);
	} 
      else 
	{
	  /* no previous block: re-initialize */
	  lastidx = ELG_REQBLK_SIZE;
	  lastreq = 0;
	}
      last_block = last_block->prev;
    } 
  else 
    {
      lastreq--;
    }  
}

void elg_check_request(struct ElgRequest* req, MPI_Status *status)
{
  if (!req) return;

  /* if receive request, write receive trace record */
  if ((req->flags & ERF_RECV) && (status->MPI_SOURCE != MPI_PROC_NULL))
    elg_mpi_recv(ELG_RANK_TO_PE(status->MPI_SOURCE, req->comm),
                 ELG_COMM_ID(req->comm), status->MPI_TAG);
      
  if (req->flags & ERF_IS_PERSISTENT)
    {
      /* if persistent request, set to inactive,
         and, if requested delete request */
      req->flags &= ~ERF_IS_ACTIVE;
      if (req->flags & ERF_DEALLOCATE) elg_request_free(req);
    }
  else
    {
      /* if non-persistent request, delete always request */
      elg_request_free(req);
    }
}

static MPI_Request *orig_req_arr = 0;
static int orig_req_arr_size = 0;

void elg_save_request_array(MPI_Request *arr_req, int arr_req_size)
{
  int i;

  if (orig_req_arr_size == 0) {
    /* -- never used: initialize -- */
    orig_req_arr = malloc(arr_req_size * sizeof(MPI_Request));
    orig_req_arr_size = arr_req_size;
  } else if (arr_req_size > orig_req_arr_size) {
    /* -- not enough room: expand -- */
    orig_req_arr = realloc(orig_req_arr, arr_req_size * sizeof(MPI_Request));
    orig_req_arr_size = arr_req_size;
  }

  /* -- copy array -- */
  for (i=0; i<arr_req_size; ++i) orig_req_arr[i] = arr_req[i];
}

struct ElgRequest* elg_saved_request_get(int i)
{
  return elg_request_get(orig_req_arr[i]);
}
