/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "elg_trc.h"
#include "elg_gen.h"
#include "elg_env.h"
#include "elg_mpireg.h"
#include "elg_mrgthrd.h"
#include "elg_omplock.h"
#include "elg_metric.h"
#include "elg_pform.h"
#include "elg_error.h"

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if (defined (ELG_MPI) || defined (ELG_OMPI))
#include <mpi.h>
#include "elg_sync.h"
#endif
#if (defined (ELG_OMPI) || defined (ELG_OMP))
#include <omp.h>
extern void POMP_Init();

/* more robust version of omp_get_max_threads which
 * - can be called outside parallel regions
 * - before main (a problem on at least SUN and NEC)
 */
static int elg_get_max_threads() {
  static int my_max_threads = -1;
  char* val;

  if ( my_max_threads < 1 ) {
    if ( (my_max_threads = omp_get_max_threads()) < 1 ) {
      if ( (val = getenv("OMP_NUM_THREADS")) == NULL ) {
	my_max_threads = elg_pform_num_cpus();
      } else {
	my_max_threads = atoi(val);
      }
    }
  }
  return my_max_threads;
}
#endif

#if (defined (ELG_OMPI) || defined (ELG_OMP))
#define ELG_MY_THREAD   omp_get_thread_num() 
#define ELG_NUM_THREADS elg_get_max_threads() 
#else 
#define ELG_MY_THREAD   0
#define ELG_NUM_THREADS 1 
#endif

static ElgGen** genv;
static char**   tmp_namev;
static long     my_node;
static int      my_trace   = 0;
static int      num_traces = 1;
static int      num_nodes = 1;
static int      init_pid = -1;

#if (defined (ELG_METR))

/* number of performance metrics */
static int num_metrics = 0;

/* vector of per-thread arrays for counter values */
static elg_ui8**  valvv;

/* vector of per-thread arrays for event sets */
static struct elg_metv** metvv;

#endif

/* id counter starting with 1 */
static elg_ui4  curid  = 0;
static elg_ui4  elg_def_string(const char* str);

void elg_last_def( );

void elg_open()
{
  static int elg_open_called = 0;
  int i;
  char tmp[PATH_MAX];
  double clckrt = 0.0; 

  /* check for double initialisation error */
  if ( elg_open_called ) return;
  elg_open_called = 1;

  /* trace file creation */

  genv = calloc(ELG_NUM_THREADS, sizeof(ElgGen*));
  if ( genv == NULL )
    elg_error();
  tmp_namev = calloc(ELG_NUM_THREADS, sizeof(char*));
  if ( tmp_namev == NULL )
    elg_error();
  
  for (i = 0; i < ELG_NUM_THREADS; i++)
    {
      sprintf(tmp, "%s/%s.%d.%d.elg", elg_env_ldir(), elg_env_fprefix(), getpid(), i);
      tmp_namev[i] = calloc(strlen(tmp) + 1, sizeof(char*));
      if ( tmp_namev[i] == NULL )
	elg_error();
      strcpy(tmp_namev[i], tmp);

#if (defined (ELG_OMPI) || defined (ELG_OMP))

      if (i == 0)  
	/* master thread gets more buffer space */
	genv[i] = ElgGen_open(tmp_namev[i] , (elg_env_bsize() / 10) * 7);  
      else
	/* worker threads get less buffer space */
	genv[i] = ElgGen_open(tmp_namev[i] , (elg_env_bsize() / 10) * 3); 

#else

      genv[i] = ElgGen_open(tmp_namev[i] , elg_env_bsize());

#endif
    }
      
  /* initialization specific to this platform */
  elg_pform_init();

#if (defined (ELG_METR))

  /* initialize hardware counters */
  num_metrics = elg_metric_open();
  clckrt      = elg_metric_clckrt();

#endif

  /* machine and SMP-node identification */

  ElgGen_write_MACHINE(genv[0], 0, elg_pform_num_nodes(), elg_def_string(elg_pform_name()));  

  my_node = elg_pform_node_id();
  ElgGen_write_NODE(genv[0], my_node, 0, elg_pform_num_cpus(), elg_def_string(elg_pform_node_name()), clckrt);  
    
  /* register MPI and OpenMP API routines if necessary */
     
#ifndef ELG_CSITE_INST
#if (defined (ELG_OMPI) || defined (ELG_MPI))
  elg_mpi_register();
#endif
#endif
#if (defined (ELG_OMPI) || defined (ELG_OMP))
  POMP_Init();
#endif
  atexit(elg_close);
  init_pid = getpid();


#if (defined (ELG_METR))

  /* return if no counters requested */
  if ( num_metrics == 0 )
    return;
    
  /* write metric-description records */
  for ( i = 0; i < num_metrics; i++ )
    ElgGen_write_METRIC(genv[0],
			i,
			elg_def_string(elg_metric_name(i)),
			elg_def_string(elg_metric_descr(i)),
			elg_metric_dtype(i),
			elg_metric_mode(i),
			elg_metric_iv(i));
  
  /* initialize per-thread arrays for counter values */
  /* note: calloc initilaizes with zero bytes */
  
  valvv = calloc(ELG_NUM_THREADS, sizeof(elg_ui8*));
  if ( valvv == NULL )
    elg_error();
  for (i = 0; i < ELG_NUM_THREADS; i++)
    {
      valvv[i] = calloc(num_metrics, sizeof(elg_ui8));
      if ( valvv[i] == NULL )
	elg_error();
    }
#if (defined (ELG_OMPI) || defined (ELG_OMP)) 
  /* initialize thread support */
  elg_metric_thread_init(omp_get_thread_num);
#endif
  
  /* initialize per-thread event sets */
  metvv = calloc(ELG_NUM_THREADS, sizeof(struct elg_metv*));
  if ( metvv == NULL )
    elg_error();
#if (defined (ELG_OMPI) || defined (ELG_OMP)) 
#pragma omp parallel
  {
#endif
    metvv[ELG_MY_THREAD] = elg_metric_create();
#if (defined (ELG_OMPI) || defined (ELG_OMP)) 
  }
#endif
#endif
  return;
}

void elg_flush()
{
  ElgGen_flush(genv[ELG_MY_THREAD]);
}

void elg_close()
{  
  static int elg_close_called = 0;
  int i, nf;
  char* local_name;

  /* catch elg_close called from child processes through atexit */
  if ( init_pid != getpid() ) return;

  /* check for double finalization error */
  if ( elg_close_called ) return;
  elg_close_called = 1;

  /* write ELG_LOCATION record */
  for (i = 0; i < ELG_NUM_THREADS; i++)
    ElgGen_write_LOCATION(genv[i], 0, 0, my_node, my_trace, i);

  /* write ELG_LAST_DEF record */  
  elg_last_def();

  /* close trace files */
  for (i = 0; i < ELG_NUM_THREADS; i++)
      ElgGen_close(genv[i]);

  /* call cleanup functions */

#if (defined (ELG_OMPI) || defined (ELG_OMP)) 
  /* OpenMP locks */
  elg_lock_close();
#endif
  /* hardware counters */
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
#if (defined (ELG_OMPI) || defined (ELG_OMP)) 
#pragma omp parallel
      {
#endif
	elg_metric_free(metvv[ELG_MY_THREAD]);
#if (defined (ELG_OMPI) || defined (ELG_OMP)) 
      }
#endif
      elg_metric_close();
    }
#endif

       
  local_name = calloc(strlen(elg_env_gdir()) +
                      strlen(elg_env_fprefix()) + 16, sizeof(char));
  if (local_name == NULL )
    elg_error();
  sprintf(local_name, "%s/%s.%d.elg",
          elg_env_gdir(), elg_env_fprefix(), my_trace);

  /* merge per-thread traces and copy per-process trace
     from local directory to current working directory */
  elg_mrgthrd(local_name, ELG_NUM_THREADS, tmp_namev);

  /* delete temporary trace files and free temporary file names */
  for (i = 0; i < ELG_NUM_THREADS; i++)
    {
      int result;

      result = remove(tmp_namev[i]);
      if (result == 0)
	elg_cntl_msg("Removed trace file %s", tmp_namev[i]);
      else
	elg_error_msg("Cannot remove trace file %s", tmp_namev[i]);
      free(tmp_namev[i]);
    }

  free(genv);
  free(tmp_namev);

  /*- merge trace files -*/

  if (my_trace != 0) {
    free(local_name);
    return;
  }
  
  if (elg_env_do_merge())
    {
      int len = strlen(elg_env_gdir()) + strlen(elg_env_fprefix()) + 16;
      char *cmd;

      /*- wait for files to be ready -*/
      for (i = 0; i < num_traces; i++)
        {
          sprintf(local_name, "%s/%s.%d.elg", elg_env_gdir(),
	                                      elg_env_fprefix(), i);
          elg_cntl_msg("Checking for %s ...", local_name);
	  nf = 0;
          while (access(local_name, R_OK) != 0 )
	    {
	      ++nf;
	      /*- if file not ready in 15 sec give up -*/
	      if ( nf > 15 ) return;
	      sleep(1);
            }
        }

      /*- do actual merge -*/
#ifdef BINDIR
      if (access(BINDIR "/elg_merge", X_OK) == 0 )
        {
	  cmd = calloc(strlen(BINDIR) + 16 + len, sizeof(char));
	  if ( cmd == NULL )
	    elg_error();
          sprintf(cmd, "%s/elg_merge %d %s/%s %d", BINDIR, num_traces,
	          elg_env_gdir(), elg_env_fprefix(), num_nodes);
        }
      else
        {
#endif
	  cmd = calloc(10 + len, sizeof(char));
	  if ( cmd == NULL )
	    elg_error();
          sprintf(cmd, "elg_merge %d %s/%s", num_traces, elg_env_gdir(),
	          elg_env_fprefix());
#ifdef BINDIR
        }
#endif
      elg_cntl_msg(cmd);
      system(cmd);
      free(cmd);
    }
  free(local_name);
}

#if (defined (ELG_MPI) || defined (ELG_OMPI))
static int longcmp(const void *a, const void *b)
{
  long x = *(long*)a;
  long y = *(long*)b;
  return x-y;
}
#endif

void elg_mpi_init()
{
#if (defined (ELG_MPI) || defined (ELG_OMPI))

  double ltime, offset;

  PMPI_Comm_rank(MPI_COMM_WORLD, &my_trace);
  PMPI_Comm_size(MPI_COMM_WORLD, &num_traces);
  
  /* 1. clock synchronization if necessary*/
  if (num_traces > 1 && !elg_pform_is_gclock())
    {
      offset = elg_offset(&ltime, MPI_COMM_WORLD);
      /* write OFFSET record */  
      ElgGen_write_OFFSET(genv[0], ltime, offset);
    }
#endif

  elg_pform_mpi_init();
}

void elg_mpi_finalize( )
{
#if (defined (ELG_MPI) || defined (ELG_OMPI))

  double ltime, offset;
  long* nodeids;
  long lastid;
  int i;

  /* 2. clock synchronization if necessary */ 
  if (num_traces > 1 && !elg_pform_is_gclock())
    {
      offset = elg_offset(&ltime, MPI_COMM_WORLD);
      /* write OFFSET record */
      ElgGen_write_OFFSET(genv[0], ltime, offset);
    }

  /* determine number of nodes */
  if (my_trace == 0)
    {
      nodeids = malloc(num_traces * sizeof(long));
      if ( nodeids == NULL )
	elg_error();
    }
  PMPI_Gather(&my_node, 1, MPI_LONG, nodeids, 1, MPI_LONG, 0, MPI_COMM_WORLD);
  if (my_trace == 0)
    {
      qsort(nodeids, num_traces, sizeof(long), longcmp);
      lastid=nodeids[0];
      for (i=1; i<num_traces; ++i)
        {
          if ( nodeids[i] != lastid )
	    {
              lastid = nodeids[i];
              num_nodes++;
            }
        }
      free(nodeids);
    }
  PMPI_Barrier(MPI_COMM_WORLD);
#endif

  elg_pform_mpi_finalize();
}


/*
 *-----------------------------------------------------------------------------
 * Defining source code entities
 *-----------------------------------------------------------------------------
 */


static elg_ui4 elg_def_string(const char* str)
{
  int stringlen = strlen(str) + 1;
  int strid = curid++;

  if (stringlen <= 250)
    {
      ElgGen_write_STRING(genv[ELG_MY_THREAD], strid, 0, str);
    }
  else
    {
      int i;
      int restlen = stringlen - 250;
      elg_ui1 cntc  = (restlen / 255) + (restlen % 255 ? 1 : 0);

      ElgGen_write_STRING(genv[ELG_MY_THREAD], strid, cntc, str);
      str += 250; 
      
      for (i = 1; i < cntc; i++)
	{
	  ElgGen_write_STRING_CNT(genv[ELG_MY_THREAD], str);
	  str += 255;
	}
      /* last contination record */
      ElgGen_write_STRING_CNT(genv[ELG_MY_THREAD], str);
    }
  return strid;  
}

elg_ui4 elg_def_file(const char* fname)
{
  elg_ui4 fid = curid++;

  ElgGen_write_FILE(genv[ELG_MY_THREAD], fid, elg_def_string(fname));
  return fid;
}

elg_ui4 elg_def_region(const char* rname,
		       elg_ui4 fid,
		       elg_ui4 begln,
		       elg_ui4 endln,
		       const char* rdesc,
		       elg_ui1 rtype)
{
  elg_ui4 rid  = curid++;

  ElgGen_write_REGION(genv[ELG_MY_THREAD],
		      rid,
		      elg_def_string(rname),
		      fid,
		      begln,
		      endln,
		      elg_def_string(rdesc),
		      rtype);
  return rid;
}

elg_ui4 elg_def_csite(elg_ui4 fid,
		      elg_ui4 lno,
		      elg_ui4 erid, 
		      elg_ui4 lrid)
{
  elg_ui4 csid  = curid++;

  if ( lrid != ELG_NO_ID ) fprintf(stderr, ">>>>>>>>>>>> lrid != ELG_NO_ID\n");

  ElgGen_write_CALL_SITE(genv[ELG_MY_THREAD],
			 csid,
			 fid,
			 lno,
			 erid,
			 lrid);
  return csid;
}

void elg_mpi_com(elg_ui4 cid,
		 elg_ui4 grpc,
		 elg_ui1 grpv[])
{
  ElgGen_write_MPI_COMM(genv[0], cid, grpc, grpv);  
}


/* 
   defines a record of type ELG_CART_TOPOLOGY 
*/

void elg_def_cart(   elg_ui4 topid,
		     elg_ui1 ndims,
		     elg_ui4 dimv[],
		     elg_ui1 periodv[] )
{   
  ElgGen_write_CART_TOPOLOGY(genv[ELG_MY_THREAD], topid, ndims, dimv, periodv);   
}

/* 
    each process calls this function to write its coordinates
    as defined by the topology
*/

void elg_def_coords(elg_ui4 topid,
		    elg_ui1 ndims,
		    elg_ui4 coordv[])
{
  ElgGen_write_CART_COORDS(genv[ELG_MY_THREAD], topid, 0, ndims, coordv);
}

void elg_last_def( )
{
  ElgGen_write_LAST_DEF(genv[0]);  
}

/*
 *-----------------------------------------------------------------------------
 * Recording events
 *-----------------------------------------------------------------------------
 */

void elg_enter(elg_ui4 rid)
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_ENTER(genv[ELG_MY_THREAD], rid, num_metrics, valvv[ELG_MY_THREAD]);
    }
  else
    {
      ElgGen_write_ENTER(genv[ELG_MY_THREAD], rid, 0, NULL);
    }
#else
  ElgGen_write_ENTER(genv[ELG_MY_THREAD], rid, 0, NULL);
#endif
}

void elg_enter_cs(elg_ui4 csid)
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_ENTER_CS(genv[ELG_MY_THREAD], csid, num_metrics, valvv[ELG_MY_THREAD]);
    }
  else
    {
      ElgGen_write_ENTER_CS(genv[ELG_MY_THREAD], csid, 0, NULL);
    }
#else
  ElgGen_write_ENTER_CS(genv[ELG_MY_THREAD], csid, 0, NULL);
#endif
}

void elg_exit()
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_EXIT(genv[ELG_MY_THREAD], num_metrics, valvv[ELG_MY_THREAD]);
    }
  else
    {
      ElgGen_write_EXIT(genv[ELG_MY_THREAD], 0, NULL);
    }
#else
  ElgGen_write_EXIT(genv[ELG_MY_THREAD], 0, NULL);
#endif
}

void elg_mpi_send(elg_ui4 dpid,
		  elg_ui4 cid,
		  elg_ui4 tag,      
		  elg_ui4 sent)
{
  ElgGen_write_MPI_SEND(genv[ELG_MY_THREAD], dpid, cid, tag, sent);
}

void elg_mpi_recv(elg_ui4 spid,
		  elg_ui4 cid,
		  elg_ui4 tag)
{ 
  ElgGen_write_MPI_RECV(genv[ELG_MY_THREAD], spid, cid, tag);
}

void elg_mpi_collexit(elg_ui4 rpid,
		      elg_ui4 cid,
		      elg_ui4 sent,
		      elg_ui4 recvd)
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_MPI_COLLEXIT(genv[ELG_MY_THREAD], num_metrics, valvv[ELG_MY_THREAD], rpid, cid, sent, recvd);
    }
  else
    {
      ElgGen_write_MPI_COLLEXIT(genv[ELG_MY_THREAD], 0, NULL, rpid, cid, sent, recvd);
    }
#else
  ElgGen_write_MPI_COLLEXIT(genv[ELG_MY_THREAD], 0, NULL, rpid, cid, sent, recvd);
#endif
}


void elg_omp_fork()
{
#if (defined (ELG_OMPI) || defined (ELG_OMP))
  if (!omp_in_parallel()) 
    ElgGen_write_OMP_FORK(genv[ELG_MY_THREAD]);
#endif
}

void elg_omp_join()
{
#if (defined (ELG_OMPI) || defined (ELG_OMP))
  if (!omp_in_parallel()) 
    ElgGen_write_OMP_JOIN(genv[ELG_MY_THREAD]);
#endif
}

void elg_omp_alock(elg_ui4 lkid)
{
  ElgGen_write_OMP_ALOCK(genv[ELG_MY_THREAD], lkid);
}

void elg_omp_rlock(elg_ui4 lkid)
{
  ElgGen_write_OMP_RLOCK(genv[ELG_MY_THREAD], lkid);  
}

void elg_omp_collexit()
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_OMP_COLLEXIT(genv[ELG_MY_THREAD], num_metrics, valvv[ELG_MY_THREAD]);
    }
  else
    {
      ElgGen_write_OMP_COLLEXIT(genv[ELG_MY_THREAD], 0, NULL);
    }
#else
  ElgGen_write_OMP_COLLEXIT(genv[ELG_MY_THREAD], 0, NULL);
#endif
}


void elg_log_off()
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_LOG_OFF(genv[ELG_MY_THREAD], num_metrics, valvv[ELG_MY_THREAD]);
    }
  else
    {
      ElgGen_write_LOG_OFF(genv[ELG_MY_THREAD], 0, NULL);
    }
#else
  ElgGen_write_LOG_OFF(genv[ELG_MY_THREAD], 0, NULL);
#endif
}

void elg_log_on()
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_LOG_ON(genv[ELG_MY_THREAD], num_metrics, valvv[ELG_MY_THREAD]);
    }
  else
    {
      ElgGen_write_LOG_ON(genv[ELG_MY_THREAD], 0, NULL);
    }
#else
  ElgGen_write_LOG_ON(genv[ELG_MY_THREAD], 0, NULL);
#endif
}

void elg_enter_dump()
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_ENTER_DUMP(genv[ELG_MY_THREAD], num_metrics, valvv[ELG_MY_THREAD]);
    }
  else
    {
      ElgGen_write_ENTER_DUMP(genv[ELG_MY_THREAD], 0, NULL);
    }
#else
  ElgGen_write_ENTER_DUMP(genv[ELG_MY_THREAD], 0, NULL);
#endif
}

void elg_exit_dump()
{
#if (defined (ELG_METR))
  if ( num_metrics > 0 )
    {
      elg_metric_read(metvv[ELG_MY_THREAD], valvv[ELG_MY_THREAD]);
      ElgGen_write_EXIT_DUMP(genv[ELG_MY_THREAD], num_metrics, valvv[ELG_MY_THREAD]);
    }
  else
    {
      ElgGen_write_EXIT_DUMP(genv[ELG_MY_THREAD], 0, NULL);
    }
#else
  ElgGen_write_EXIT_DUMP(genv[ELG_MY_THREAD], 0, NULL);
#endif
}




















