/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (C) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "Pattern.h"

#include <algorithm>
#include <cstdlib>
#include <earl.h>
#include <string>

#include "Analyzer.h"
#include "CallTree.h"
#include "Profile.h"

using namespace earl;
using namespace std;

/****************************************************************************
* Declaration of auxiliary functions                                                                                    
****************************************************************************/

bool is_mpi_appl(const EventTrace* trace);
bool is_mpi_api(const Region* region);
bool is_mpi_barrier(const Region* region);
bool is_mpi_block_send(const Region* region); // potentially blocking send
bool is_mpi_collcom(const Region* region);
bool is_mpi_io(const Region* region);
bool is_mpi_12n(const Region* region);
bool is_mpi_n21(const Region* region);
bool is_mpi_n2n(const Region* region);
bool is_mpi_p2p(const Region* region);
bool is_mpi_recv(const Region* region);
bool is_mpi_testx(const Region* region);
bool is_mpi_waitall(const Region* region);
bool is_mpi_waitsome(const Region* region);
bool is_omp_appl(const EventTrace* trace);
bool is_omp_api(const Region* region); 
bool is_omp_lock(const Region* region); 
bool is_omp_barrier(const Region* region); 
bool is_omp_ibarrier(const Region* region);
bool is_omp_ebarrier(const Region* region); 
bool is_omp_flush(const Region* region); 
bool is_omp_critical(const Region* region); 
bool is_omp_critical_sblock(const Region* region); 
bool is_omp_parallel(const Region* region); 
bool is_function(const Region* region); 

int  get_met_id(string met_name, const EventTrace* trace);
bool is_met_avail(string met_name, const EventTrace* trace);

string lower(string str);

/****************************************************************************
* Pattern creation and deletion                                                                                      
****************************************************************************/

void get_patternv(std::vector<Pattern*>& pv, EventTrace* trace)
{
  pv.push_back(new Time(trace));
  pv.push_back(new Execution(trace));
  pv.push_back(new IdleThreads(trace));
  pv.push_back(new MPI(trace));
  pv.push_back(new MPICommunication(trace));
  pv.push_back(new MPIIO(trace));
  pv.push_back(new MPISynchronization(trace));
  pv.push_back(new MPIBarrierCompletion(trace));
  pv.push_back(new MPIWaitAtBarrier(trace));
  pv.push_back(new Collective(trace));
  pv.push_back(new EarlyReduce(trace));
  pv.push_back(new LateBroadcast(trace));
  pv.push_back(new WaitAtN2N(trace));
  pv.push_back(new PointToPoint(trace));
  pv.push_back(new LateReceiver(trace));
  pv.push_back(new LateSender(trace));
  pv.push_back(new MsgsWrongOrderLS(trace));
  pv.push_back(new MsgsWrongOrderLR(trace));
  pv.push_back(new OMP(trace));
  pv.push_back(new OMPFlush(trace));
  pv.push_back(new OMPFork(trace));
  pv.push_back(new OMPSynchronization(trace));
  pv.push_back(new OMPBarrier(trace));
  pv.push_back(new OMPEBarrier(trace));
  pv.push_back(new OMPWaitAtEBarrier(trace));
  pv.push_back(new OMPIBarrier(trace));
  pv.push_back(new OMPWaitAtIBarrier(trace));
  pv.push_back(new OMPLock(trace));
  pv.push_back(new OMPAPILock(trace));
  pv.push_back(new OMPCriticalLock(trace));
  // Non-time-based HWC metrics
  pv.push_back(new FloatingPoint(trace));
  pv.push_back(new L1DataCache(trace));
}

/****************************************************************************
* Time                                                                                      
****************************************************************************/

void   
Time::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;

  ctree.get_keyv(cnode_idv);

  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	{
	  const Location* loc = trace->get_loc(loc_id);

	  // if slave thread outside a parallel region 
	  if ( profile.is_multhrd(cnode_id, loc_id) == false && loc->get_thrd()->get_thrd_id() != 0 )
	    {
	      long   master_id  = loc->get_proc()->get_thrd(0)->get_loc()->get_id();
	      double master_val = profile.get_time(cnode_id, master_id);

	      severity.set_val(cnode_id, loc_id, master_val);
	    }
	  else
	    {
	      severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
	    }
	}
    }
}

/****************************************************************************
* Execution                                                                                      
****************************************************************************/

void   
Execution::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;

  ctree.get_keyv(cnode_idv);

  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* Idle Threads                                                                                    
****************************************************************************/

bool 
IdleThreads::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
IdleThreads::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;

  ctree.get_keyv(cnode_idv);

  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	{
	  const Location* loc = trace->get_loc(loc_id);
	  
	  // if slave thread outside a parallel region 
	  if ( profile.is_multhrd(cnode_id, loc_id) == false && loc->get_thrd()->get_thrd_id() != 0 )
	    {
	      long   master_id  = loc->get_proc()->get_thrd(0)->get_loc()->get_id();
	      double master_val = profile.get_time(cnode_id, master_id);

	      severity.set_val(cnode_id, loc_id, master_val);
	    }
	}
    }
}

/****************************************************************************
*
* Hardware Counter                                                                                    
*
****************************************************************************/

/****************************************************************************
* Low Floating Point                                                                                   
****************************************************************************/

bool 
FloatingPoint::isenabled() 
{
  return is_met_avail("FLOATING_POINT", trace);
}                                

void 
FloatingPoint::end(Profile& profile, CallTree& ctree)
{
  int met_id = get_met_id("FLOATING_POINT", trace);
  
  vector<long> cnode_idv;
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	severity.set_val(cnode_id, loc_id, profile.get_metval(met_id, cnode_id, loc_id));
    }  
}

/****************************************************************************
* L1 Data Cache                                                                                   
****************************************************************************/

bool 
L1DataCache::isenabled() 
{ 
  return is_met_avail("L1_D_MISS", trace);
}                                

void 
L1DataCache::end(Profile& profile, CallTree& ctree)
{
  int met_id = get_met_id("L1_D_MISS", trace);
  
  vector<long> cnode_idv;
  ctree.get_keyv(cnode_idv);

  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	severity.set_val(cnode_id, loc_id, profile.get_metval(met_id, cnode_id, loc_id));
    }  
}

/****************************************************************************
*
* MPI                                                                                  
*
****************************************************************************/

/****************************************************************************
* MPI                                                                                    
****************************************************************************/

bool 
MPI::isenabled() 
{ 
  return is_mpi_appl(trace);
}                                

void 
MPI::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_event(MPICEXIT, this);
} 

void 
MPI::end(Profile& profile, CallTree& ctree) 
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_mpi_api(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

cmp_event* 
MPI::mpicexit(earl::Event& event) 
{ 
  vector<Event> coll;

  trace->mpicoll(coll, event.get_pos());

  // if collective operation complete
  if ( coll.size() > 0 )
    {
      mpi_coll* mc = new mpi_coll();
      
      mc->region = event.get_reg();
      mc->exitv  = coll;
      for ( long i = 0; i < mc->exitv.size(); i++ )
	{
	  mc->enterv.push_back(mc->exitv[i].get_enterptr());
	  if ( event.get_root() == mc->exitv[i].get_loc() )
	    mc->root_idx = i;
	}
      return mc;
    }
  return NULL;
}

/****************************************************************************
* MPI Communication                                                                                    
****************************************************************************/

bool 
MPICommunication::isenabled()
{ 
  return is_mpi_appl(trace);
}
                                
void 
MPICommunication::end(Profile& profile, CallTree& ctree) 
{ 
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_mpi_p2p(ctree.get_callee(cnode_id)) || is_mpi_collcom(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* MPI IO                                                                                    
****************************************************************************/

bool 
MPIIO::isenabled()
{
  return is_mpi_appl(trace);
}                                

void 
MPIIO::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_mpi_io(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* MPI Synchronization                                                                                    
****************************************************************************/

bool 
MPISynchronization::isenabled()
{
  return is_mpi_appl(trace);
}                                

void 
MPISynchronization::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_mpi_barrier(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* Wait at Barrier (MPI)                                                                            
****************************************************************************/

bool 
MPIWaitAtBarrier::isenabled()
{
  return is_mpi_appl(trace);
}                                

void 
MPIWaitAtBarrier::regcb(Analyzer& analyzer)
{
  analyzer.subscribe_cmpnd(MPI_COLL, this);
} 

cmp_event* 
MPIWaitAtBarrier::cmpnd(const cmp_event* ce)
{
  const mpi_coll* mc = (const mpi_coll*) ce;
 
  Event enter = mc->enterv[0];
  if ( is_mpi_barrier(enter.get_reg()) )
    {
      vector<double> start_timev;
      
      for ( int i = 0; i < mc->enterv.size(); i++ )
	start_timev.push_back(mc->enterv[i].get_time());
      double max_start_time = *max_element(start_timev.begin(), start_timev.end());
      for ( int i = 0; i < mc->enterv.size(); i++ )
	{
	  long loc_id       = mc->enterv[i].get_loc()->get_id();
	  long cnode_id     = mc->enterv[i].get_cnodeptr().get_pos();
	  double idle_time = max_start_time - start_timev[i];
	  
	  severity.add_val(cnode_id, loc_id, idle_time);	      
	}
    }
  return NULL;
}   

/****************************************************************************
* Barrier Completion (MPI)                                                                            
****************************************************************************/

bool 
MPIBarrierCompletion::isenabled()
{
  return is_mpi_appl(trace);
}                                

void 
MPIBarrierCompletion::regcb(Analyzer& analyzer)
{
  analyzer.subscribe_cmpnd(MPI_COLL, this);
} 

cmp_event* 
MPIBarrierCompletion::cmpnd(const cmp_event* ce)
{
  const mpi_coll* mc = (const mpi_coll*) ce;
 
  Event enter = mc->enterv[0];
  if ( is_mpi_barrier(enter.get_reg()) )
    {
      vector<double> end_timev;
      
      for ( int i = 0; i < mc->exitv.size(); i++ )
	end_timev.push_back(mc->exitv[i].get_time());
      double min_end_time = *min_element(end_timev.begin(), end_timev.end());
      for ( int i = 0; i < mc->exitv.size(); i++ )
	{
	  long loc_id      = mc->enterv[i].get_loc()->get_id();
	  long cnode_id    = mc->enterv[i].get_cnodeptr().get_pos();
	  double comp_time = end_timev[i] - min_end_time;
	  
	  severity.add_val(cnode_id, loc_id, comp_time);	      
	}
    }
  return NULL;
}   

/****************************************************************************
* MPI Point-to-Point                                                                                   
****************************************************************************/

bool 
PointToPoint::isenabled() 
{ 
  return is_mpi_appl(trace);
}                                

void 
PointToPoint::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_event(RECV, this);
} 

void 
PointToPoint::end(Profile& profile, CallTree& ctree) 
{ 
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      // if callee P2P function
      if ( is_mpi_p2p(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

cmp_event* 
PointToPoint::recv(earl::Event& event) 
{ 
  recv_op* ro = new recv_op();
  
  ro->recv = event;
  ro->enter_recv = event.get_enterptr();
  ro->region = ro->enter_recv.get_reg();
  return ro;
}

/****************************************************************************
* MPI Late Sender                                                                                   
****************************************************************************/

bool 
LateSender::isenabled() 
{ 
  return is_mpi_appl(trace);
}                                

void 
LateSender::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_cmpnd(RECV_OP, this);
} 

cmp_event* 
LateSender::cmpnd(const cmp_event* ce) 
{ 
  const recv_op* ro = (const recv_op*) ce;
  
  if ( ! ( is_mpi_testx(ro->region) ||
	   is_mpi_waitall(ro->region) || 
	   is_mpi_waitsome(ro->region) ))
    {
      late_send* ls = new late_send(ro);
      
      ls->send = ls->recv.get_sendptr();
      ls->enter_send = ls->send.get_enterptr();
      ls->idle_time = ls->enter_send.get_time() - ls->enter_recv.get_time();
      if ( ls->idle_time > 0 )
	{
	  long loc_id = ls->enter_recv.get_loc()->get_id();
	  long cnode_id = ls->enter_recv.get_cnodeptr().get_pos();
	  
	  severity.add_val(cnode_id, loc_id, ls->idle_time);
	  return ls;
	}
      delete ls;
    }
  return NULL;
}   

/****************************************************************************
* Messages in Wrong Order (Late Sender)                                                                    
****************************************************************************/

bool
MsgsWrongOrderLS::isenabled() 
{ 
  return is_mpi_appl(trace);
}                                

void 
MsgsWrongOrderLS::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_cmpnd(LATE_SEND, this);
} 

cmp_event* 
MsgsWrongOrderLS::cmpnd(const cmp_event* ce) 
{ 
  const late_send* ls = (const late_send*) ce;

  long pos    = ls->recv.get_pos();
  long loc_id = ls->recv.get_loc()->get_id();
  
  vector<Event> queuev;
  trace->queue(queuev, pos, -1, loc_id);
  if ( ! queuev.empty() && queuev[0] < ls->send )
    {
      long loc_id   = ls->enter_recv.get_loc()->get_id();
      long cnode_id = ls->enter_recv.get_cnodeptr().get_pos();
      
      severity.add_val(cnode_id, loc_id, ls->idle_time);
    }
  return NULL;
} 

/****************************************************************************
* MPI Late Receiver                                                                                   
****************************************************************************/

bool 
LateReceiver::isenabled() 
{ 
  return is_mpi_appl(trace);
}                                

void 
LateReceiver::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_cmpnd(RECV_OP, this);
} 

cmp_event* 
LateReceiver::cmpnd(const cmp_event* ce) 
{ 
  const recv_op* ro = (const recv_op*) ce;
 
  if ( is_mpi_recv(ro->region) )
    {
      late_recv* lr = new late_recv(ro);
      
      lr->send = lr->recv.get_sendptr();
      lr->enter_send = lr->send.get_enterptr();
      if ( is_mpi_block_send(lr->enter_send.get_reg()) )
	{
	  vector<Event> stackv;

	  lr->idle_time = lr->enter_recv.get_time() - lr->enter_send.get_time();
	  trace->stack(stackv, lr->enter_recv.get_pos(), lr->enter_send.get_loc()->get_id());

	  if ( lr->idle_time > 0 && find(stackv.begin(), stackv.end(), lr->enter_send) != stackv.end() )  
	    {
	      long loc_id = lr->enter_send.get_loc()->get_id();
	      long cnode_id = lr->enter_send.get_cnodeptr().get_pos();
	  
	      severity.add_val(cnode_id, loc_id, lr->idle_time);
	      return lr;
	    }
	  delete lr;
	}
    }
  return NULL;
}   

/****************************************************************************
* Messages in Wrong Order (Late Receiver)                                             
****************************************************************************/

bool MsgsWrongOrderLR::isenabled() 
{ 
  return is_mpi_appl(trace);
}                                

void 
MsgsWrongOrderLR::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_cmpnd(LATE_RECV, this);
} 

cmp_event* 
MsgsWrongOrderLR::cmpnd(const cmp_event* ce) 
{ 
  const late_recv* lr = (const late_recv*) ce;

  long pos    = lr->recv.get_pos();
  long src_id = lr->recv.get_src()->get_id();
  
  vector<Event> queuev;
  trace->queue(queuev, pos, src_id, -1);
  if ( ! queuev.empty() && queuev[0] < lr->send )
    {
      long loc_id   = lr->enter_send.get_loc()->get_id();
      long cnode_id = lr->enter_send.get_cnodeptr().get_pos();
      
      severity.add_val(cnode_id, loc_id, lr->idle_time);
    }
  return NULL;
} 

/****************************************************************************
* Collective                                                                                
****************************************************************************/

bool 
Collective::isenabled()
{
  return is_mpi_appl(trace);
}
                        
void 
Collective::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      // if callee is collective communication function
      if ( is_mpi_collcom(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* Early Reduce                                                                               
****************************************************************************/

bool 
EarlyReduce::isenabled()
{
  return is_mpi_appl(trace);
}

void 
EarlyReduce::regcb(Analyzer& analyzer)
{
  analyzer.subscribe_cmpnd(MPI_COLL, this);
}

cmp_event* 
EarlyReduce::cmpnd(const cmp_event* ce)  
{
  const mpi_coll* mc = (const mpi_coll*) ce;
 
  Event enter = mc->enterv[0];
  if ( is_mpi_n21(enter.get_reg()) )
    {
      vector<double> sender_startv;
      
      Event root_enter = mc->enterv[mc->root_idx]; 
      for ( int i = 0; i < mc->enterv.size(); i++ )
	if ( mc->enterv[i] != root_enter )
	  sender_startv.push_back(mc->enterv[i].get_time());
      double idle_time = *min_element(sender_startv.begin(), sender_startv.end()) - root_enter.get_time();
      if ( idle_time > 0 )
	{
	  long loc_id   = root_enter.get_loc()->get_id();
	  long cnode_id = root_enter.get_cnodeptr().get_pos();
	  
	  severity.add_val(cnode_id, loc_id, idle_time);
	}
    }
  return NULL;
}

/****************************************************************************
* Late Broadcast                                                                              
****************************************************************************/

bool 
LateBroadcast::isenabled() 
{
  return is_mpi_appl(trace);
}                                

void 
LateBroadcast::regcb(Analyzer& analyzer) 
{
  analyzer.subscribe_cmpnd(MPI_COLL, this);
} 

cmp_event* 
LateBroadcast::cmpnd(const cmp_event* ce) 
{
  const mpi_coll* mc = (const mpi_coll*) ce;
 
  Event enter = mc->enterv[0];
  if ( is_mpi_12n(enter.get_reg()) )
    {
      double root_start = mc->enterv[mc->root_idx].get_time();
      for ( int i = 0; i < mc->enterv.size(); i++ )
	{
	  double idle_time = root_start - mc->enterv[i].get_time();
	  if ( idle_time > 0 )
	    {
	      long loc_id   = mc->enterv[i].get_loc()->get_id();
	      long cnode_id = mc->enterv[i].get_cnodeptr().get_pos();
	      
	      severity.add_val(cnode_id, loc_id, idle_time);	      
	    }
	}
    }
  return NULL;
}   

/****************************************************************************
* Wait at N to N                                                                            
****************************************************************************/

bool 
WaitAtN2N::isenabled()
{
  return is_mpi_appl(trace);
}                                

void 
WaitAtN2N::regcb(Analyzer& analyzer)
{
  analyzer.subscribe_cmpnd(MPI_COLL, this);
} 

cmp_event* 
WaitAtN2N::cmpnd(const cmp_event* ce)
{
  const mpi_coll* mc = (const mpi_coll*) ce;
 
  Event enter = mc->enterv[0];
  if ( is_mpi_n2n(enter.get_reg()) )
    {
      vector<double> start_timev;
      
      for ( int i = 0; i < mc->enterv.size(); i++ )
	start_timev.push_back(mc->enterv[i].get_time());
      double max_start_time = *max_element(start_timev.begin(), start_timev.end());
      for ( int i = 0; i < mc->enterv.size(); i++ )
	{
	  long loc_id      = mc->enterv[i].get_loc()->get_id();
	  long cnode_id    = mc->enterv[i].get_cnodeptr().get_pos();
	  double idle_time = max_start_time - start_timev[i];
	  
	  severity.add_val(cnode_id, loc_id, idle_time);	      
	}
    }
  return NULL;
}   

/****************************************************************************
*
* OpenMP                                                                                  
*
****************************************************************************/

/****************************************************************************
* OpenMP                                                                            
****************************************************************************/

bool 
OMP::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMP::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_cmpnd(TEAM_CREATE, this);
}

void 
OMP::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_omp_api(ctree.get_callee(cnode_id)) ||
	   is_omp_barrier(ctree.get_callee(cnode_id)) ||
	   is_omp_flush(ctree.get_callee(cnode_id)) ||
	   is_omp_critical(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

cmp_event* 
OMP::cmpnd(const cmp_event* ce) 
{
  const team_create* tc = (const team_create*) ce;
 
  long loc_id   = tc->enter.get_loc()->get_id();
  long cnode_id = tc->enter.get_cedgeptr().get_pos();

  severity.add_val(cnode_id, loc_id, tc->time);
  return NULL;
} 

/****************************************************************************
* OMP Fork                                                                           
****************************************************************************/

bool 
OMPFork::isenabled() 
{ 
  return is_omp_appl(trace);
}
                               
void 
OMPFork::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_event(ENTER, this);
  analyzer.subscribe_event(FORK, this);
}

cmp_event* 
OMPFork::enter(earl::Event& event)
{
  if ( is_omp_parallel(event.get_reg()) )
    {
      long loc_id   = event.get_loc()->get_id();
      long proc_id  = event.get_loc()->get_proc()->get_id();
      long thrd_id  = event.get_loc()->get_thrd()->get_thrd_id();
      if ( thrd_id != 0 )
	return NULL;
      if ( ! event.get_cedgeptr().null() )
	{
	  long cnode_id = event.get_cedgeptr().get_pos();
	  team_create* tc = new team_create();
	  tc->enter = event;
	  if ( forkv[proc_id].null() )
	    {
	      cerr << "EXPERT: Error in OMPFork::enter(earl::Event& event)." << endl;
	      std::exit(EXIT_FAILURE);
	  }
	  tc->time  = event.get_time() - forkv[proc_id].get_time();
	  severity.add_val(cnode_id, loc_id, tc->time);
	  return tc;
	  
	}
    }
  return NULL;
}

cmp_event* 
OMPFork::fork(earl::Event& event)
{
  long proc_id  = event.get_loc()->get_proc()->get_id();
  forkv[proc_id] = event;
  return NULL;
}

/****************************************************************************
* OMP Flush                                                                          
****************************************************************************/

bool 
OMPFlush::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPFlush::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_omp_flush(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* OMP Synchronization                                                                          
****************************************************************************/

bool 
OMPSynchronization::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPSynchronization::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_omp_lock(ctree.get_callee(cnode_id)) || 
	   is_omp_barrier(ctree.get_callee(cnode_id)) || 
	   is_omp_critical(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* OMP Barrier                                                                         
****************************************************************************/

bool 
OMPBarrier::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPBarrier::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_event(OMPCEXIT, this);
}

void 
OMPBarrier::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_omp_barrier(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

cmp_event* 
OMPBarrier::ompcexit(earl::Event& event)
{
  vector<Event> coll;

  trace->ompcoll(coll, event.get_pos());

  // if collective operation complete
  if ( coll.size() > 0 )
    {
      omp_coll* oc = new omp_coll();
      
      oc->region = event.get_reg();
      oc->exitv  = coll;
      for ( long i = 0; i < oc->exitv.size(); i++ )
	oc->enterv.push_back(oc->exitv[i].get_enterptr());
      return oc;
    }
  return NULL;
} 

/****************************************************************************
* OMP Explicit Barrier                                                                         
****************************************************************************/

bool 
OMPEBarrier::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPEBarrier::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_omp_ebarrier(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* Wait at Barrier (OMP, explicit)                                                                         
****************************************************************************/

bool 
OMPWaitAtEBarrier::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPWaitAtEBarrier::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_cmpnd(OMP_COLL, this);
}

cmp_event* 
OMPWaitAtEBarrier::cmpnd(const cmp_event* ce) 
{ 
  const omp_coll* oc = (const omp_coll*) ce;
 
  Event enter = oc->enterv[0];
  if ( is_omp_ebarrier(enter.get_reg()) )
    {
      vector<double> start_timev;
      
      for ( int i = 0; i < oc->enterv.size(); i++ )
	start_timev.push_back(oc->enterv[i].get_time());
      double max_start_time = *max_element(start_timev.begin(), start_timev.end());
      for ( int i = 0; i < oc->enterv.size(); i++ )
	{
	  long loc_id       = oc->enterv[i].get_loc()->get_id();
	  long cnode_id     = oc->enterv[i].get_cnodeptr().get_pos();
	  double idle_time = max_start_time - start_timev[i];
	  
	  severity.add_val(cnode_id, loc_id, idle_time);	      
	}
    }
  return NULL;
} 

/****************************************************************************
* OMP Implicit Barrier                                                                         
****************************************************************************/

bool 
OMPIBarrier::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPIBarrier::end(Profile& profile, CallTree& ctree)
{
  vector<long> cnode_idv;
  
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < cnode_idv.size(); i++ )
    { 
      long cnode_id =  cnode_idv[i];
      if ( is_omp_ibarrier(ctree.get_callee(cnode_id)) )
	for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
	  severity.set_val(cnode_id, loc_id, profile.get_time(cnode_id, loc_id));
    }
}

/****************************************************************************
* Wait at Barrier (OMP, implicit)                                                                         
****************************************************************************/

bool 
OMPWaitAtIBarrier::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPWaitAtIBarrier::regcb(Analyzer& analyzer) 
{
  analyzer.subscribe_cmpnd(OMP_COLL, this);
}

cmp_event* 
OMPWaitAtIBarrier::cmpnd(const cmp_event* ce) 
{ 
  const omp_coll* oc = (const omp_coll*) ce;
 
  Event enter = oc->enterv[0];
  if ( is_omp_ibarrier(enter.get_reg()) )
    {
      vector<double> start_timev;
      
      for ( int i = 0; i < oc->enterv.size(); i++ )
	start_timev.push_back(oc->enterv[i].get_time());
      double max_start_time = *max_element(start_timev.begin(), start_timev.end());
      for ( int i = 0; i < oc->enterv.size(); i++ )
	{
	  long loc_id       = oc->enterv[i].get_loc()->get_id();
	  long cnode_id     = oc->enterv[i].get_cnodeptr().get_pos();
	  double idle_time = max_start_time - start_timev[i];
	  
	  severity.add_val(cnode_id, loc_id, idle_time);	      
	}
    }
  return NULL;
}

/****************************************************************************
* OMP Lock                                                                     
****************************************************************************/

bool 
OMPLock::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPLock::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_event(ALOCK, this);
}

cmp_event* 
OMPLock::alock(earl::Event& event) 
{ 
  if ( ! event.get_lockptr().null() )
    {
      lock_comp* lc = new lock_comp();
  
      lc->alock = event;
      lc->enter_alock = event.get_enterptr();
      lc->rlock = event.get_lockptr();
      lc->idle_time = lc->rlock.get_time() - lc->alock.get_time();
      if ( lc->idle_time > 0)
	{
	  long loc_id   = lc->enter_alock.get_loc()->get_id();
	  long cnode_id = lc->enter_alock.get_cnodeptr().get_pos();
	  
	  severity.add_val(cnode_id, loc_id, lc->idle_time);	      
	  return lc;
	}
      delete lc;
    }
  return NULL;
}

/****************************************************************************
* OMP Lock (API)                                                                     
****************************************************************************/

bool 
OMPAPILock::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPAPILock::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_cmpnd(LOCK_COMP, this);
}

cmp_event* 
OMPAPILock::cmpnd(const cmp_event* ce) 
{ 
  const lock_comp* lc = (const lock_comp*) ce;

  if ( is_function(lc->enter_alock.get_reg()) )
    {
      long loc_id   = lc->enter_alock.get_loc()->get_id();
      long cnode_id = lc->enter_alock.get_cnodeptr().get_pos();
      
      severity.add_val(cnode_id, loc_id, lc->idle_time);	      
    }
  return NULL;
}

/****************************************************************************
* OMP Lock (Critical Section)                                                                     
****************************************************************************/

bool 
OMPCriticalLock::isenabled() 
{ 
  return is_omp_appl(trace);
}                               

void 
OMPCriticalLock::regcb(Analyzer& analyzer) 
{ 
  analyzer.subscribe_cmpnd(LOCK_COMP, this);
}

cmp_event* 
OMPCriticalLock::cmpnd(const cmp_event* ce) 
{ 
   const lock_comp* lc = (const lock_comp*) ce;

  if ( is_omp_critical(lc->enter_alock.get_reg()) )
    {
      long loc_id   = lc->enter_alock.get_loc()->get_id();
      long cnode_id = lc->enter_alock.get_cnodeptr().get_pos();
      
      severity.add_val(cnode_id, loc_id, lc->idle_time);	      
    }
  return NULL;
}

/****************************************************************************
*
* Auxiliary functions                                                                                       
*
****************************************************************************/

/****************************************************************************
* MPI
****************************************************************************/

bool is_mpi_appl(const EventTrace* trace)
{
  return trace->get_nprocs() > 1;
}

bool is_mpi_api(const Region* region)
{
  if ( region->get_name().length() < 4 )
    return false;
  
  string prefix(region->get_name(), 0, 3);

  return lower(prefix) == "mpi";
}

bool is_mpi_barrier(const Region* region)
{
  if ( region->get_name().length() < 11 )
    return false;

  string prefix(region->get_name(), 0, 11);

  return lower(prefix) == "mpi_barrier";
}

bool is_mpi_block_send(const Region* region)
{
   return lower(region->get_name()) == "mpi_send" ||  lower(region->get_name()) == "mpi_ssend";
}

bool is_mpi_collcom(const Region* region)
{
  if ( ! is_mpi_api(region) )
    return false;

  string suffix = lower(region->get_name().substr(3)); 

  if ( suffix.find("_all") != string::npos ||
       suffix.find("_op") != string::npos ||
       suffix.find("bcast") != string::npos ||
       suffix.find("gather") != string::npos ||
       suffix.find("reduce") != string::npos ||
       suffix.find("scan") != string::npos ||
       suffix.find("scatter") != string::npos )
    return true;
  else
    return false;
}

bool is_mpi_io(const Region* region)
{
  if ( region->get_name().length() < 8 )
    return false;

  string prefix(region->get_name(), 0, 8);

  return lower(prefix) == "mpi_file";
}

bool is_mpi_12n(const Region* region)
{
   if ( ! is_mpi_api(region) )
    return false;

  string suffix = lower(region->get_name().substr(4));

  if ( suffix == "bcast" ||
       suffix == "scatter" ||
       suffix == "scatterv" )
    return true;
  else
    return false;  
}

bool is_mpi_n21(const Region* region)
{
  if ( ! is_mpi_api(region) )
    return false;
  
  string suffix = lower(region->get_name().substr(4));

  if ( suffix == "gather" ||
       suffix == "gatherv" ||
       suffix == "reduce" )
    return true;
  else
    return false;  
}

bool is_mpi_n2n(const Region* region)
{
  if ( ! is_mpi_api(region) )
    return false;

  string suffix = lower(region->get_name().substr(3));

  if ( suffix.find("_all") != string::npos || suffix == "_reduce_scatter" )
    return true;
  else
    return false;  
}


bool is_mpi_p2p(const Region* region)
{
  if ( ! is_mpi_api(region) )
    return false;

  string suffix = lower(region->get_name().substr(4));
  
  if ( suffix.find("buffer") != string::npos ||
       suffix.find("cancel") != string::npos ||
       suffix.find("get_count") != string::npos ||
       suffix.find("probe") != string::npos ||
       suffix.find("recv") != string::npos ||
       suffix.find("request") != string::npos ||
       suffix.find("send") != string::npos ||
       suffix.find("start") != string::npos ||
       suffix.find("i_test") != string::npos || 
       suffix.find("wait") != string::npos )
    return true;
  else
    return false;
}

bool is_mpi_recv(const Region* region)
{
  return lower(region->get_name()) == "mpi_recv";
}
    
bool is_mpi_testx(const Region* region)
{
  if ( region->get_name().length() < 8 )
    return false;

  string prefix(region->get_name(), 0, 8);

  return prefix == "mpi_test";
}

bool is_mpi_waitall(const Region* region)
{
  if ( region->get_name().length() != 11 )
    return false;

  return lower(region->get_name()) == "mpi_waitall";
}

bool is_mpi_waitsome(const Region* region)
{
  if ( region->get_name().length() != 12 )
    return false;

  return lower(region->get_name()) == "mpi_waitsome";
}


/****************************************************************************
* OpenMP
****************************************************************************/

bool is_omp_appl(const EventTrace* trace)
{
  return trace->get_nthrds() > trace->get_nprocs();
}

bool is_omp_api(const Region* region) 
{
  if ( region->get_name().length() < 4 )
    return false;
  
  string prefix(region->get_name(), 0, 3);

  return lower(prefix) == "omp";
}

bool is_omp_lock(const Region* region) 
{
  if ( ! is_mpi_api(region) )
    return false;

  string suffix = region->get_name().substr(region->get_name().length() - 4);
  
  return suffix == "lock";
}

bool is_omp_barrier(const Region* region) 
{
  return region->get_rtype() == "OMP_BARRIER" || region->get_rtype() == "OMP_IBARRIER";
}

bool is_omp_ibarrier(const Region* region)
{
  return region->get_rtype() == "OMP_IBARRIER"; 
}

bool is_omp_ebarrier(const Region* region) 
{
  return region->get_rtype() == "OMP_BARRIER";
} 

bool is_omp_flush(const Region* region) 
{
  return region->get_rtype() == "OMP_FLUSH"; 
}

bool is_omp_critical(const Region* region) 
{
  return region->get_rtype() == "OMP_CRITICAL"; 
}

bool is_omp_critical_sblock(const Region* region) 
{
  return region->get_rtype() == "OMP_CRITICAL_SBLOCK"; 
}

bool is_omp_parallel(const Region* region) 
{
  return region->get_rtype() == "OMP_PARALLEL"; 
}

bool is_function(const Region* region) 
{
  return region->get_rtype() == "FUNCTION"; 
}


/****************************************************************************
* Miscellaneous
****************************************************************************/

int get_met_id(string met_name, const EventTrace* trace)
{
  for ( int i = 0; i < trace->get_nmets(); i++ )
    {
      if ( met_name == trace->get_met(i)->get_name() )
        return i;
    }
  return -1;
}

bool is_met_avail(string met_name, const EventTrace* trace)
{
  if ( get_met_id(met_name, trace) == -1 )
    return false;
  else
    return true;
}

string lower(string str)
{
  string result = str;

  for (int i = 0; i < result.length(); i++ )
    result[i] = tolower(result[i]);

  return result;
}
