/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (C) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "Report.h"

#include "CallTree.h"
#include "Pattern.h"
#include "PatternTree.h"

using namespace std;

Report::Report(const PatternTree& ptree, 
	       const CallTree&    ctree,
	       const earl::EventTrace& trace)
{
  // mapping of EXPERT IDs onto CUBE objects

  map<long, const cube::Metric*> patm;
  map<string, const cube::Module*> modm;
  map<const earl::Region*, const cube::Region*> regm;
  map<const earl::Callsite*, const cube::Callsite*> csitem;
  map<long, const cube::Cnode*> cnodem;
  map<const earl::Machine*, const cube::Machine*> machm;
  map<const earl::Node*, const cube::Node*> nodem;
  map<const earl::Process*, const cube::Process*> procm;
  map<const earl::Thread*, const cube::Thread*> thrdm;
  map<const earl::Location*, const cube::Thread*> locm;

  vector<long> keyv;

  // define mirror sites

  cb.def_mirror("http://icl.cs.utk.edu/projectsfiles/kojak/software/kojak/");
  cb.def_mirror("http://www.fz-juelich.de/zam/kojak/");

  // define CUBE aggregation mode

  cb.def_attr("CUBE_CT_AGGR", "SUM");

  // specify mapping of absent information

  patm[-1] = NULL;
  modm["UNKNOWN"] = cb.def_module("UNKNOWN");
  cnodem[-1] = NULL;

  // define CUBE metrics

  ptree.get_dfkeyv(keyv);
  for ( long i = 0; i < keyv.size(); i++ )
    {
      Pattern* pat = ptree.get_data(keyv[i]);
      patm[pat->get_id()] = cb.def_met(pat->get_name(), pat->get_uom(), pat->get_url(), pat->get_descr(), patm[pat->get_parent_id()]);
    }

  // define CUBE modules

  for ( long file_id = 0; file_id < trace.get_nfiles(); file_id++ )
    {
      string file = trace.get_file(file_id);
      modm[file] = cb.def_module(file);
    }

  // define CUBE regions

  for ( long reg_id = 0; reg_id < trace.get_nregs(); reg_id++ )
    {
      const earl::Region* reg = trace.get_reg(reg_id);
      regm[reg] = cb.def_region(reg->get_name(), 
				reg->get_begln(), 
				reg->get_endln(),
				"",
				reg->get_descr(), 
				modm[reg->get_file()]);
    }
  
  // define CUBE call sites

  for ( long csite_id = 0; csite_id < trace.get_ncsites(); csite_id++ )
    {
      const earl::Callsite* csite = trace.get_csite(csite_id);
      csitem[csite] = cb.def_csite(modm[csite->get_file()], 
				   csite->get_line(), 
				   regm[csite->get_callee()]);
    }

  // define CUBE call paths (i.e., call-tree nodes)

  ctree.get_dfkeyv(keyv);
  for ( long i = 0; i < keyv.size(); i++ )
    {
      const earl::Callsite* csite  = ctree.get_csite(keyv[i]);
      const earl::Region*   callee = ctree.get_callee(keyv[i]);

      const cube::Callsite* cube_csite;

      if ( csite == NULL )
	cube_csite = cb.def_csite(modm["UNKNOWN"], -1, regm[callee]);
      else
	cube_csite = csitem[csite];

      cnodem[keyv[i]] = cb.def_cnode(cube_csite, cnodem[ctree.get_parent(keyv[i])]);
    }

  // define CUBE machines and nodes

  for ( long mach_id = 0; mach_id < trace.get_nmachs(); mach_id++ )
    {
      const earl::Machine* mach = trace.get_mach(mach_id);
      machm[mach] = cb.def_mach(mach->get_name());
				     
      for ( long node_id = 0; node_id < mach->get_nnodes(); node_id++ )
	{
	  const earl::Node* node = trace.get_node(mach_id, node_id);
	  nodem[node] = cb.def_node(node->get_name(), machm[mach]);
	}				       
    }
  
  // define CUBE processes and threads

  for ( long proc_id = 0; proc_id < trace.get_nprocs(); proc_id++ )
    {
      const earl::Process* proc = trace.get_proc(proc_id);
      procm[proc] = cb.def_proc(proc->get_name(),
				proc->get_id(),
				nodem[proc->get_node()]);
      
      for ( long thrd_id = 0; thrd_id < proc->get_nthrds(); thrd_id++ )
	{
	  const earl::Thread* thrd = trace.get_thrd(proc_id, thrd_id);
	  thrdm[thrd] = cb.def_thrd(thrd->get_name(), 
				    thrd->get_thrd_id(), 
				    procm[proc]);
	  
	  locm[thrd->get_loc()] = thrdm[thrd];
	}				       
    }

  // define CUBE severity
  
  vector<long> pat_idv, cnode_idv;

  ptree.get_keyv(pat_idv);
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < pat_idv.size(); i++ )
    {
      long pat_id = pat_idv[i];

      for ( long j = 0; j < cnode_idv.size(); j++ )
	{
	  long cnode_id = cnode_idv[j];

	  for ( long loc_id = 0; loc_id < trace.get_nlocs(); loc_id++ )
	    cb.set_sev(patm[pat_id], 
		       cnodem[cnode_id], 
		       locm[trace.get_loc(loc_id)], 
		       ptree.get_data(pat_id)->get_sev(cnode_id, loc_id));
	}
    }
}

void 
Report::write(std::string path)
{
  cb.write(path);
}


